#include "miniz.h"

#ifndef MINIZ_HEADER_FILE_ONLY

typedef unsigned char mz_validate_uint16[sizeof(mz_uint16) == 2 ? 1 : -1];
typedef unsigned char mz_validate_uint32[sizeof(mz_uint32) == 4 ? 1 : -1];
typedef unsigned char mz_validate_uint64[sizeof(mz_uint64) == 8 ? 1 : -1];

#  include <assert.h>
#  include <cstring>
#  include <string.h>

#  define MZ_ASSERT(x) assert(x)

#  ifdef MINIZ_NO_MALLOC
#    define MZ_MALLOC(x) NULL
#    define MZ_FREE(x) (void)x, ((void)0)
#    define MZ_REALLOC(p, x) NULL
#  else
#    define MZ_MALLOC(x) malloc(x)
#    define MZ_FREE(x) free(x)
#    define MZ_REALLOC(p, x) realloc(p, x)
#  endif

#  define MZ_MAX(a, b) (((a) > (b)) ? (a) : (b))
#  define MZ_MIN(a, b) (((a) < (b)) ? (a) : (b))
#  define MZ_CLEAR_OBJ(obj) memset(&(obj), 0, sizeof(obj))

#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN
#    define MZ_READ_LE16(p) *((const mz_uint16*)(p))
#    define MZ_READ_LE32(p) *((const mz_uint32*)(p))
#  else
#    define MZ_READ_LE16(p) \
      ((mz_uint32)(((const mz_uint8*)(p))[0]) | \
       ((mz_uint32)(((const mz_uint8*)(p))[1]) << 8U))
#    define MZ_READ_LE32(p) \
      ((mz_uint32)(((const mz_uint8*)(p))[0]) | \
       ((mz_uint32)(((const mz_uint8*)(p))[1]) << 8U) | \
       ((mz_uint32)(((const mz_uint8*)(p))[2]) << 16U) | \
       ((mz_uint32)(((const mz_uint8*)(p))[3]) << 24U))
#  endif

#  ifdef _MSC_VER
#    define MZ_FORCEINLINE __forceinline
#  elif defined(__GNUC__)
#    define MZ_FORCEINLINE inline __attribute__((__always_inline__))
#  else
#    define MZ_FORCEINLINE inline
#  endif

#  ifdef __cplusplus
extern "C"
{
#  endif

  // ------------------- zlib-style API's

  mz_ulong mz_adler32(mz_ulong adler, const unsigned char* ptr, size_t buf_len)
  {
    mz_uint32 i, s1 = (mz_uint32)(adler & 0xffff),
                 s2 = (mz_uint32)(adler >> 16);
    size_t block_len = buf_len % 5552;
    if (!ptr)
      return MZ_ADLER32_INIT;
    while (buf_len)
    {
      for (i = 0; i + 7 < block_len; i += 8, ptr += 8)
      {
        s1 += ptr[0], s2 += s1;
        s1 += ptr[1], s2 += s1;
        s1 += ptr[2], s2 += s1;
        s1 += ptr[3], s2 += s1;
        s1 += ptr[4], s2 += s1;
        s1 += ptr[5], s2 += s1;
        s1 += ptr[6], s2 += s1;
        s1 += ptr[7], s2 += s1;
      }
      for (; i < block_len; ++i)
        s1 += *ptr++, s2 += s1;
      s1 %= 65521U, s2 %= 65521U;
      buf_len -= block_len;
      block_len = 5552;
    }
    return (s2 << 16) + s1;
  }

  // Karl Malbrain's compact CRC-32. See "A compact CCITT crc16 and crc32 C
  // implementation that balances processor cache usage against speed":
  // http://www.geocities.com/malbrain/
  mz_ulong mz_crc32(mz_ulong crc, const mz_uint8* ptr, size_t buf_len)
  {
    static const mz_uint32 s_crc32[16] = {
      0,
      0x1db71064,
      0x3b6e20c8,
      0x26d930ac,
      0x76dc4190,
      0x6b6b51f4,
      0x4db26158,
      0x5005713c,
      0xedb88320,
      0xf00f9344,
      0xd6d6a3e8,
      0xcb61b38c,
      0x9b64c2b0,
      0x86d3d2d4,
      0xa00ae278,
      0xbdbdf21c};
    mz_uint32 crcu32 = (mz_uint32)crc;
    if (!ptr)
      return MZ_CRC32_INIT;
    crcu32 = ~crcu32;
    while (buf_len--)
    {
      mz_uint8 b = *ptr++;
      crcu32 = (crcu32 >> 4) ^ s_crc32[(crcu32 & 0xF) ^ (b & 0xF)];
      crcu32 = (crcu32 >> 4) ^ s_crc32[(crcu32 & 0xF) ^ (b >> 4)];
    }
    return ~crcu32;
  }

  void mz_free(void* p)
  {
    MZ_FREE(p);
  }

#  ifndef MINIZ_NO_ZLIB_APIS

  static void* def_alloc_func(void* opaque, size_t items, size_t size)
  {
    (void)opaque, (void)items, (void)size;
    return MZ_MALLOC(items * size);
  }
  static void def_free_func(void* opaque, void* address)
  {
    (void)opaque, (void)address;
    MZ_FREE(address);
  }
  static void*
  def_realloc_func(void* opaque, void* address, size_t items, size_t size)
  {
    (void)opaque, (void)address, (void)items, (void)size;
    return MZ_REALLOC(address, items * size);
  }

  const char* mz_version(void)
  {
    return MZ_VERSION;
  }

  int mz_deflateInit(mz_streamp pStream, int level)
  {
    return mz_deflateInit2(
      pStream,
      level,
      MZ_DEFLATED,
      MZ_DEFAULT_WINDOW_BITS,
      9,
      MZ_DEFAULT_STRATEGY);
  }

  int mz_deflateInit2(
    mz_streamp pStream,
    int level,
    int method,
    int window_bits,
    int mem_level,
    int strategy)
  {
    tdefl_compressor* pComp;
    mz_uint comp_flags = TDEFL_COMPUTE_ADLER32 |
      tdefl_create_comp_flags_from_zip_params(level, window_bits, strategy);

    if (!pStream)
      return MZ_STREAM_ERROR;
    if (
      (method != MZ_DEFLATED) || ((mem_level < 1) || (mem_level > 9)) ||
      ((window_bits != MZ_DEFAULT_WINDOW_BITS) &&
       (-window_bits != MZ_DEFAULT_WINDOW_BITS)))
      return MZ_PARAM_ERROR;

    pStream->data_type = 0;
    pStream->adler = MZ_ADLER32_INIT;
    pStream->msg = NULL;
    pStream->reserved = 0;
    pStream->total_in = 0;
    pStream->total_out = 0;
    if (!pStream->zalloc)
      pStream->zalloc = def_alloc_func;
    if (!pStream->zfree)
      pStream->zfree = def_free_func;

    pComp = (tdefl_compressor*)pStream->zalloc(
      pStream->opaque, 1, sizeof(tdefl_compressor));
    if (!pComp)
      return MZ_MEM_ERROR;

    pStream->state = (struct mz_internal_state*)pComp;

    if (tdefl_init(pComp, NULL, NULL, comp_flags) != TDEFL_STATUS_OKAY)
    {
      mz_deflateEnd(pStream);
      return MZ_PARAM_ERROR;
    }

    return MZ_OK;
  }

  int mz_deflateReset(mz_streamp pStream)
  {
    if (
      (!pStream) || (!pStream->state) || (!pStream->zalloc) ||
      (!pStream->zfree))
      return MZ_STREAM_ERROR;
    pStream->total_in = pStream->total_out = 0;
    tdefl_init(
      (tdefl_compressor*)pStream->state,
      NULL,
      NULL,
      ((tdefl_compressor*)pStream->state)->m_flags);
    return MZ_OK;
  }

  int mz_deflate(mz_streamp pStream, int flush)
  {
    size_t in_bytes, out_bytes;
    mz_ulong orig_total_in, orig_total_out;
    int mz_status = MZ_OK;

    if (
      (!pStream) || (!pStream->state) || (flush < 0) || (flush > MZ_FINISH) ||
      (!pStream->next_out))
      return MZ_STREAM_ERROR;
    if (!pStream->avail_out)
      return MZ_BUF_ERROR;

    if (flush == MZ_PARTIAL_FLUSH)
      flush = MZ_SYNC_FLUSH;

    if (
      ((tdefl_compressor*)pStream->state)->m_prev_return_status ==
      TDEFL_STATUS_DONE)
      return (flush == MZ_FINISH) ? MZ_STREAM_END : MZ_BUF_ERROR;

    orig_total_in = pStream->total_in;
    orig_total_out = pStream->total_out;
    for (;;)
    {
      tdefl_status defl_status;
      in_bytes = pStream->avail_in;
      out_bytes = pStream->avail_out;

      defl_status = tdefl_compress(
        (tdefl_compressor*)pStream->state,
        pStream->next_in,
        &in_bytes,
        pStream->next_out,
        &out_bytes,
        (tdefl_flush)flush);
      pStream->next_in += (mz_uint)in_bytes;
      pStream->avail_in -= (mz_uint)in_bytes;
      pStream->total_in += (mz_uint)in_bytes;
      pStream->adler = tdefl_get_adler32((tdefl_compressor*)pStream->state);

      pStream->next_out += (mz_uint)out_bytes;
      pStream->avail_out -= (mz_uint)out_bytes;
      pStream->total_out += (mz_uint)out_bytes;

      if (defl_status < 0)
      {
        mz_status = MZ_STREAM_ERROR;
        break;
      }
      else if (defl_status == TDEFL_STATUS_DONE)
      {
        mz_status = MZ_STREAM_END;
        break;
      }
      else if (!pStream->avail_out)
        break;
      else if ((!pStream->avail_in) && (flush != MZ_FINISH))
      {
        if (
          (flush) || (pStream->total_in != orig_total_in) ||
          (pStream->total_out != orig_total_out))
          break;
        return MZ_BUF_ERROR; // Can't make forward progress without some input.
      }
    }
    return mz_status;
  }

  int mz_deflateEnd(mz_streamp pStream)
  {
    if (!pStream)
      return MZ_STREAM_ERROR;
    if (pStream->state)
    {
      pStream->zfree(pStream->opaque, pStream->state);
      pStream->state = NULL;
    }
    return MZ_OK;
  }

  mz_ulong mz_deflateBound(mz_streamp pStream, mz_ulong source_len)
  {
    (void)pStream;
    // This is really over conservative. (And lame, but it's actually pretty
    // tricky to compute a true upper bound given the way tdefl's blocking
    // works.)
    return MZ_MAX(
      128 + (source_len * 110) / 100,
      128 + source_len + ((source_len / (31 * 1024)) + 1) * 5);
  }

  int mz_compress2(
    unsigned char* pDest,
    mz_ulong* pDest_len,
    const unsigned char* pSource,
    mz_ulong source_len,
    int level)
  {
    int status;
    mz_stream stream;
    memset(&stream, 0, sizeof(stream));

    // In case mz_ulong is 64-bits (argh I hate longs).
    if ((source_len | *pDest_len) > 0xFFFFFFFFU)
      return MZ_PARAM_ERROR;

    stream.next_in = pSource;
    stream.avail_in = (mz_uint32)source_len;
    stream.next_out = pDest;
    stream.avail_out = (mz_uint32)*pDest_len;

    status = mz_deflateInit(&stream, level);
    if (status != MZ_OK)
      return status;

    status = mz_deflate(&stream, MZ_FINISH);
    if (status != MZ_STREAM_END)
    {
      mz_deflateEnd(&stream);
      return (status == MZ_OK) ? MZ_BUF_ERROR : status;
    }

    *pDest_len = stream.total_out;
    return mz_deflateEnd(&stream);
  }

  int mz_compress(
    unsigned char* pDest,
    mz_ulong* pDest_len,
    const unsigned char* pSource,
    mz_ulong source_len)
  {
    return mz_compress2(
      pDest, pDest_len, pSource, source_len, MZ_DEFAULT_COMPRESSION);
  }

  mz_ulong mz_compressBound(mz_ulong source_len)
  {
    return mz_deflateBound(NULL, source_len);
  }

  typedef struct
  {
    tinfl_decompressor m_decomp;
    mz_uint m_dict_ofs, m_dict_avail, m_first_call, m_has_flushed;
    int m_window_bits;
    mz_uint8 m_dict[TINFL_LZ_DICT_SIZE];
    tinfl_status m_last_status;
  } inflate_state;

  int mz_inflateInit2(mz_streamp pStream, int window_bits)
  {
    inflate_state* pDecomp;
    if (!pStream)
      return MZ_STREAM_ERROR;
    if (
      (window_bits != MZ_DEFAULT_WINDOW_BITS) &&
      (-window_bits != MZ_DEFAULT_WINDOW_BITS))
      return MZ_PARAM_ERROR;

    pStream->data_type = 0;
    pStream->adler = 0;
    pStream->msg = NULL;
    pStream->total_in = 0;
    pStream->total_out = 0;
    pStream->reserved = 0;
    if (!pStream->zalloc)
      pStream->zalloc = def_alloc_func;
    if (!pStream->zfree)
      pStream->zfree = def_free_func;

    pDecomp = (inflate_state*)pStream->zalloc(
      pStream->opaque, 1, sizeof(inflate_state));
    if (!pDecomp)
      return MZ_MEM_ERROR;

    pStream->state = (struct mz_internal_state*)pDecomp;

    tinfl_init(&pDecomp->m_decomp);
    pDecomp->m_dict_ofs = 0;
    pDecomp->m_dict_avail = 0;
    pDecomp->m_last_status = TINFL_STATUS_NEEDS_MORE_INPUT;
    pDecomp->m_first_call = 1;
    pDecomp->m_has_flushed = 0;
    pDecomp->m_window_bits = window_bits;

    return MZ_OK;
  }

  int mz_inflateInit(mz_streamp pStream)
  {
    return mz_inflateInit2(pStream, MZ_DEFAULT_WINDOW_BITS);
  }

  int mz_inflate(mz_streamp pStream, int flush)
  {
    inflate_state* pState;
    mz_uint n, first_call, decomp_flags = TINFL_FLAG_COMPUTE_ADLER32;
    size_t in_bytes, out_bytes, orig_avail_in;
    tinfl_status status;

    if ((!pStream) || (!pStream->state))
      return MZ_STREAM_ERROR;
    if (flush == MZ_PARTIAL_FLUSH)
      flush = MZ_SYNC_FLUSH;
    if ((flush) && (flush != MZ_SYNC_FLUSH) && (flush != MZ_FINISH))
      return MZ_STREAM_ERROR;

    pState = (inflate_state*)pStream->state;
    if (pState->m_window_bits > 0)
      decomp_flags |= TINFL_FLAG_PARSE_ZLIB_HEADER;
    orig_avail_in = pStream->avail_in;

    first_call = pState->m_first_call;
    pState->m_first_call = 0;
    if (pState->m_last_status < 0)
      return MZ_DATA_ERROR;

    if (pState->m_has_flushed && (flush != MZ_FINISH))
      return MZ_STREAM_ERROR;
    pState->m_has_flushed |= (flush == MZ_FINISH);

    if ((flush == MZ_FINISH) && (first_call))
    {
      // MZ_FINISH on the first call implies that the input and output buffers
      // are large enough to hold the entire compressed/decompressed file.
      decomp_flags |= TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF;
      in_bytes = pStream->avail_in;
      out_bytes = pStream->avail_out;
      status = tinfl_decompress(
        &pState->m_decomp,
        pStream->next_in,
        &in_bytes,
        pStream->next_out,
        pStream->next_out,
        &out_bytes,
        decomp_flags);
      pState->m_last_status = status;
      pStream->next_in += (mz_uint)in_bytes;
      pStream->avail_in -= (mz_uint)in_bytes;
      pStream->total_in += (mz_uint)in_bytes;
      pStream->adler = tinfl_get_adler32(&pState->m_decomp);
      pStream->next_out += (mz_uint)out_bytes;
      pStream->avail_out -= (mz_uint)out_bytes;
      pStream->total_out += (mz_uint)out_bytes;

      if (status < 0)
        return MZ_DATA_ERROR;
      else if (status != TINFL_STATUS_DONE)
      {
        pState->m_last_status = TINFL_STATUS_FAILED;
        return MZ_BUF_ERROR;
      }
      return MZ_STREAM_END;
    }
    // flush != MZ_FINISH then we must assume there's more input.
    if (flush != MZ_FINISH)
      decomp_flags |= TINFL_FLAG_HAS_MORE_INPUT;

    if (pState->m_dict_avail)
    {
      n = MZ_MIN(pState->m_dict_avail, pStream->avail_out);
      memcpy(pStream->next_out, pState->m_dict + pState->m_dict_ofs, n);
      pStream->next_out += n;
      pStream->avail_out -= n;
      pStream->total_out += n;
      pState->m_dict_avail -= n;
      pState->m_dict_ofs = (pState->m_dict_ofs + n) & (TINFL_LZ_DICT_SIZE - 1);
      return ((pState->m_last_status == TINFL_STATUS_DONE) &&
              (!pState->m_dict_avail)) ?
        MZ_STREAM_END :
        MZ_OK;
    }

    for (;;)
    {
      in_bytes = pStream->avail_in;
      out_bytes = TINFL_LZ_DICT_SIZE - pState->m_dict_ofs;

      status = tinfl_decompress(
        &pState->m_decomp,
        pStream->next_in,
        &in_bytes,
        pState->m_dict,
        pState->m_dict + pState->m_dict_ofs,
        &out_bytes,
        decomp_flags);
      pState->m_last_status = status;

      pStream->next_in += (mz_uint)in_bytes;
      pStream->avail_in -= (mz_uint)in_bytes;
      pStream->total_in += (mz_uint)in_bytes;
      pStream->adler = tinfl_get_adler32(&pState->m_decomp);

      pState->m_dict_avail = (mz_uint)out_bytes;

      n = MZ_MIN(pState->m_dict_avail, pStream->avail_out);
      memcpy(pStream->next_out, pState->m_dict + pState->m_dict_ofs, n);
      pStream->next_out += n;
      pStream->avail_out -= n;
      pStream->total_out += n;
      pState->m_dict_avail -= n;
      pState->m_dict_ofs = (pState->m_dict_ofs + n) & (TINFL_LZ_DICT_SIZE - 1);

      if (status < 0)
        return MZ_DATA_ERROR; // Stream is corrupted (there could be some
                              // uncompressed data left in the output dictionary
                              // - oh well).
      else if ((status == TINFL_STATUS_NEEDS_MORE_INPUT) && (!orig_avail_in))
        return MZ_BUF_ERROR; // Signal caller that we can't make forward
                             // progress without supplying more input or by
                             // setting flush to MZ_FINISH.
      else if (flush == MZ_FINISH)
      {
        // The output buffer MUST be large to hold the remaining uncompressed
        // data when flush==MZ_FINISH.
        if (status == TINFL_STATUS_DONE)
          return pState->m_dict_avail ? MZ_BUF_ERROR : MZ_STREAM_END;
        // status here must be TINFL_STATUS_HAS_MORE_OUTPUT, which means there's
        // at least 1 more byte on the way. If there's no more room left in the
        // output buffer then something is wrong.
        else if (!pStream->avail_out)
          return MZ_BUF_ERROR;
      }
      else if (
        (status == TINFL_STATUS_DONE) || (!pStream->avail_in) ||
        (!pStream->avail_out) || (pState->m_dict_avail))
        break;
    }

    return ((status == TINFL_STATUS_DONE) && (!pState->m_dict_avail)) ?
      MZ_STREAM_END :
      MZ_OK;
  }

  int mz_inflateEnd(mz_streamp pStream)
  {
    if (!pStream)
      return MZ_STREAM_ERROR;
    if (pStream->state)
    {
      pStream->zfree(pStream->opaque, pStream->state);
      pStream->state = NULL;
    }
    return MZ_OK;
  }

  int mz_uncompress(
    unsigned char* pDest,
    mz_ulong* pDest_len,
    const unsigned char* pSource,
    mz_ulong source_len)
  {
    mz_stream stream;
    int status;
    memset(&stream, 0, sizeof(stream));

    // In case mz_ulong is 64-bits (argh I hate longs).
    if ((source_len | *pDest_len) > 0xFFFFFFFFU)
      return MZ_PARAM_ERROR;

    stream.next_in = pSource;
    stream.avail_in = (mz_uint32)source_len;
    stream.next_out = pDest;
    stream.avail_out = (mz_uint32)*pDest_len;

    status = mz_inflateInit(&stream);
    if (status != MZ_OK)
      return status;

    status = mz_inflate(&stream, MZ_FINISH);
    if (status != MZ_STREAM_END)
    {
      mz_inflateEnd(&stream);
      return ((status == MZ_BUF_ERROR) && (!stream.avail_in)) ? MZ_DATA_ERROR :
                                                                status;
    }
    *pDest_len = stream.total_out;

    return mz_inflateEnd(&stream);
  }

  const char* mz_error(int err)
  {
    static struct
    {
      int m_err;
      const char* m_pDesc;
    } s_error_descs[] = {
      {MZ_OK, ""},
      {MZ_STREAM_END, "stream end"},
      {MZ_NEED_DICT, "need dictionary"},
      {MZ_ERRNO, "file error"},
      {MZ_STREAM_ERROR, "stream error"},
      {MZ_DATA_ERROR, "data error"},
      {MZ_MEM_ERROR, "out of memory"},
      {MZ_BUF_ERROR, "buf error"},
      {MZ_VERSION_ERROR, "version error"},
      {MZ_PARAM_ERROR, "parameter error"}};
    mz_uint i;
    for (i = 0; i < sizeof(s_error_descs) / sizeof(s_error_descs[0]); ++i)
      if (s_error_descs[i].m_err == err)
        return s_error_descs[i].m_pDesc;
    return NULL;
  }

#  endif // MINIZ_NO_ZLIB_APIS

  // ------------------- Low-level Decompression (completely independent from
  // all compression API's)

#  define TINFL_MEMCPY(d, s, l) memcpy(d, s, l)
#  define TINFL_MEMSET(p, c, l) memset(p, c, l)

#  define TINFL_CR_BEGIN \
    switch (r->m_state) \
    { \
      case 0:
#  define TINFL_CR_RETURN(state_index, result) \
    do \
    { \
      status = result; \
      r->m_state = state_index; \
      goto common_exit; \
      case state_index:; \
    } \
    MZ_MACRO_END
#  define TINFL_CR_RETURN_FOREVER(state_index, result) \
    do \
    { \
      for (;;) \
      { \
        TINFL_CR_RETURN(state_index, result); \
      } \
    } \
    MZ_MACRO_END
#  define TINFL_CR_FINISH }

// TODO: If the caller has indicated that there's no more input, and we attempt
// to read beyond the input buf, then something is wrong with the input because
// the inflator never reads ahead more than it needs to. Currently
// TINFL_GET_BYTE() pads the end of the stream with 0's in this scenario.
#  define TINFL_GET_BYTE(state_index, c) \
    do \
    { \
      if (pIn_buf_cur >= pIn_buf_end) \
      { \
        for (;;) \
        { \
          if (decomp_flags & TINFL_FLAG_HAS_MORE_INPUT) \
          { \
            TINFL_CR_RETURN(state_index, TINFL_STATUS_NEEDS_MORE_INPUT); \
            if (pIn_buf_cur < pIn_buf_end) \
            { \
              c = *pIn_buf_cur++; \
              break; \
            } \
          } \
          else \
          { \
            c = 0; \
            break; \
          } \
        } \
      } \
      else \
        c = *pIn_buf_cur++; \
    } \
    MZ_MACRO_END

#  define TINFL_NEED_BITS(state_index, n) \
    do \
    { \
      mz_uint c; \
      TINFL_GET_BYTE(state_index, c); \
      bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); \
      num_bits += 8; \
    } while (num_bits < (mz_uint)(n))
#  define TINFL_SKIP_BITS(state_index, n) \
    do \
    { \
      if (num_bits < (mz_uint)(n)) \
      { \
        TINFL_NEED_BITS(state_index, n); \
      } \
      bit_buf >>= (n); \
      num_bits -= (n); \
    } \
    MZ_MACRO_END
#  define TINFL_GET_BITS(state_index, b, n) \
    do \
    { \
      if (num_bits < (mz_uint)(n)) \
      { \
        TINFL_NEED_BITS(state_index, n); \
      } \
      b = bit_buf & ((1 << (n)) - 1); \
      bit_buf >>= (n); \
      num_bits -= (n); \
    } \
    MZ_MACRO_END

// TINFL_HUFF_BITBUF_FILL() is only used rarely, when the number of bytes
// remaining in the input buffer falls below 2. It reads just enough bytes from
// the input stream that are needed to decode the next Huffman code (and
// absolutely no more). It works by trying to fully decode a Huffman code by
// using whatever bits are currently present in the bit buffer. If this fails,
// it reads another byte, and tries again until it succeeds or until the bit
// buffer contains >=15 bits (deflate's max. Huffman code size).
#  define TINFL_HUFF_BITBUF_FILL(state_index, pHuff) \
    do \
    { \
      temp = (pHuff)->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]; \
      if (temp >= 0) \
      { \
        code_len = temp >> 9; \
        if ((code_len) && (num_bits >= code_len)) \
          break; \
      } \
      else if (num_bits > TINFL_FAST_LOOKUP_BITS) \
      { \
        code_len = TINFL_FAST_LOOKUP_BITS; \
        do \
        { \
          temp = (pHuff)->m_tree[~temp + ((bit_buf >> code_len++) & 1)]; \
        } while ((temp < 0) && (num_bits >= (code_len + 1))); \
        if (temp >= 0) \
          break; \
      } \
      TINFL_GET_BYTE(state_index, c); \
      bit_buf |= (((tinfl_bit_buf_t)c) << num_bits); \
      num_bits += 8; \
    } while (num_bits < 15);

// TINFL_HUFF_DECODE() decodes the next Huffman coded symbol. It's more complex
// than you would initially expect because the zlib API expects the decompressor
// to never read beyond the final byte of the deflate stream. (In other words,
// when this macro wants to read another byte from the input, it REALLY needs
// another byte in order to fully decode the next Huffman code.) Handling this
// properly is particularly important on raw deflate (non-zlib) streams, which
// aren't followed by a byte aligned adler-32. The slow path is only executed at
// the very end of the input buffer.
#  define TINFL_HUFF_DECODE(state_index, sym, pHuff) \
    do \
    { \
      int temp; \
      mz_uint code_len, c; \
      if (num_bits < 15) \
      { \
        if ((pIn_buf_end - pIn_buf_cur) < 2) \
        { \
          TINFL_HUFF_BITBUF_FILL(state_index, pHuff); \
        } \
        else \
        { \
          bit_buf |= (((tinfl_bit_buf_t)pIn_buf_cur[0]) << num_bits) | \
            (((tinfl_bit_buf_t)pIn_buf_cur[1]) << (num_bits + 8)); \
          pIn_buf_cur += 2; \
          num_bits += 16; \
        } \
      } \
      if ( \
        (temp = (pHuff)->m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >= \
        0) \
        code_len = temp >> 9, temp &= 511; \
      else \
      { \
        code_len = TINFL_FAST_LOOKUP_BITS; \
        do \
        { \
          temp = (pHuff)->m_tree[~temp + ((bit_buf >> code_len++) & 1)]; \
        } while (temp < 0); \
      } \
      sym = temp; \
      bit_buf >>= code_len; \
      num_bits -= code_len; \
    } \
    MZ_MACRO_END

#  ifdef _MSC_VER
#    pragma warning(push)
#    pragma warning(disable : 4334)
#  endif

  tinfl_status tinfl_decompress(
    tinfl_decompressor* r,
    const mz_uint8* pIn_buf_next,
    size_t* pIn_buf_size,
    mz_uint8* pOut_buf_start,
    mz_uint8* pOut_buf_next,
    size_t* pOut_buf_size,
    const mz_uint32 decomp_flags)
  {
    static const int s_length_base[31] = {
      3,  4,  5,  6,  7,  8,  9,  10,  11,  13,  15,  17,  19,  23, 27, 31,
      35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0,  0};
    static const int s_length_extra[31] = {0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
                                           1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4,
                                           4, 4, 5, 5, 5, 5, 0, 0, 0};
    static const int s_dist_base[32] = {
      1,    2,    3,    4,    5,    7,     9,     13,    17,  25,   33,
      49,   65,   97,   129,  193,  257,   385,   513,   769, 1025, 1537,
      2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577, 0,   0};
    static const int s_dist_extra[32] = {0, 0, 0,  0,  1,  1,  2,  2,  3,  3,
                                         4, 4, 5,  5,  6,  6,  7,  7,  8,  8,
                                         9, 9, 10, 10, 11, 11, 12, 12, 13, 13};
    static const mz_uint8 s_length_dezigzag[19] = {
      16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};
    static const int s_min_table_sizes[3] = {257, 1, 4};

    tinfl_status status = TINFL_STATUS_FAILED;
    mz_uint32 num_bits, dist, counter, num_extra;
    tinfl_bit_buf_t bit_buf;
    const mz_uint8 *pIn_buf_cur = pIn_buf_next,
                   *const pIn_buf_end = pIn_buf_next + *pIn_buf_size;
    mz_uint8 *pOut_buf_cur = pOut_buf_next,
             *const pOut_buf_end = pOut_buf_next + *pOut_buf_size;
    size_t out_buf_size_mask =
             (decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF) ?
      (size_t)-1 :
      ((pOut_buf_next - pOut_buf_start) + *pOut_buf_size) - 1,
           dist_from_out_buf_start;

    // Ensure the output buffer's size is a power of 2, unless the output buffer
    // is large enough to hold the entire output file (in which case it doesn't
    // matter).
    if (
      ((out_buf_size_mask + 1) & out_buf_size_mask) ||
      (pOut_buf_next < pOut_buf_start))
    {
      *pIn_buf_size = *pOut_buf_size = 0;
      return TINFL_STATUS_BAD_PARAM;
    }

    num_bits = r->m_num_bits;
    bit_buf = r->m_bit_buf;
    dist = r->m_dist;
    counter = r->m_counter;
    num_extra = r->m_num_extra;
    dist_from_out_buf_start = r->m_dist_from_out_buf_start;
    TINFL_CR_BEGIN

    bit_buf = num_bits = dist = counter = num_extra = r->m_zhdr0 = r->m_zhdr1 =
      0;
    r->m_z_adler32 = r->m_check_adler32 = 1;
    if (decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER)
    {
      TINFL_GET_BYTE(1, r->m_zhdr0);
      TINFL_GET_BYTE(2, r->m_zhdr1);
      counter =
        (((r->m_zhdr0 * 256 + r->m_zhdr1) % 31 != 0) || (r->m_zhdr1 & 32) ||
         ((r->m_zhdr0 & 15) != 8));
      if (!(decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF))
        counter |=
          (((1U << (8U + (r->m_zhdr0 >> 4))) > 32768U) ||
           ((out_buf_size_mask + 1) <
            (size_t)(1U << (8U + (r->m_zhdr0 >> 4)))));
      if (counter)
      {
        TINFL_CR_RETURN_FOREVER(36, TINFL_STATUS_FAILED);
      }
    }

    do
    {
      TINFL_GET_BITS(3, r->m_final, 3);
      r->m_type = r->m_final >> 1;
      if (r->m_type == 0)
      {
        TINFL_SKIP_BITS(5, num_bits & 7);
        for (counter = 0; counter < 4; ++counter)
        {
          if (num_bits)
            TINFL_GET_BITS(6, r->m_raw_header[counter], 8);
          else
            TINFL_GET_BYTE(7, r->m_raw_header[counter]);
        }
        if (
          (counter = (r->m_raw_header[0] | (r->m_raw_header[1] << 8))) !=
          (mz_uint)(0xFFFF ^ (r->m_raw_header[2] | (r->m_raw_header[3] << 8))))
        {
          TINFL_CR_RETURN_FOREVER(39, TINFL_STATUS_FAILED);
        }
        while ((counter) && (num_bits))
        {
          TINFL_GET_BITS(51, dist, 8);
          while (pOut_buf_cur >= pOut_buf_end)
          {
            TINFL_CR_RETURN(52, TINFL_STATUS_HAS_MORE_OUTPUT);
          }
          *pOut_buf_cur++ = (mz_uint8)dist;
          counter--;
        }
        while (counter)
        {
          size_t n;
          while (pOut_buf_cur >= pOut_buf_end)
          {
            TINFL_CR_RETURN(9, TINFL_STATUS_HAS_MORE_OUTPUT);
          }
          while (pIn_buf_cur >= pIn_buf_end)
          {
            if (decomp_flags & TINFL_FLAG_HAS_MORE_INPUT)
            {
              TINFL_CR_RETURN(38, TINFL_STATUS_NEEDS_MORE_INPUT);
            }
            else
            {
              TINFL_CR_RETURN_FOREVER(40, TINFL_STATUS_FAILED);
            }
          }
          n = MZ_MIN(
            MZ_MIN(
              (size_t)(pOut_buf_end - pOut_buf_cur),
              (size_t)(pIn_buf_end - pIn_buf_cur)),
            counter);
          TINFL_MEMCPY(pOut_buf_cur, pIn_buf_cur, n);
          pIn_buf_cur += n;
          pOut_buf_cur += n;
          counter -= (mz_uint)n;
        }
      }
      else if (r->m_type == 3)
      {
        TINFL_CR_RETURN_FOREVER(10, TINFL_STATUS_FAILED);
      }
      else
      {
        if (r->m_type == 1)
        {
          mz_uint8* p = r->m_tables[0].m_code_size;
          mz_uint i;
          r->m_table_sizes[0] = 288;
          r->m_table_sizes[1] = 32;
          TINFL_MEMSET(r->m_tables[1].m_code_size, 5, 32);
          for (i = 0; i <= 143; ++i)
            *p++ = 8;
          for (; i <= 255; ++i)
            *p++ = 9;
          for (; i <= 279; ++i)
            *p++ = 7;
          for (; i <= 287; ++i)
            *p++ = 8;
        }
        else
        {
          for (counter = 0; counter < 3; counter++)
          {
            TINFL_GET_BITS(11, r->m_table_sizes[counter], "\05\05\04"[counter]);
            r->m_table_sizes[counter] += s_min_table_sizes[counter];
          }
          MZ_CLEAR_OBJ(r->m_tables[2].m_code_size);
          for (counter = 0; counter < r->m_table_sizes[2]; counter++)
          {
            mz_uint s;
            TINFL_GET_BITS(14, s, 3);
            r->m_tables[2].m_code_size[s_length_dezigzag[counter]] =
              (mz_uint8)s;
          }
          r->m_table_sizes[2] = 19;
        }
        for (; (int)r->m_type >= 0; r->m_type--)
        {
          int tree_next, tree_cur;
          tinfl_huff_table* pTable;
          mz_uint i, j, used_syms, total, sym_index, next_code[17],
            total_syms[16];
          pTable = &r->m_tables[r->m_type];
          MZ_CLEAR_OBJ(total_syms);
          MZ_CLEAR_OBJ(pTable->m_look_up);
          MZ_CLEAR_OBJ(pTable->m_tree);
          for (i = 0; i < r->m_table_sizes[r->m_type]; ++i)
            total_syms[pTable->m_code_size[i]]++;
          used_syms = 0, total = 0;
          next_code[0] = next_code[1] = 0;
          for (i = 1; i <= 15; ++i)
          {
            used_syms += total_syms[i];
            next_code[i + 1] = (total = ((total + total_syms[i]) << 1));
          }
          if ((65536 != total) && (used_syms > 1))
          {
            TINFL_CR_RETURN_FOREVER(35, TINFL_STATUS_FAILED);
          }
          for (tree_next = -1, sym_index = 0;
               sym_index < r->m_table_sizes[r->m_type];
               ++sym_index)
          {
            mz_uint rev_code = 0, l, cur_code,
                    code_size = pTable->m_code_size[sym_index];
            if (!code_size)
              continue;
            cur_code = next_code[code_size]++;
            for (l = code_size; l > 0; l--, cur_code >>= 1)
              rev_code = (rev_code << 1) | (cur_code & 1);
            if (code_size <= TINFL_FAST_LOOKUP_BITS)
            {
              mz_int16 k = (mz_int16)((code_size << 9) | sym_index);
              while (rev_code < TINFL_FAST_LOOKUP_SIZE)
              {
                pTable->m_look_up[rev_code] = k;
                rev_code += (1 << code_size);
              }
              continue;
            }
            if (
              0 ==
              (tree_cur =
                 pTable->m_look_up[rev_code & (TINFL_FAST_LOOKUP_SIZE - 1)]))
            {
              pTable->m_look_up[rev_code & (TINFL_FAST_LOOKUP_SIZE - 1)] =
                (mz_int16)tree_next;
              tree_cur = tree_next;
              tree_next -= 2;
            }
            rev_code >>= (TINFL_FAST_LOOKUP_BITS - 1);
            for (j = code_size; j > (TINFL_FAST_LOOKUP_BITS + 1); j--)
            {
              tree_cur -= ((rev_code >>= 1) & 1);
              if (!pTable->m_tree[-tree_cur - 1])
              {
                pTable->m_tree[-tree_cur - 1] = (mz_int16)tree_next;
                tree_cur = tree_next;
                tree_next -= 2;
              }
              else
                tree_cur = pTable->m_tree[-tree_cur - 1];
            }
            tree_cur -= ((rev_code >>= 1) & 1);
            pTable->m_tree[-tree_cur - 1] = (mz_int16)sym_index;
          }
          if (r->m_type == 2)
          {
            for (counter = 0;
                 counter < (r->m_table_sizes[0] + r->m_table_sizes[1]);)
            {
              mz_uint s;
              TINFL_HUFF_DECODE(16, dist, &r->m_tables[2]);
              if (dist < 16)
              {
                r->m_len_codes[counter++] = (mz_uint8)dist;
                continue;
              }
              if ((dist == 16) && (!counter))
              {
                TINFL_CR_RETURN_FOREVER(17, TINFL_STATUS_FAILED);
              }
              num_extra = "\02\03\07"[dist - 16];
              TINFL_GET_BITS(18, s, num_extra);
              s += "\03\03\013"[dist - 16];
              TINFL_MEMSET(
                r->m_len_codes + counter,
                (dist == 16) ? r->m_len_codes[counter - 1] : 0,
                s);
              counter += s;
            }
            if ((r->m_table_sizes[0] + r->m_table_sizes[1]) != counter)
            {
              TINFL_CR_RETURN_FOREVER(21, TINFL_STATUS_FAILED);
            }
            TINFL_MEMCPY(
              r->m_tables[0].m_code_size, r->m_len_codes, r->m_table_sizes[0]);
            TINFL_MEMCPY(
              r->m_tables[1].m_code_size,
              r->m_len_codes + r->m_table_sizes[0],
              r->m_table_sizes[1]);
          }
        }
        for (;;)
        {
          mz_uint8* pSrc;
          for (;;)
          {
            if (
              ((pIn_buf_end - pIn_buf_cur) < 4) ||
              ((pOut_buf_end - pOut_buf_cur) < 2))
            {
              TINFL_HUFF_DECODE(23, counter, &r->m_tables[0]);
              if (counter >= 256)
                break;
              while (pOut_buf_cur >= pOut_buf_end)
              {
                TINFL_CR_RETURN(24, TINFL_STATUS_HAS_MORE_OUTPUT);
              }
              *pOut_buf_cur++ = (mz_uint8)counter;
            }
            else
            {
              int sym2;
              mz_uint code_len;
#  if TINFL_USE_64BIT_BITBUF
              if (num_bits < 30)
              {
                bit_buf |=
                  (((tinfl_bit_buf_t)MZ_READ_LE32(pIn_buf_cur)) << num_bits);
                pIn_buf_cur += 4;
                num_bits += 32;
              }
#  else
            if (num_bits < 15)
            {
              bit_buf |=
                (((tinfl_bit_buf_t)MZ_READ_LE16(pIn_buf_cur)) << num_bits);
              pIn_buf_cur += 2;
              num_bits += 16;
            }
#  endif
              if (
                (sym2 = r->m_tables[0]
                          .m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >=
                0)
                code_len = sym2 >> 9;
              else
              {
                code_len = TINFL_FAST_LOOKUP_BITS;
                do
                {
                  sym2 = r->m_tables[0]
                           .m_tree[~sym2 + ((bit_buf >> code_len++) & 1)];
                } while (sym2 < 0);
              }
              counter = sym2;
              bit_buf >>= code_len;
              num_bits -= code_len;
              if (counter & 256)
                break;

#  if !TINFL_USE_64BIT_BITBUF
              if (num_bits < 15)
              {
                bit_buf |=
                  (((tinfl_bit_buf_t)MZ_READ_LE16(pIn_buf_cur)) << num_bits);
                pIn_buf_cur += 2;
                num_bits += 16;
              }
#  endif
              if (
                (sym2 = r->m_tables[0]
                          .m_look_up[bit_buf & (TINFL_FAST_LOOKUP_SIZE - 1)]) >=
                0)
                code_len = sym2 >> 9;
              else
              {
                code_len = TINFL_FAST_LOOKUP_BITS;
                do
                {
                  sym2 = r->m_tables[0]
                           .m_tree[~sym2 + ((bit_buf >> code_len++) & 1)];
                } while (sym2 < 0);
              }
              bit_buf >>= code_len;
              num_bits -= code_len;

              pOut_buf_cur[0] = (mz_uint8)counter;
              if (sym2 & 256)
              {
                pOut_buf_cur++;
                counter = sym2;
                break;
              }
              pOut_buf_cur[1] = (mz_uint8)sym2;
              pOut_buf_cur += 2;
            }
          }
          if ((counter &= 511) == 256)
            break;

          num_extra = s_length_extra[counter - 257];
          counter = s_length_base[counter - 257];
          if (num_extra)
          {
            mz_uint extra_bits;
            TINFL_GET_BITS(25, extra_bits, num_extra);
            counter += extra_bits;
          }

          TINFL_HUFF_DECODE(26, dist, &r->m_tables[1]);
          num_extra = s_dist_extra[dist];
          dist = s_dist_base[dist];
          if (num_extra)
          {
            mz_uint extra_bits;
            TINFL_GET_BITS(27, extra_bits, num_extra);
            dist += extra_bits;
          }

          dist_from_out_buf_start = pOut_buf_cur - pOut_buf_start;
          if (
            (dist > dist_from_out_buf_start) &&
            (decomp_flags & TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF))
          {
            TINFL_CR_RETURN_FOREVER(37, TINFL_STATUS_FAILED);
          }

          pSrc = pOut_buf_start +
            ((dist_from_out_buf_start - dist) & out_buf_size_mask);

          if ((MZ_MAX(pOut_buf_cur, pSrc) + counter) > pOut_buf_end)
          {
            while (counter--)
            {
              while (pOut_buf_cur >= pOut_buf_end)
              {
                TINFL_CR_RETURN(53, TINFL_STATUS_HAS_MORE_OUTPUT);
              }
              *pOut_buf_cur++ = pOut_buf_start
                [(dist_from_out_buf_start++ - dist) & out_buf_size_mask];
            }
            continue;
          }
#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES
          else if ((counter >= 9) && (counter <= dist))
          {
            const mz_uint8* pSrc_end = pSrc + (counter & ~7);
            do
            {
              ((mz_uint32*)pOut_buf_cur)[0] = ((const mz_uint32*)pSrc)[0];
              ((mz_uint32*)pOut_buf_cur)[1] = ((const mz_uint32*)pSrc)[1];
              pOut_buf_cur += 8;
            } while ((pSrc += 8) < pSrc_end);
            if ((counter &= 7) < 3)
            {
              if (counter)
              {
                pOut_buf_cur[0] = pSrc[0];
                if (counter > 1)
                  pOut_buf_cur[1] = pSrc[1];
                pOut_buf_cur += counter;
              }
              continue;
            }
          }
#  endif
          do
          {
            pOut_buf_cur[0] = pSrc[0];
            pOut_buf_cur[1] = pSrc[1];
            pOut_buf_cur[2] = pSrc[2];
            pOut_buf_cur += 3;
            pSrc += 3;
          } while ((int)(counter -= 3) > 2);
          if ((int)counter > 0)
          {
            pOut_buf_cur[0] = pSrc[0];
            if ((int)counter > 1)
              pOut_buf_cur[1] = pSrc[1];
            pOut_buf_cur += counter;
          }
        }
      }
    } while (!(r->m_final & 1));
    if (decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER)
    {
      TINFL_SKIP_BITS(32, num_bits & 7);
      for (counter = 0; counter < 4; ++counter)
      {
        mz_uint s;
        if (num_bits)
          TINFL_GET_BITS(41, s, 8);
        else
          TINFL_GET_BYTE(42, s);
        r->m_z_adler32 = (r->m_z_adler32 << 8) | s;
      }
    }
    TINFL_CR_RETURN_FOREVER(34, TINFL_STATUS_DONE);
    TINFL_CR_FINISH

  common_exit:
    r->m_num_bits = num_bits;
    r->m_bit_buf = bit_buf;
    r->m_dist = dist;
    r->m_counter = counter;
    r->m_num_extra = num_extra;
    r->m_dist_from_out_buf_start = dist_from_out_buf_start;
    *pIn_buf_size = pIn_buf_cur - pIn_buf_next;
    *pOut_buf_size = pOut_buf_cur - pOut_buf_next;
    if (
      (decomp_flags &
       (TINFL_FLAG_PARSE_ZLIB_HEADER | TINFL_FLAG_COMPUTE_ADLER32)) &&
      (status >= 0))
    {
      const mz_uint8* ptr = pOut_buf_next;
      size_t buf_len = *pOut_buf_size;
      mz_uint32 i, s1 = r->m_check_adler32 & 0xffff,
                   s2 = r->m_check_adler32 >> 16;
      size_t block_len = buf_len % 5552;
      while (buf_len)
      {
        for (i = 0; i + 7 < block_len; i += 8, ptr += 8)
        {
          s1 += ptr[0], s2 += s1;
          s1 += ptr[1], s2 += s1;
          s1 += ptr[2], s2 += s1;
          s1 += ptr[3], s2 += s1;
          s1 += ptr[4], s2 += s1;
          s1 += ptr[5], s2 += s1;
          s1 += ptr[6], s2 += s1;
          s1 += ptr[7], s2 += s1;
        }
        for (; i < block_len; ++i)
          s1 += *ptr++, s2 += s1;
        s1 %= 65521U, s2 %= 65521U;
        buf_len -= block_len;
        block_len = 5552;
      }
      r->m_check_adler32 = (s2 << 16) + s1;
      if (
        (status == TINFL_STATUS_DONE) &&
        (decomp_flags & TINFL_FLAG_PARSE_ZLIB_HEADER) &&
        (r->m_check_adler32 != r->m_z_adler32))
        status = TINFL_STATUS_ADLER32_MISMATCH;
    }
    return status;
  }

#  ifdef _MSC_VER
#    pragma warning(pop)
#  endif

  // Higher level helper functions.
  void* tinfl_decompress_mem_to_heap(
    const void* pSrc_buf, size_t src_buf_len, size_t* pOut_len, int flags)
  {
    tinfl_decompressor decomp;
    void *pBuf = NULL, *pNew_buf;
    size_t src_buf_ofs = 0, out_buf_capacity = 0;
    *pOut_len = 0;
    tinfl_init(&decomp);
    for (;;)
    {
      size_t src_buf_size = src_buf_len - src_buf_ofs,
             dst_buf_size = out_buf_capacity - *pOut_len, new_out_buf_capacity;
      tinfl_status status = tinfl_decompress(
        &decomp,
        (const mz_uint8*)pSrc_buf + src_buf_ofs,
        &src_buf_size,
        (mz_uint8*)pBuf,
        pBuf ? (mz_uint8*)pBuf + *pOut_len : NULL,
        &dst_buf_size,
        (flags & ~TINFL_FLAG_HAS_MORE_INPUT) |
          TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF);
      if ((status < 0) || (status == TINFL_STATUS_NEEDS_MORE_INPUT))
      {
        MZ_FREE(pBuf);
        *pOut_len = 0;
        return NULL;
      }
      src_buf_ofs += src_buf_size;
      *pOut_len += dst_buf_size;
      if (status == TINFL_STATUS_DONE)
        break;
      new_out_buf_capacity = out_buf_capacity * 2;
      if (new_out_buf_capacity < 128)
        new_out_buf_capacity = 128;
      pNew_buf = MZ_REALLOC(pBuf, new_out_buf_capacity);
      if (!pNew_buf)
      {
        MZ_FREE(pBuf);
        *pOut_len = 0;
        return NULL;
      }
      pBuf = pNew_buf;
      out_buf_capacity = new_out_buf_capacity;
    }
    return pBuf;
  }

  size_t tinfl_decompress_mem_to_mem(
    void* pOut_buf,
    size_t out_buf_len,
    const void* pSrc_buf,
    size_t src_buf_len,
    int flags)
  {
    tinfl_decompressor decomp;
    tinfl_status status;
    tinfl_init(&decomp);
    status = tinfl_decompress(
      &decomp,
      (const mz_uint8*)pSrc_buf,
      &src_buf_len,
      (mz_uint8*)pOut_buf,
      (mz_uint8*)pOut_buf,
      &out_buf_len,
      (flags & ~TINFL_FLAG_HAS_MORE_INPUT) |
        TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF);
    return (status != TINFL_STATUS_DONE) ? TINFL_DECOMPRESS_MEM_TO_MEM_FAILED :
                                           out_buf_len;
  }

  int tinfl_decompress_mem_to_callback(
    const void* pIn_buf,
    size_t* pIn_buf_size,
    tinfl_put_buf_func_ptr pPut_buf_func,
    void* pPut_buf_user,
    int flags)
  {
    int result = 0;
    tinfl_decompressor decomp;
    mz_uint8* pDict = (mz_uint8*)MZ_MALLOC(TINFL_LZ_DICT_SIZE);
    size_t in_buf_ofs = 0, dict_ofs = 0;
    if (!pDict)
      return TINFL_STATUS_FAILED;
    tinfl_init(&decomp);
    for (;;)
    {
      size_t in_buf_size = *pIn_buf_size - in_buf_ofs,
             dst_buf_size = TINFL_LZ_DICT_SIZE - dict_ofs;
      tinfl_status status = tinfl_decompress(
        &decomp,
        (const mz_uint8*)pIn_buf + in_buf_ofs,
        &in_buf_size,
        pDict,
        pDict + dict_ofs,
        &dst_buf_size,
        (flags &
         ~(TINFL_FLAG_HAS_MORE_INPUT |
           TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF)));
      in_buf_ofs += in_buf_size;
      if (
        (dst_buf_size) &&
        (!(*pPut_buf_func)(pDict + dict_ofs, (int)dst_buf_size, pPut_buf_user)))
        break;
      if (status != TINFL_STATUS_HAS_MORE_OUTPUT)
      {
        result = (status == TINFL_STATUS_DONE);
        break;
      }
      dict_ofs = (dict_ofs + dst_buf_size) & (TINFL_LZ_DICT_SIZE - 1);
    }
    MZ_FREE(pDict);
    *pIn_buf_size = in_buf_ofs;
    return result;
  }

  // ------------------- Low-level Compression (independent from all
  // decompression API's)

  // Purposely making these tables static for faster init and thread safety.
  static const mz_uint16 s_tdefl_len_sym[256] = {
    257, 258, 259, 260, 261, 262, 263, 264, 265, 265, 266, 266, 267, 267, 268,
    268, 269, 269, 269, 269, 270, 270, 270, 270, 271, 271, 271, 271, 272, 272,
    272, 272, 273, 273, 273, 273, 273, 273, 273, 273, 274, 274, 274, 274, 274,
    274, 274, 274, 275, 275, 275, 275, 275, 275, 275, 275, 276, 276, 276, 276,
    276, 276, 276, 276, 277, 277, 277, 277, 277, 277, 277, 277, 277, 277, 277,
    277, 277, 277, 277, 277, 278, 278, 278, 278, 278, 278, 278, 278, 278, 278,
    278, 278, 278, 278, 278, 278, 279, 279, 279, 279, 279, 279, 279, 279, 279,
    279, 279, 279, 279, 279, 279, 279, 280, 280, 280, 280, 280, 280, 280, 280,
    280, 280, 280, 280, 280, 280, 280, 280, 281, 281, 281, 281, 281, 281, 281,
    281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 281,
    281, 281, 281, 281, 281, 281, 281, 281, 281, 281, 282, 282, 282, 282, 282,
    282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282,
    282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 282, 283, 283, 283,
    283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283,
    283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 283, 284,
    284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284,
    284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284, 284,
    285};

  static const mz_uint8 s_tdefl_len_extra[256] = {
    0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
    2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 0};

  static const mz_uint8 s_tdefl_small_dist_sym[512] = {
    0,  1,  2,  3,  4,  4,  5,  5,  6,  6,  6,  6,  7,  7,  7,  7,  8,  8,  8,
    8,  8,  8,  8,  8,  9,  9,  9,  9,  9,  9,  9,  9,  10, 10, 10, 10, 10, 10,
    10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11,
    11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
    12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
    12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
    13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14,
    14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
    14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
    14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
    14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
    15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
    15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
    15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17};

  static const mz_uint8 s_tdefl_small_dist_extra[512] = {
    0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7};

  static const mz_uint8 s_tdefl_large_dist_sym[128] = {
    0,  0,  18, 19, 20, 20, 21, 21, 22, 22, 22, 22, 23, 23, 23, 23, 24, 24, 24,
    24, 24, 24, 24, 24, 25, 25, 25, 25, 25, 25, 25, 25, 26, 26, 26, 26, 26, 26,
    26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 27, 27, 27, 27, 27, 27, 27, 27, 27,
    27, 27, 27, 27, 27, 27, 27, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
    28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
    28, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29,
    29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29};

  static const mz_uint8 s_tdefl_large_dist_extra[128] = {
    0,  0,  8,  8,  9,  9,  9,  9,  10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11,
    11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12,
    12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
    12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
    13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
    13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
    13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13};

  // Radix sorts tdefl_sym_freq[] array by 16-bit key m_key. Returns ptr to
  // sorted values.
  typedef struct
  {
    mz_uint16 m_key, m_sym_index;
  } tdefl_sym_freq;
  static tdefl_sym_freq* tdefl_radix_sort_syms(
    mz_uint num_syms, tdefl_sym_freq* pSyms0, tdefl_sym_freq* pSyms1)
  {
    mz_uint32 total_passes = 2, pass_shift, pass, i, hist[256 * 2];
    tdefl_sym_freq *pCur_syms = pSyms0, *pNew_syms = pSyms1;
    MZ_CLEAR_OBJ(hist);
    for (i = 0; i < num_syms; i++)
    {
      mz_uint freq = pSyms0[i].m_key;
      hist[freq & 0xFF]++;
      hist[256 + ((freq >> 8) & 0xFF)]++;
    }
    while ((total_passes > 1) && (num_syms == hist[(total_passes - 1) * 256]))
      total_passes--;
    for (pass_shift = 0, pass = 0; pass < total_passes; pass++, pass_shift += 8)
    {
      const mz_uint32* pHist = &hist[pass << 8];
      mz_uint offsets[256], cur_ofs = 0;
      for (i = 0; i < 256; i++)
      {
        offsets[i] = cur_ofs;
        cur_ofs += pHist[i];
      }
      for (i = 0; i < num_syms; i++)
        pNew_syms[offsets[(pCur_syms[i].m_key >> pass_shift) & 0xFF]++] =
          pCur_syms[i];
      {
        tdefl_sym_freq* t = pCur_syms;
        pCur_syms = pNew_syms;
        pNew_syms = t;
      }
    }
    return pCur_syms;
  }

  // tdefl_calculate_minimum_redundancy() originally written by: Alistair
  // Moffat, alistair@cs.mu.oz.au, Jyrki Katajainen, jyrki@diku.dk, November
  // 1996.
  static void tdefl_calculate_minimum_redundancy(tdefl_sym_freq* A, int n)
  {
    int root, leaf, next, avbl, used, dpth;
    if (n == 0)
      return;
    else if (n == 1)
    {
      A[0].m_key = 1;
      return;
    }
    A[0].m_key += A[1].m_key;
    root = 0;
    leaf = 2;
    for (next = 1; next < n - 1; next++)
    {
      if (leaf >= n || A[root].m_key < A[leaf].m_key)
      {
        A[next].m_key = A[root].m_key;
        A[root++].m_key = (mz_uint16)next;
      }
      else
        A[next].m_key = A[leaf++].m_key;
      if (leaf >= n || (root < next && A[root].m_key < A[leaf].m_key))
      {
        A[next].m_key = (mz_uint16)(A[next].m_key + A[root].m_key);
        A[root++].m_key = (mz_uint16)next;
      }
      else
        A[next].m_key = (mz_uint16)(A[next].m_key + A[leaf++].m_key);
    }
    A[n - 2].m_key = 0;
    for (next = n - 3; next >= 0; next--)
      A[next].m_key = A[A[next].m_key].m_key + 1;
    avbl = 1;
    used = dpth = 0;
    root = n - 2;
    next = n - 1;
    while (avbl > 0)
    {
      while (root >= 0 && (int)A[root].m_key == dpth)
      {
        used++;
        root--;
      }
      while (avbl > used)
      {
        A[next--].m_key = (mz_uint16)(dpth);
        avbl--;
      }
      avbl = 2 * used;
      dpth++;
      used = 0;
    }
  }

  // Limits canonical Huffman code table's max code size.
  enum
  {
    TDEFL_MAX_SUPPORTED_HUFF_CODESIZE = 32
  };
  static void tdefl_huffman_enforce_max_code_size(
    int* pNum_codes, int code_list_len, int max_code_size)
  {
    int i;
    mz_uint32 total = 0;
    if (code_list_len <= 1)
      return;
    for (i = max_code_size + 1; i <= TDEFL_MAX_SUPPORTED_HUFF_CODESIZE; i++)
      pNum_codes[max_code_size] += pNum_codes[i];
    for (i = max_code_size; i > 0; i--)
      total += (((mz_uint32)pNum_codes[i]) << (max_code_size - i));
    while (total != (1UL << max_code_size))
    {
      pNum_codes[max_code_size]--;
      for (i = max_code_size - 1; i > 0; i--)
        if (pNum_codes[i])
        {
          pNum_codes[i]--;
          pNum_codes[i + 1] += 2;
          break;
        }
      total--;
    }
  }

  static void tdefl_optimize_huffman_table(
    tdefl_compressor* d,
    int table_num,
    int table_len,
    int code_size_limit,
    int static_table)
  {
    int i, j, l, num_codes[1 + TDEFL_MAX_SUPPORTED_HUFF_CODESIZE];
    mz_uint next_code[TDEFL_MAX_SUPPORTED_HUFF_CODESIZE + 1];
    MZ_CLEAR_OBJ(num_codes);
    if (static_table)
    {
      for (i = 0; i < table_len; i++)
        num_codes[d->m_huff_code_sizes[table_num][i]]++;
    }
    else
    {
      tdefl_sym_freq syms0[TDEFL_MAX_HUFF_SYMBOLS],
        syms1[TDEFL_MAX_HUFF_SYMBOLS], *pSyms;
      int num_used_syms = 0;
      const mz_uint16* pSym_count = &d->m_huff_count[table_num][0];
      for (i = 0; i < table_len; i++)
        if (pSym_count[i])
        {
          syms0[num_used_syms].m_key = (mz_uint16)pSym_count[i];
          syms0[num_used_syms++].m_sym_index = (mz_uint16)i;
        }

      pSyms = tdefl_radix_sort_syms(num_used_syms, syms0, syms1);
      tdefl_calculate_minimum_redundancy(pSyms, num_used_syms);

      for (i = 0; i < num_used_syms; i++)
        num_codes[pSyms[i].m_key]++;

      tdefl_huffman_enforce_max_code_size(
        num_codes, num_used_syms, code_size_limit);

      MZ_CLEAR_OBJ(d->m_huff_code_sizes[table_num]);
      MZ_CLEAR_OBJ(d->m_huff_codes[table_num]);
      for (i = 1, j = num_used_syms; i <= code_size_limit; i++)
        for (l = num_codes[i]; l > 0; l--)
          d->m_huff_code_sizes[table_num][pSyms[--j].m_sym_index] =
            (mz_uint8)(i);
    }

    next_code[1] = 0;
    for (j = 0, i = 2; i <= code_size_limit; i++)
      next_code[i] = j = ((j + num_codes[i - 1]) << 1);

    for (i = 0; i < table_len; i++)
    {
      mz_uint rev_code = 0, code, code_size;
      if ((code_size = d->m_huff_code_sizes[table_num][i]) == 0)
        continue;
      code = next_code[code_size]++;
      for (l = code_size; l > 0; l--, code >>= 1)
        rev_code = (rev_code << 1) | (code & 1);
      d->m_huff_codes[table_num][i] = (mz_uint16)rev_code;
    }
  }

#  define TDEFL_PUT_BITS(b, l) \
    do \
    { \
      mz_uint bits = b; \
      mz_uint len = l; \
      MZ_ASSERT(bits <= ((1U << len) - 1U)); \
      d->m_bit_buffer |= (bits << d->m_bits_in); \
      d->m_bits_in += len; \
      while (d->m_bits_in >= 8) \
      { \
        if (d->m_pOutput_buf < d->m_pOutput_buf_end) \
          *d->m_pOutput_buf++ = (mz_uint8)(d->m_bit_buffer); \
        d->m_bit_buffer >>= 8; \
        d->m_bits_in -= 8; \
      } \
    } \
    MZ_MACRO_END

#  define TDEFL_RLE_PREV_CODE_SIZE() \
    { \
      if (rle_repeat_count) \
      { \
        if (rle_repeat_count < 3) \
        { \
          d->m_huff_count[2][prev_code_size] = (mz_uint16)( \
            d->m_huff_count[2][prev_code_size] + rle_repeat_count); \
          while (rle_repeat_count--) \
            packed_code_sizes[num_packed_code_sizes++] = prev_code_size; \
        } \
        else \
        { \
          d->m_huff_count[2][16] = (mz_uint16)(d->m_huff_count[2][16] + 1); \
          packed_code_sizes[num_packed_code_sizes++] = 16; \
          packed_code_sizes[num_packed_code_sizes++] = \
            (mz_uint8)(rle_repeat_count - 3); \
        } \
        rle_repeat_count = 0; \
      } \
    }

#  define TDEFL_RLE_ZERO_CODE_SIZE() \
    { \
      if (rle_z_count) \
      { \
        if (rle_z_count < 3) \
        { \
          d->m_huff_count[2][0] = \
            (mz_uint16)(d->m_huff_count[2][0] + rle_z_count); \
          while (rle_z_count--) \
            packed_code_sizes[num_packed_code_sizes++] = 0; \
        } \
        else if (rle_z_count <= 10) \
        { \
          d->m_huff_count[2][17] = (mz_uint16)(d->m_huff_count[2][17] + 1); \
          packed_code_sizes[num_packed_code_sizes++] = 17; \
          packed_code_sizes[num_packed_code_sizes++] = \
            (mz_uint8)(rle_z_count - 3); \
        } \
        else \
        { \
          d->m_huff_count[2][18] = (mz_uint16)(d->m_huff_count[2][18] + 1); \
          packed_code_sizes[num_packed_code_sizes++] = 18; \
          packed_code_sizes[num_packed_code_sizes++] = \
            (mz_uint8)(rle_z_count - 11); \
        } \
        rle_z_count = 0; \
      } \
    }

  static mz_uint8 s_tdefl_packed_code_size_syms_swizzle[] = {
    16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

  static void tdefl_start_dynamic_block(tdefl_compressor* d)
  {
    int num_lit_codes, num_dist_codes, num_bit_lengths;
    mz_uint i, total_code_sizes_to_pack, num_packed_code_sizes, rle_z_count,
      rle_repeat_count, packed_code_sizes_index;
    mz_uint8
      code_sizes_to_pack[TDEFL_MAX_HUFF_SYMBOLS_0 + TDEFL_MAX_HUFF_SYMBOLS_1],
      packed_code_sizes[TDEFL_MAX_HUFF_SYMBOLS_0 + TDEFL_MAX_HUFF_SYMBOLS_1],
      prev_code_size = 0xFF;

    d->m_huff_count[0][256] = 1;

    tdefl_optimize_huffman_table(d, 0, TDEFL_MAX_HUFF_SYMBOLS_0, 15, MZ_FALSE);
    tdefl_optimize_huffman_table(d, 1, TDEFL_MAX_HUFF_SYMBOLS_1, 15, MZ_FALSE);

    for (num_lit_codes = 286; num_lit_codes > 257; num_lit_codes--)
      if (d->m_huff_code_sizes[0][num_lit_codes - 1])
        break;
    for (num_dist_codes = 30; num_dist_codes > 1; num_dist_codes--)
      if (d->m_huff_code_sizes[1][num_dist_codes - 1])
        break;

    memcpy(code_sizes_to_pack, &d->m_huff_code_sizes[0][0], num_lit_codes);
    memcpy(
      code_sizes_to_pack + num_lit_codes,
      &d->m_huff_code_sizes[1][0],
      num_dist_codes);
    total_code_sizes_to_pack = num_lit_codes + num_dist_codes;
    num_packed_code_sizes = 0;
    rle_z_count = 0;
    rle_repeat_count = 0;

    memset(
      &d->m_huff_count[2][0],
      0,
      sizeof(d->m_huff_count[2][0]) * TDEFL_MAX_HUFF_SYMBOLS_2);
    for (i = 0; i < total_code_sizes_to_pack; i++)
    {
      mz_uint8 code_size = code_sizes_to_pack[i];
      if (!code_size)
      {
        TDEFL_RLE_PREV_CODE_SIZE();
        if (++rle_z_count == 138)
        {
          TDEFL_RLE_ZERO_CODE_SIZE();
        }
      }
      else
      {
        TDEFL_RLE_ZERO_CODE_SIZE();
        if (code_size != prev_code_size)
        {
          TDEFL_RLE_PREV_CODE_SIZE();
          d->m_huff_count[2][code_size] =
            (mz_uint16)(d->m_huff_count[2][code_size] + 1);
          packed_code_sizes[num_packed_code_sizes++] = code_size;
        }
        else if (++rle_repeat_count == 6)
        {
          TDEFL_RLE_PREV_CODE_SIZE();
        }
      }
      prev_code_size = code_size;
    }
    if (rle_repeat_count)
    {
      TDEFL_RLE_PREV_CODE_SIZE();
    }
    else
    {
      TDEFL_RLE_ZERO_CODE_SIZE();
    }

    tdefl_optimize_huffman_table(d, 2, TDEFL_MAX_HUFF_SYMBOLS_2, 7, MZ_FALSE);

    TDEFL_PUT_BITS(2, 2);

    TDEFL_PUT_BITS(num_lit_codes - 257, 5);
    TDEFL_PUT_BITS(num_dist_codes - 1, 5);

    for (num_bit_lengths = 18; num_bit_lengths >= 0; num_bit_lengths--)
      if (d->m_huff_code_sizes
            [2][s_tdefl_packed_code_size_syms_swizzle[num_bit_lengths]])
        break;
    num_bit_lengths = MZ_MAX(4, (num_bit_lengths + 1));
    TDEFL_PUT_BITS(num_bit_lengths - 4, 4);
    for (i = 0; (int)i < num_bit_lengths; i++)
      TDEFL_PUT_BITS(
        d->m_huff_code_sizes[2][s_tdefl_packed_code_size_syms_swizzle[i]], 3);

    for (packed_code_sizes_index = 0;
         packed_code_sizes_index < num_packed_code_sizes;)
    {
      mz_uint code = packed_code_sizes[packed_code_sizes_index++];
      MZ_ASSERT(code < TDEFL_MAX_HUFF_SYMBOLS_2);
      TDEFL_PUT_BITS(d->m_huff_codes[2][code], d->m_huff_code_sizes[2][code]);
      if (code >= 16)
        TDEFL_PUT_BITS(
          packed_code_sizes[packed_code_sizes_index++], "\02\03\07"[code - 16]);
    }
  }

  static void tdefl_start_static_block(tdefl_compressor* d)
  {
    mz_uint i;
    mz_uint8* p = &d->m_huff_code_sizes[0][0];

    for (i = 0; i <= 143; ++i)
      *p++ = 8;
    for (; i <= 255; ++i)
      *p++ = 9;
    for (; i <= 279; ++i)
      *p++ = 7;
    for (; i <= 287; ++i)
      *p++ = 8;

    memset(d->m_huff_code_sizes[1], 5, 32);

    tdefl_optimize_huffman_table(d, 0, 288, 15, MZ_TRUE);
    tdefl_optimize_huffman_table(d, 1, 32, 15, MZ_TRUE);

    TDEFL_PUT_BITS(1, 2);
  }

  static const mz_uint mz_bitmasks[17] = {
    0x0000,
    0x0001,
    0x0003,
    0x0007,
    0x000F,
    0x001F,
    0x003F,
    0x007F,
    0x00FF,
    0x01FF,
    0x03FF,
    0x07FF,
    0x0FFF,
    0x1FFF,
    0x3FFF,
    0x7FFF,
    0xFFFF};

#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN && \
    MINIZ_HAS_64BIT_REGISTERS
  static mz_bool tdefl_compress_lz_codes(tdefl_compressor* d)
  {
    mz_uint flags;
    mz_uint8* pLZ_codes;
    mz_uint8* pOutput_buf = d->m_pOutput_buf;
    mz_uint8* pLZ_code_buf_end = d->m_pLZ_code_buf;
    mz_uint64 bit_buffer = d->m_bit_buffer;
    mz_uint bits_in = d->m_bits_in;

#    define TDEFL_PUT_BITS_FAST(b, l) \
      { \
        bit_buffer |= (((mz_uint64)(b)) << bits_in); \
        bits_in += (l); \
      }

    flags = 1;
    for (pLZ_codes = d->m_lz_code_buf; pLZ_codes < pLZ_code_buf_end;
         flags >>= 1)
    {
      if (flags == 1)
        flags = *pLZ_codes++ | 0x100;

      if (flags & 1)
      {
        mz_uint s0, s1, n0, n1, sym, num_extra_bits;
        mz_uint match_len = pLZ_codes[0],
                match_dist = *(const mz_uint16*)(pLZ_codes + 1);
        pLZ_codes += 3;

        MZ_ASSERT(d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]);
        TDEFL_PUT_BITS_FAST(
          d->m_huff_codes[0][s_tdefl_len_sym[match_len]],
          d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]);
        TDEFL_PUT_BITS_FAST(
          match_len & mz_bitmasks[s_tdefl_len_extra[match_len]],
          s_tdefl_len_extra[match_len]);

        // This sequence coaxes MSVC into using cmov's vs. jmp's.
        s0 = s_tdefl_small_dist_sym[match_dist & 511];
        n0 = s_tdefl_small_dist_extra[match_dist & 511];
        s1 = s_tdefl_large_dist_sym[match_dist >> 8];
        n1 = s_tdefl_large_dist_extra[match_dist >> 8];
        sym = (match_dist < 512) ? s0 : s1;
        num_extra_bits = (match_dist < 512) ? n0 : n1;

        MZ_ASSERT(d->m_huff_code_sizes[1][sym]);
        TDEFL_PUT_BITS_FAST(
          d->m_huff_codes[1][sym], d->m_huff_code_sizes[1][sym]);
        TDEFL_PUT_BITS_FAST(
          match_dist & mz_bitmasks[num_extra_bits], num_extra_bits);
      }
      else
      {
        mz_uint lit = *pLZ_codes++;
        MZ_ASSERT(d->m_huff_code_sizes[0][lit]);
        TDEFL_PUT_BITS_FAST(
          d->m_huff_codes[0][lit], d->m_huff_code_sizes[0][lit]);

        if (((flags & 2) == 0) && (pLZ_codes < pLZ_code_buf_end))
        {
          flags >>= 1;
          lit = *pLZ_codes++;
          MZ_ASSERT(d->m_huff_code_sizes[0][lit]);
          TDEFL_PUT_BITS_FAST(
            d->m_huff_codes[0][lit], d->m_huff_code_sizes[0][lit]);

          if (((flags & 2) == 0) && (pLZ_codes < pLZ_code_buf_end))
          {
            flags >>= 1;
            lit = *pLZ_codes++;
            MZ_ASSERT(d->m_huff_code_sizes[0][lit]);
            TDEFL_PUT_BITS_FAST(
              d->m_huff_codes[0][lit], d->m_huff_code_sizes[0][lit]);
          }
        }
      }

      if (pOutput_buf >= d->m_pOutput_buf_end)
        return MZ_FALSE;

      *(mz_uint64*)pOutput_buf = bit_buffer;
      pOutput_buf += (bits_in >> 3);
      bit_buffer >>= (bits_in & ~7);
      bits_in &= 7;
    }

#    undef TDEFL_PUT_BITS_FAST

    d->m_pOutput_buf = pOutput_buf;
    d->m_bits_in = 0;
    d->m_bit_buffer = 0;

    while (bits_in)
    {
      mz_uint32 n = MZ_MIN(bits_in, 16);
      TDEFL_PUT_BITS((mz_uint)bit_buffer & mz_bitmasks[n], n);
      bit_buffer >>= n;
      bits_in -= n;
    }

    TDEFL_PUT_BITS(d->m_huff_codes[0][256], d->m_huff_code_sizes[0][256]);

    return (d->m_pOutput_buf < d->m_pOutput_buf_end);
  }
#  else
static mz_bool tdefl_compress_lz_codes(tdefl_compressor* d)
{
  mz_uint flags;
  mz_uint8* pLZ_codes;

  flags = 1;
  for (pLZ_codes = d->m_lz_code_buf; pLZ_codes < d->m_pLZ_code_buf; flags >>= 1)
  {
    if (flags == 1)
      flags = *pLZ_codes++ | 0x100;
    if (flags & 1)
    {
      mz_uint sym, num_extra_bits;
      mz_uint match_len = pLZ_codes[0],
              match_dist = (pLZ_codes[1] | (pLZ_codes[2] << 8));
      pLZ_codes += 3;

      MZ_ASSERT(d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]);
      TDEFL_PUT_BITS(
        d->m_huff_codes[0][s_tdefl_len_sym[match_len]],
        d->m_huff_code_sizes[0][s_tdefl_len_sym[match_len]]);
      TDEFL_PUT_BITS(
        match_len & mz_bitmasks[s_tdefl_len_extra[match_len]],
        s_tdefl_len_extra[match_len]);

      if (match_dist < 512)
      {
        sym = s_tdefl_small_dist_sym[match_dist];
        num_extra_bits = s_tdefl_small_dist_extra[match_dist];
      }
      else
      {
        sym = s_tdefl_large_dist_sym[match_dist >> 8];
        num_extra_bits = s_tdefl_large_dist_extra[match_dist >> 8];
      }
      MZ_ASSERT(d->m_huff_code_sizes[1][sym]);
      TDEFL_PUT_BITS(d->m_huff_codes[1][sym], d->m_huff_code_sizes[1][sym]);
      TDEFL_PUT_BITS(match_dist & mz_bitmasks[num_extra_bits], num_extra_bits);
    }
    else
    {
      mz_uint lit = *pLZ_codes++;
      MZ_ASSERT(d->m_huff_code_sizes[0][lit]);
      TDEFL_PUT_BITS(d->m_huff_codes[0][lit], d->m_huff_code_sizes[0][lit]);
    }
  }

  TDEFL_PUT_BITS(d->m_huff_codes[0][256], d->m_huff_code_sizes[0][256]);

  return (d->m_pOutput_buf < d->m_pOutput_buf_end);
}
#  endif // MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN &&
         // MINIZ_HAS_64BIT_REGISTERS

  static mz_bool tdefl_compress_block(tdefl_compressor* d, mz_bool static_block)
  {
    if (static_block)
      tdefl_start_static_block(d);
    else
      tdefl_start_dynamic_block(d);
    return tdefl_compress_lz_codes(d);
  }

  static int tdefl_flush_block(tdefl_compressor* d, int flush)
  {
    mz_uint saved_bit_buf, saved_bits_in;
    mz_uint8* pSaved_output_buf;
    mz_bool comp_block_succeeded = MZ_FALSE;
    int n,
      use_raw_block = ((d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS) != 0) &&
      (d->m_lookahead_pos - d->m_lz_code_buf_dict_pos) <= d->m_dict_size;
    mz_uint8* pOutput_buf_start =
      ((d->m_pPut_buf_func == NULL) &&
       ((*d->m_pOut_buf_size - d->m_out_buf_ofs) >= TDEFL_OUT_BUF_SIZE)) ?
      ((mz_uint8*)d->m_pOut_buf + d->m_out_buf_ofs) :
      d->m_output_buf;

    d->m_pOutput_buf = pOutput_buf_start;
    d->m_pOutput_buf_end = d->m_pOutput_buf + TDEFL_OUT_BUF_SIZE - 16;

    MZ_ASSERT(!d->m_output_flush_remaining);
    d->m_output_flush_ofs = 0;
    d->m_output_flush_remaining = 0;

    *d->m_pLZ_flags = (mz_uint8)(*d->m_pLZ_flags >> d->m_num_flags_left);
    d->m_pLZ_code_buf -= (d->m_num_flags_left == 8);

    if ((d->m_flags & TDEFL_WRITE_ZLIB_HEADER) && (!d->m_block_index))
    {
      TDEFL_PUT_BITS(0x78, 8);
      TDEFL_PUT_BITS(0x01, 8);
    }

    TDEFL_PUT_BITS(flush == TDEFL_FINISH, 1);

    pSaved_output_buf = d->m_pOutput_buf;
    saved_bit_buf = d->m_bit_buffer;
    saved_bits_in = d->m_bits_in;

    if (!use_raw_block)
      comp_block_succeeded = tdefl_compress_block(
        d,
        (d->m_flags & TDEFL_FORCE_ALL_STATIC_BLOCKS) ||
          (d->m_total_lz_bytes < 48));

    // If the block gets expanded, forget the current contents of the output
    // buffer and send a raw block instead.
    if (
      ((use_raw_block) ||
       ((d->m_total_lz_bytes) &&
        ((d->m_pOutput_buf - pSaved_output_buf + 1U) >=
         d->m_total_lz_bytes))) &&
      ((d->m_lookahead_pos - d->m_lz_code_buf_dict_pos) <= d->m_dict_size))
    {
      mz_uint i;
      d->m_pOutput_buf = pSaved_output_buf;
      d->m_bit_buffer = saved_bit_buf, d->m_bits_in = saved_bits_in;
      TDEFL_PUT_BITS(0, 2);
      if (d->m_bits_in)
      {
        TDEFL_PUT_BITS(0, 8 - d->m_bits_in);
      }
      for (i = 2; i; --i, d->m_total_lz_bytes ^= 0xFFFF)
      {
        TDEFL_PUT_BITS(d->m_total_lz_bytes & 0xFFFF, 16);
      }
      for (i = 0; i < d->m_total_lz_bytes; ++i)
      {
        TDEFL_PUT_BITS(
          d->m_dict[(d->m_lz_code_buf_dict_pos + i) & TDEFL_LZ_DICT_SIZE_MASK],
          8);
      }
    }
    // Check for the extremely unlikely (if not impossible) case of the
    // compressed block not fitting into the output buffer when using dynamic
    // codes.
    else if (!comp_block_succeeded)
    {
      d->m_pOutput_buf = pSaved_output_buf;
      d->m_bit_buffer = saved_bit_buf, d->m_bits_in = saved_bits_in;
      tdefl_compress_block(d, MZ_TRUE);
    }

    if (flush)
    {
      if (flush == TDEFL_FINISH)
      {
        if (d->m_bits_in)
        {
          TDEFL_PUT_BITS(0, 8 - d->m_bits_in);
        }
        if (d->m_flags & TDEFL_WRITE_ZLIB_HEADER)
        {
          mz_uint i, a = d->m_adler32;
          for (i = 0; i < 4; i++)
          {
            TDEFL_PUT_BITS((a >> 24) & 0xFF, 8);
            a <<= 8;
          }
        }
      }
      else
      {
        mz_uint i, z = 0;
        TDEFL_PUT_BITS(0, 3);
        if (d->m_bits_in)
        {
          TDEFL_PUT_BITS(0, 8 - d->m_bits_in);
        }
        for (i = 2; i; --i, z ^= 0xFFFF)
        {
          TDEFL_PUT_BITS(z & 0xFFFF, 16);
        }
      }
    }

    MZ_ASSERT(d->m_pOutput_buf < d->m_pOutput_buf_end);

    memset(
      &d->m_huff_count[0][0],
      0,
      sizeof(d->m_huff_count[0][0]) * TDEFL_MAX_HUFF_SYMBOLS_0);
    memset(
      &d->m_huff_count[1][0],
      0,
      sizeof(d->m_huff_count[1][0]) * TDEFL_MAX_HUFF_SYMBOLS_1);

    d->m_pLZ_code_buf = d->m_lz_code_buf + 1;
    d->m_pLZ_flags = d->m_lz_code_buf;
    d->m_num_flags_left = 8;
    d->m_lz_code_buf_dict_pos += d->m_total_lz_bytes;
    d->m_total_lz_bytes = 0;
    d->m_block_index++;

    if ((n = (int)(d->m_pOutput_buf - pOutput_buf_start)) != 0)
    {
      if (d->m_pPut_buf_func)
      {
        *d->m_pIn_buf_size = d->m_pSrc - (const mz_uint8*)d->m_pIn_buf;
        if (!(*d->m_pPut_buf_func)(d->m_output_buf, n, d->m_pPut_buf_user))
          return (d->m_prev_return_status = TDEFL_STATUS_PUT_BUF_FAILED);
      }
      else if (pOutput_buf_start == d->m_output_buf)
      {
        int bytes_to_copy = (int)MZ_MIN(
          (size_t)n, (size_t)(*d->m_pOut_buf_size - d->m_out_buf_ofs));
        memcpy(
          (mz_uint8*)d->m_pOut_buf + d->m_out_buf_ofs,
          d->m_output_buf,
          bytes_to_copy);
        d->m_out_buf_ofs += bytes_to_copy;
        if ((n -= bytes_to_copy) != 0)
        {
          d->m_output_flush_ofs = bytes_to_copy;
          d->m_output_flush_remaining = n;
        }
      }
      else
      {
        d->m_out_buf_ofs += n;
      }
    }

    return d->m_output_flush_remaining;
  }

#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES
#    define TDEFL_READ_UNALIGNED_WORD(p) *(const mz_uint16*)(p)
  static MZ_FORCEINLINE void tdefl_find_match(
    tdefl_compressor* d,
    mz_uint lookahead_pos,
    mz_uint max_dist,
    mz_uint max_match_len,
    mz_uint* pMatch_dist,
    mz_uint* pMatch_len)
  {
    mz_uint dist, pos = lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK,
                  match_len = *pMatch_len, probe_pos = pos, next_probe_pos,
                  probe_len;
    mz_uint num_probes_left = d->m_max_probes[match_len >= 32];
    const mz_uint16 *s = (const mz_uint16*)(d->m_dict + pos), *p, *q;
    mz_uint16 c01 = TDEFL_READ_UNALIGNED_WORD(&d->m_dict[pos + match_len - 1]),
              s01 = TDEFL_READ_UNALIGNED_WORD(s);
    MZ_ASSERT(max_match_len <= TDEFL_MAX_MATCH_LEN);
    if (max_match_len <= match_len)
      return;
    for (;;)
    {
      for (;;)
      {
        if (--num_probes_left == 0)
          return;
#    define TDEFL_PROBE \
      next_probe_pos = d->m_next[probe_pos]; \
      if ( \
        (!next_probe_pos) || \
        ((dist = (mz_uint16)(lookahead_pos - next_probe_pos)) > max_dist)) \
        return; \
      probe_pos = next_probe_pos & TDEFL_LZ_DICT_SIZE_MASK; \
      if ( \
        TDEFL_READ_UNALIGNED_WORD(&d->m_dict[probe_pos + match_len - 1]) == \
        c01) \
        break;
        TDEFL_PROBE;
        TDEFL_PROBE;
        TDEFL_PROBE;
      }
      if (!dist)
        break;
      q = (const mz_uint16*)(d->m_dict + probe_pos);
      if (TDEFL_READ_UNALIGNED_WORD(q) != s01)
        continue;
      p = s;
      probe_len = 32;
      do
      {
      } while (
        (TDEFL_READ_UNALIGNED_WORD(++p) == TDEFL_READ_UNALIGNED_WORD(++q)) &&
        (TDEFL_READ_UNALIGNED_WORD(++p) == TDEFL_READ_UNALIGNED_WORD(++q)) &&
        (TDEFL_READ_UNALIGNED_WORD(++p) == TDEFL_READ_UNALIGNED_WORD(++q)) &&
        (TDEFL_READ_UNALIGNED_WORD(++p) == TDEFL_READ_UNALIGNED_WORD(++q)) &&
        (--probe_len > 0));
      if (!probe_len)
      {
        *pMatch_dist = dist;
        *pMatch_len = MZ_MIN(max_match_len, TDEFL_MAX_MATCH_LEN);
        break;
      }
      else if (
        (probe_len = ((mz_uint)(p - s) * 2) +
           (mz_uint)(*(const mz_uint8*)p == *(const mz_uint8*)q)) > match_len)
      {
        *pMatch_dist = dist;
        if (
          (*pMatch_len = match_len = MZ_MIN(max_match_len, probe_len)) ==
          max_match_len)
          break;
        c01 = TDEFL_READ_UNALIGNED_WORD(&d->m_dict[pos + match_len - 1]);
      }
    }
  }
#  else
static MZ_FORCEINLINE void tdefl_find_match(
  tdefl_compressor* d,
  mz_uint lookahead_pos,
  mz_uint max_dist,
  mz_uint max_match_len,
  mz_uint* pMatch_dist,
  mz_uint* pMatch_len)
{
  mz_uint dist, pos = lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK,
                match_len = *pMatch_len, probe_pos = pos, next_probe_pos,
                probe_len;
  mz_uint num_probes_left = d->m_max_probes[match_len >= 32];
  const mz_uint8 *s = d->m_dict + pos, *p, *q;
  mz_uint8 c0 = d->m_dict[pos + match_len], c1 = d->m_dict[pos + match_len - 1];
  MZ_ASSERT(max_match_len <= TDEFL_MAX_MATCH_LEN);
  if (max_match_len <= match_len)
    return;
  for (;;)
  {
    for (;;)
    {
      if (--num_probes_left == 0)
        return;
#    define TDEFL_PROBE \
      next_probe_pos = d->m_next[probe_pos]; \
      if ( \
        (!next_probe_pos) || \
        ((dist = (mz_uint16)(lookahead_pos - next_probe_pos)) > max_dist)) \
        return; \
      probe_pos = next_probe_pos & TDEFL_LZ_DICT_SIZE_MASK; \
      if ( \
        (d->m_dict[probe_pos + match_len] == c0) && \
        (d->m_dict[probe_pos + match_len - 1] == c1)) \
        break;
      TDEFL_PROBE;
      TDEFL_PROBE;
      TDEFL_PROBE;
    }
    if (!dist)
      break;
    p = s;
    q = d->m_dict + probe_pos;
    for (probe_len = 0; probe_len < max_match_len; probe_len++)
      if (*p++ != *q++)
        break;
    if (probe_len > match_len)
    {
      *pMatch_dist = dist;
      if ((*pMatch_len = match_len = probe_len) == max_match_len)
        return;
      c0 = d->m_dict[pos + match_len];
      c1 = d->m_dict[pos + match_len - 1];
    }
  }
}
#  endif // #if MINIZ_USE_UNALIGNED_LOADS_AND_STORES

#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN
  static mz_bool tdefl_compress_fast(tdefl_compressor* d)
  {
    // Faster, minimally featured LZRW1-style match+parse loop with better
    // register utilization. Intended for applications where raw throughput is
    // valued more highly than ratio.
    mz_uint lookahead_pos = d->m_lookahead_pos,
            lookahead_size = d->m_lookahead_size, dict_size = d->m_dict_size,
            total_lz_bytes = d->m_total_lz_bytes,
            num_flags_left = d->m_num_flags_left;
    mz_uint8 *pLZ_code_buf = d->m_pLZ_code_buf, *pLZ_flags = d->m_pLZ_flags;
    mz_uint cur_pos = lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK;

    while ((d->m_src_buf_left) || ((d->m_flush) && (lookahead_size)))
    {
      const mz_uint TDEFL_COMP_FAST_LOOKAHEAD_SIZE = 4096;
      mz_uint dst_pos =
        (lookahead_pos + lookahead_size) & TDEFL_LZ_DICT_SIZE_MASK;
      mz_uint num_bytes_to_process = (mz_uint)MZ_MIN(
        d->m_src_buf_left, TDEFL_COMP_FAST_LOOKAHEAD_SIZE - lookahead_size);
      d->m_src_buf_left -= num_bytes_to_process;
      lookahead_size += num_bytes_to_process;

      while (num_bytes_to_process)
      {
        mz_uint32 n =
          MZ_MIN(TDEFL_LZ_DICT_SIZE - dst_pos, num_bytes_to_process);
        memcpy(d->m_dict + dst_pos, d->m_pSrc, n);
        if (dst_pos < (TDEFL_MAX_MATCH_LEN - 1))
          memcpy(
            d->m_dict + TDEFL_LZ_DICT_SIZE + dst_pos,
            d->m_pSrc,
            MZ_MIN(n, (TDEFL_MAX_MATCH_LEN - 1) - dst_pos));
        d->m_pSrc += n;
        dst_pos = (dst_pos + n) & TDEFL_LZ_DICT_SIZE_MASK;
        num_bytes_to_process -= n;
      }

      dict_size = MZ_MIN(TDEFL_LZ_DICT_SIZE - lookahead_size, dict_size);
      if ((!d->m_flush) && (lookahead_size < TDEFL_COMP_FAST_LOOKAHEAD_SIZE))
        break;

      while (lookahead_size >= 4)
      {
        mz_uint cur_match_dist, cur_match_len = 1;
        mz_uint8* pCur_dict = d->m_dict + cur_pos;
        mz_uint first_trigram = (*(const mz_uint32*)pCur_dict) & 0xFFFFFF;
        mz_uint hash =
          (first_trigram ^ (first_trigram >> (24 - (TDEFL_LZ_HASH_BITS - 8)))) &
          TDEFL_LEVEL1_HASH_SIZE_MASK;
        mz_uint probe_pos = d->m_hash[hash];
        d->m_hash[hash] = (mz_uint16)lookahead_pos;

        if (
          ((cur_match_dist = (mz_uint16)(lookahead_pos - probe_pos)) <=
           dict_size) &&
          ((*(const mz_uint32*)(d->m_dict + (probe_pos &= TDEFL_LZ_DICT_SIZE_MASK)) &
            0xFFFFFF) == first_trigram))
        {
          const mz_uint16* p = (const mz_uint16*)pCur_dict;
          const mz_uint16* q = (const mz_uint16*)(d->m_dict + probe_pos);
          mz_uint32 probe_len = 32;
          do
          {
          } while ((TDEFL_READ_UNALIGNED_WORD(++p) ==
                    TDEFL_READ_UNALIGNED_WORD(++q)) &&
                   (TDEFL_READ_UNALIGNED_WORD(++p) ==
                    TDEFL_READ_UNALIGNED_WORD(++q)) &&
                   (TDEFL_READ_UNALIGNED_WORD(++p) ==
                    TDEFL_READ_UNALIGNED_WORD(++q)) &&
                   (TDEFL_READ_UNALIGNED_WORD(++p) ==
                    TDEFL_READ_UNALIGNED_WORD(++q)) &&
                   (--probe_len > 0));
          cur_match_len = ((mz_uint)(p - (const mz_uint16*)pCur_dict) * 2) +
            (mz_uint)(*(const mz_uint8*)p == *(const mz_uint8*)q);
          if (!probe_len)
            cur_match_len = cur_match_dist ? TDEFL_MAX_MATCH_LEN : 0;

          if (
            (cur_match_len < TDEFL_MIN_MATCH_LEN) ||
            ((cur_match_len == TDEFL_MIN_MATCH_LEN) &&
             (cur_match_dist >= 8U * 1024U)))
          {
            cur_match_len = 1;
            *pLZ_code_buf++ = (mz_uint8)first_trigram;
            *pLZ_flags = (mz_uint8)(*pLZ_flags >> 1);
            d->m_huff_count[0][(mz_uint8)first_trigram]++;
          }
          else
          {
            mz_uint32 s0, s1;
            cur_match_len = MZ_MIN(cur_match_len, lookahead_size);

            MZ_ASSERT(
              (cur_match_len >= TDEFL_MIN_MATCH_LEN) && (cur_match_dist >= 1) &&
              (cur_match_dist <= TDEFL_LZ_DICT_SIZE));

            cur_match_dist--;

            pLZ_code_buf[0] = (mz_uint8)(cur_match_len - TDEFL_MIN_MATCH_LEN);
            *(mz_uint16*)(&pLZ_code_buf[1]) = (mz_uint16)cur_match_dist;
            pLZ_code_buf += 3;
            *pLZ_flags = (mz_uint8)((*pLZ_flags >> 1) | 0x80);

            s0 = s_tdefl_small_dist_sym[cur_match_dist & 511];
            s1 = s_tdefl_large_dist_sym[cur_match_dist >> 8];
            d->m_huff_count[1][(cur_match_dist < 512) ? s0 : s1]++;

            d->m_huff_count
              [0][s_tdefl_len_sym[cur_match_len - TDEFL_MIN_MATCH_LEN]]++;
          }
        }
        else
        {
          *pLZ_code_buf++ = (mz_uint8)first_trigram;
          *pLZ_flags = (mz_uint8)(*pLZ_flags >> 1);
          d->m_huff_count[0][(mz_uint8)first_trigram]++;
        }

        if (--num_flags_left == 0)
        {
          num_flags_left = 8;
          pLZ_flags = pLZ_code_buf++;
        }

        total_lz_bytes += cur_match_len;
        lookahead_pos += cur_match_len;
        dict_size = MZ_MIN(dict_size + cur_match_len, TDEFL_LZ_DICT_SIZE);
        cur_pos = (cur_pos + cur_match_len) & TDEFL_LZ_DICT_SIZE_MASK;
        MZ_ASSERT(lookahead_size >= cur_match_len);
        lookahead_size -= cur_match_len;

        if (pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8])
        {
          int n;
          d->m_lookahead_pos = lookahead_pos;
          d->m_lookahead_size = lookahead_size;
          d->m_dict_size = dict_size;
          d->m_total_lz_bytes = total_lz_bytes;
          d->m_pLZ_code_buf = pLZ_code_buf;
          d->m_pLZ_flags = pLZ_flags;
          d->m_num_flags_left = num_flags_left;
          if ((n = tdefl_flush_block(d, 0)) != 0)
            return (n < 0) ? MZ_FALSE : MZ_TRUE;
          total_lz_bytes = d->m_total_lz_bytes;
          pLZ_code_buf = d->m_pLZ_code_buf;
          pLZ_flags = d->m_pLZ_flags;
          num_flags_left = d->m_num_flags_left;
        }
      }

      while (lookahead_size)
      {
        mz_uint8 lit = d->m_dict[cur_pos];

        total_lz_bytes++;
        *pLZ_code_buf++ = lit;
        *pLZ_flags = (mz_uint8)(*pLZ_flags >> 1);
        if (--num_flags_left == 0)
        {
          num_flags_left = 8;
          pLZ_flags = pLZ_code_buf++;
        }

        d->m_huff_count[0][lit]++;

        lookahead_pos++;
        dict_size = MZ_MIN(dict_size + 1, TDEFL_LZ_DICT_SIZE);
        cur_pos = (cur_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK;
        lookahead_size--;

        if (pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8])
        {
          int n;
          d->m_lookahead_pos = lookahead_pos;
          d->m_lookahead_size = lookahead_size;
          d->m_dict_size = dict_size;
          d->m_total_lz_bytes = total_lz_bytes;
          d->m_pLZ_code_buf = pLZ_code_buf;
          d->m_pLZ_flags = pLZ_flags;
          d->m_num_flags_left = num_flags_left;
          if ((n = tdefl_flush_block(d, 0)) != 0)
            return (n < 0) ? MZ_FALSE : MZ_TRUE;
          total_lz_bytes = d->m_total_lz_bytes;
          pLZ_code_buf = d->m_pLZ_code_buf;
          pLZ_flags = d->m_pLZ_flags;
          num_flags_left = d->m_num_flags_left;
        }
      }
    }

    d->m_lookahead_pos = lookahead_pos;
    d->m_lookahead_size = lookahead_size;
    d->m_dict_size = dict_size;
    d->m_total_lz_bytes = total_lz_bytes;
    d->m_pLZ_code_buf = pLZ_code_buf;
    d->m_pLZ_flags = pLZ_flags;
    d->m_num_flags_left = num_flags_left;
    return MZ_TRUE;
  }
#  endif // MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN

  static MZ_FORCEINLINE void
  tdefl_record_literal(tdefl_compressor* d, mz_uint8 lit)
  {
    d->m_total_lz_bytes++;
    *d->m_pLZ_code_buf++ = lit;
    *d->m_pLZ_flags = (mz_uint8)(*d->m_pLZ_flags >> 1);
    if (--d->m_num_flags_left == 0)
    {
      d->m_num_flags_left = 8;
      d->m_pLZ_flags = d->m_pLZ_code_buf++;
    }
    d->m_huff_count[0][lit]++;
  }

  static MZ_FORCEINLINE void
  tdefl_record_match(tdefl_compressor* d, mz_uint match_len, mz_uint match_dist)
  {
    mz_uint32 s0, s1;

    MZ_ASSERT(
      (match_len >= TDEFL_MIN_MATCH_LEN) && (match_dist >= 1) &&
      (match_dist <= TDEFL_LZ_DICT_SIZE));

    d->m_total_lz_bytes += match_len;

    d->m_pLZ_code_buf[0] = (mz_uint8)(match_len - TDEFL_MIN_MATCH_LEN);

    match_dist -= 1;
    d->m_pLZ_code_buf[1] = (mz_uint8)(match_dist & 0xFF);
    d->m_pLZ_code_buf[2] = (mz_uint8)(match_dist >> 8);
    d->m_pLZ_code_buf += 3;

    *d->m_pLZ_flags = (mz_uint8)((*d->m_pLZ_flags >> 1) | 0x80);
    if (--d->m_num_flags_left == 0)
    {
      d->m_num_flags_left = 8;
      d->m_pLZ_flags = d->m_pLZ_code_buf++;
    }

    s0 = s_tdefl_small_dist_sym[match_dist & 511];
    s1 = s_tdefl_large_dist_sym[(match_dist >> 8) & 127];
    d->m_huff_count[1][(match_dist < 512) ? s0 : s1]++;

    if (match_len >= TDEFL_MIN_MATCH_LEN)
      d->m_huff_count[0][s_tdefl_len_sym[match_len - TDEFL_MIN_MATCH_LEN]]++;
  }

  static mz_bool tdefl_compress_normal(tdefl_compressor* d)
  {
    const mz_uint8* pSrc = d->m_pSrc;
    size_t src_buf_left = d->m_src_buf_left;
    tdefl_flush flush = d->m_flush;

    while ((src_buf_left) || ((flush) && (d->m_lookahead_size)))
    {
      mz_uint len_to_move, cur_match_dist, cur_match_len, cur_pos;
      // Update dictionary and hash chains. Keeps the lookahead size equal to
      // TDEFL_MAX_MATCH_LEN.
      if ((d->m_lookahead_size + d->m_dict_size) >= (TDEFL_MIN_MATCH_LEN - 1))
      {
        mz_uint dst_pos = (d->m_lookahead_pos + d->m_lookahead_size) &
          TDEFL_LZ_DICT_SIZE_MASK,
                ins_pos = d->m_lookahead_pos + d->m_lookahead_size - 2;
        mz_uint hash = (d->m_dict[ins_pos & TDEFL_LZ_DICT_SIZE_MASK]
                        << TDEFL_LZ_HASH_SHIFT) ^
          d->m_dict[(ins_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK];
        mz_uint num_bytes_to_process = (mz_uint)MZ_MIN(
          src_buf_left, TDEFL_MAX_MATCH_LEN - d->m_lookahead_size);
        const mz_uint8* pSrc_end = pSrc + num_bytes_to_process;
        src_buf_left -= num_bytes_to_process;
        d->m_lookahead_size += num_bytes_to_process;
        while (pSrc != pSrc_end)
        {
          mz_uint8 c = *pSrc++;
          d->m_dict[dst_pos] = c;
          if (dst_pos < (TDEFL_MAX_MATCH_LEN - 1))
            d->m_dict[TDEFL_LZ_DICT_SIZE + dst_pos] = c;
          hash = ((hash << TDEFL_LZ_HASH_SHIFT) ^ c) & (TDEFL_LZ_HASH_SIZE - 1);
          d->m_next[ins_pos & TDEFL_LZ_DICT_SIZE_MASK] = d->m_hash[hash];
          d->m_hash[hash] = (mz_uint16)(ins_pos);
          dst_pos = (dst_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK;
          ins_pos++;
        }
      }
      else
      {
        while ((src_buf_left) && (d->m_lookahead_size < TDEFL_MAX_MATCH_LEN))
        {
          mz_uint8 c = *pSrc++;
          mz_uint dst_pos = (d->m_lookahead_pos + d->m_lookahead_size) &
            TDEFL_LZ_DICT_SIZE_MASK;
          src_buf_left--;
          d->m_dict[dst_pos] = c;
          if (dst_pos < (TDEFL_MAX_MATCH_LEN - 1))
            d->m_dict[TDEFL_LZ_DICT_SIZE + dst_pos] = c;
          if ((++d->m_lookahead_size + d->m_dict_size) >= TDEFL_MIN_MATCH_LEN)
          {
            mz_uint ins_pos =
              d->m_lookahead_pos + (d->m_lookahead_size - 1) - 2;
            mz_uint hash = ((d->m_dict[ins_pos & TDEFL_LZ_DICT_SIZE_MASK]
                             << (TDEFL_LZ_HASH_SHIFT * 2)) ^
                            (d->m_dict[(ins_pos + 1) & TDEFL_LZ_DICT_SIZE_MASK]
                             << TDEFL_LZ_HASH_SHIFT) ^
                            c) &
              (TDEFL_LZ_HASH_SIZE - 1);
            d->m_next[ins_pos & TDEFL_LZ_DICT_SIZE_MASK] = d->m_hash[hash];
            d->m_hash[hash] = (mz_uint16)(ins_pos);
          }
        }
      }
      d->m_dict_size =
        MZ_MIN(TDEFL_LZ_DICT_SIZE - d->m_lookahead_size, d->m_dict_size);
      if ((!flush) && (d->m_lookahead_size < TDEFL_MAX_MATCH_LEN))
        break;

      // Simple lazy/greedy parsing state machine.
      len_to_move = 1;
      cur_match_dist = 0;
      cur_match_len =
        d->m_saved_match_len ? d->m_saved_match_len : (TDEFL_MIN_MATCH_LEN - 1);
      cur_pos = d->m_lookahead_pos & TDEFL_LZ_DICT_SIZE_MASK;
      if (d->m_flags & (TDEFL_RLE_MATCHES | TDEFL_FORCE_ALL_RAW_BLOCKS))
      {
        if ((d->m_dict_size) && (!(d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS)))
        {
          mz_uint8 c = d->m_dict[(cur_pos - 1) & TDEFL_LZ_DICT_SIZE_MASK];
          cur_match_len = 0;
          while (cur_match_len < d->m_lookahead_size)
          {
            if (d->m_dict[cur_pos + cur_match_len] != c)
              break;
            cur_match_len++;
          }
          if (cur_match_len < TDEFL_MIN_MATCH_LEN)
            cur_match_len = 0;
          else
            cur_match_dist = 1;
        }
      }
      else
      {
        tdefl_find_match(
          d,
          d->m_lookahead_pos,
          d->m_dict_size,
          d->m_lookahead_size,
          &cur_match_dist,
          &cur_match_len);
      }
      if (
        ((cur_match_len == TDEFL_MIN_MATCH_LEN) &&
         (cur_match_dist >= 8U * 1024U)) ||
        (cur_pos == cur_match_dist) ||
        ((d->m_flags & TDEFL_FILTER_MATCHES) && (cur_match_len <= 5)))
      {
        cur_match_dist = cur_match_len = 0;
      }
      if (d->m_saved_match_len)
      {
        if (cur_match_len > d->m_saved_match_len)
        {
          tdefl_record_literal(d, (mz_uint8)d->m_saved_lit);
          if (cur_match_len >= 128)
          {
            tdefl_record_match(d, cur_match_len, cur_match_dist);
            d->m_saved_match_len = 0;
            len_to_move = cur_match_len;
          }
          else
          {
            d->m_saved_lit = d->m_dict[cur_pos];
            d->m_saved_match_dist = cur_match_dist;
            d->m_saved_match_len = cur_match_len;
          }
        }
        else
        {
          tdefl_record_match(d, d->m_saved_match_len, d->m_saved_match_dist);
          len_to_move = d->m_saved_match_len - 1;
          d->m_saved_match_len = 0;
        }
      }
      else if (!cur_match_dist)
        tdefl_record_literal(
          d, d->m_dict[MZ_MIN(cur_pos, sizeof(d->m_dict) - 1)]);
      else if (
        (d->m_greedy_parsing) || (d->m_flags & TDEFL_RLE_MATCHES) ||
        (cur_match_len >= 128))
      {
        tdefl_record_match(d, cur_match_len, cur_match_dist);
        len_to_move = cur_match_len;
      }
      else
      {
        d->m_saved_lit = d->m_dict[MZ_MIN(cur_pos, sizeof(d->m_dict) - 1)];
        d->m_saved_match_dist = cur_match_dist;
        d->m_saved_match_len = cur_match_len;
      }
      // Move the lookahead forward by len_to_move bytes.
      d->m_lookahead_pos += len_to_move;
      MZ_ASSERT(d->m_lookahead_size >= len_to_move);
      d->m_lookahead_size -= len_to_move;
      d->m_dict_size = MZ_MIN(d->m_dict_size + len_to_move, TDEFL_LZ_DICT_SIZE);
      // Check if it's time to flush the current LZ codes to the internal output
      // buffer.
      if (
        (d->m_pLZ_code_buf > &d->m_lz_code_buf[TDEFL_LZ_CODE_BUF_SIZE - 8]) ||
        ((d->m_total_lz_bytes > 31 * 1024) &&
         (((((mz_uint)(d->m_pLZ_code_buf - d->m_lz_code_buf) * 115) >> 7) >=
           d->m_total_lz_bytes) ||
          (d->m_flags & TDEFL_FORCE_ALL_RAW_BLOCKS))))
      {
        int n;
        d->m_pSrc = pSrc;
        d->m_src_buf_left = src_buf_left;
        if ((n = tdefl_flush_block(d, 0)) != 0)
          return (n < 0) ? MZ_FALSE : MZ_TRUE;
      }
    }

    d->m_pSrc = pSrc;
    d->m_src_buf_left = src_buf_left;
    return MZ_TRUE;
  }

  static tdefl_status tdefl_flush_output_buffer(tdefl_compressor* d)
  {
    if (d->m_pIn_buf_size)
    {
      *d->m_pIn_buf_size = d->m_pSrc - (const mz_uint8*)d->m_pIn_buf;
    }

    if (d->m_pOut_buf_size)
    {
      size_t n = MZ_MIN(
        *d->m_pOut_buf_size - d->m_out_buf_ofs, d->m_output_flush_remaining);
      memcpy(
        (mz_uint8*)d->m_pOut_buf + d->m_out_buf_ofs,
        d->m_output_buf + d->m_output_flush_ofs,
        n);
      d->m_output_flush_ofs += (mz_uint)n;
      d->m_output_flush_remaining -= (mz_uint)n;
      d->m_out_buf_ofs += n;

      *d->m_pOut_buf_size = d->m_out_buf_ofs;
    }

    return (d->m_finished && !d->m_output_flush_remaining) ? TDEFL_STATUS_DONE :
                                                             TDEFL_STATUS_OKAY;
  }

  tdefl_status tdefl_compress(
    tdefl_compressor* d,
    const void* pIn_buf,
    size_t* pIn_buf_size,
    void* pOut_buf,
    size_t* pOut_buf_size,
    tdefl_flush flush)
  {
    if (!d)
    {
      if (pIn_buf_size)
        *pIn_buf_size = 0;
      if (pOut_buf_size)
        *pOut_buf_size = 0;
      return TDEFL_STATUS_BAD_PARAM;
    }

    d->m_pIn_buf = pIn_buf;
    d->m_pIn_buf_size = pIn_buf_size;
    d->m_pOut_buf = pOut_buf;
    d->m_pOut_buf_size = pOut_buf_size;
    d->m_pSrc = (const mz_uint8*)(pIn_buf);
    d->m_src_buf_left = pIn_buf_size ? *pIn_buf_size : 0;
    d->m_out_buf_ofs = 0;
    d->m_flush = flush;

    if (
      ((d->m_pPut_buf_func != NULL) ==
       ((pOut_buf != NULL) || (pOut_buf_size != NULL))) ||
      (d->m_prev_return_status != TDEFL_STATUS_OKAY) ||
      (d->m_wants_to_finish && (flush != TDEFL_FINISH)) ||
      (pIn_buf_size && *pIn_buf_size && !pIn_buf) ||
      (pOut_buf_size && *pOut_buf_size && !pOut_buf))
    {
      if (pIn_buf_size)
        *pIn_buf_size = 0;
      if (pOut_buf_size)
        *pOut_buf_size = 0;
      return (d->m_prev_return_status = TDEFL_STATUS_BAD_PARAM);
    }
    d->m_wants_to_finish |= (flush == TDEFL_FINISH);

    if ((d->m_output_flush_remaining) || (d->m_finished))
      return (d->m_prev_return_status = tdefl_flush_output_buffer(d));

#  if MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN
    if (
      ((d->m_flags & TDEFL_MAX_PROBES_MASK) == 1) &&
      ((d->m_flags & TDEFL_GREEDY_PARSING_FLAG) != 0) &&
      ((d->m_flags &
        (TDEFL_FILTER_MATCHES | TDEFL_FORCE_ALL_RAW_BLOCKS |
         TDEFL_RLE_MATCHES)) == 0))
    {
      if (!tdefl_compress_fast(d))
        return d->m_prev_return_status;
    }
    else
#  endif // #if MINIZ_USE_UNALIGNED_LOADS_AND_STORES && MINIZ_LITTLE_ENDIAN
    {
      if (!tdefl_compress_normal(d))
        return d->m_prev_return_status;
    }

    if (
      (d->m_flags & (TDEFL_WRITE_ZLIB_HEADER | TDEFL_COMPUTE_ADLER32)) &&
      (pIn_buf))
      d->m_adler32 = (mz_uint32)mz_adler32(
        d->m_adler32,
        (const mz_uint8*)pIn_buf,
        d->m_pSrc - (const mz_uint8*)pIn_buf);

    if (
      (flush) && (!d->m_lookahead_size) && (!d->m_src_buf_left) &&
      (!d->m_output_flush_remaining))
    {
      if (tdefl_flush_block(d, flush) < 0)
        return d->m_prev_return_status;
      d->m_finished = (flush == TDEFL_FINISH);
      if (flush == TDEFL_FULL_FLUSH)
      {
        MZ_CLEAR_OBJ(d->m_hash);
        MZ_CLEAR_OBJ(d->m_next);
        d->m_dict_size = 0;
      }
    }

    return (d->m_prev_return_status = tdefl_flush_output_buffer(d));
  }

  tdefl_status tdefl_compress_buffer(
    tdefl_compressor* d,
    const void* pIn_buf,
    size_t in_buf_size,
    tdefl_flush flush)
  {
    MZ_ASSERT(d->m_pPut_buf_func);
    return tdefl_compress(d, pIn_buf, &in_buf_size, NULL, NULL, flush);
  }

  tdefl_status tdefl_init(
    tdefl_compressor* d,
    tdefl_put_buf_func_ptr pPut_buf_func,
    void* pPut_buf_user,
    int flags)
  {
    d->m_pPut_buf_func = pPut_buf_func;
    d->m_pPut_buf_user = pPut_buf_user;
    d->m_flags = (mz_uint)(flags);
    d->m_max_probes[0] = 1 + ((flags & 0xFFF) + 2) / 3;
    d->m_greedy_parsing = (flags & TDEFL_GREEDY_PARSING_FLAG) != 0;
    d->m_max_probes[1] = 1 + (((flags & 0xFFF) >> 2) + 2) / 3;
    if (!(flags & TDEFL_NONDETERMINISTIC_PARSING_FLAG))
      MZ_CLEAR_OBJ(d->m_hash);
    d->m_lookahead_pos = d->m_lookahead_size = d->m_dict_size =
      d->m_total_lz_bytes = d->m_lz_code_buf_dict_pos = d->m_bits_in = 0;
    d->m_output_flush_ofs = d->m_output_flush_remaining = d->m_finished =
      d->m_block_index = d->m_bit_buffer = d->m_wants_to_finish = 0;
    d->m_pLZ_code_buf = d->m_lz_code_buf + 1;
    d->m_pLZ_flags = d->m_lz_code_buf;
    d->m_num_flags_left = 8;
    d->m_pOutput_buf = d->m_output_buf;
    d->m_pOutput_buf_end = d->m_output_buf;
    d->m_prev_return_status = TDEFL_STATUS_OKAY;
    d->m_saved_match_dist = d->m_saved_match_len = d->m_saved_lit = 0;
    d->m_adler32 = 1;
    d->m_pIn_buf = NULL;
    d->m_pOut_buf = NULL;
    d->m_pIn_buf_size = NULL;
    d->m_pOut_buf_size = NULL;
    d->m_flush = TDEFL_NO_FLUSH;
    d->m_pSrc = NULL;
    d->m_src_buf_left = 0;
    d->m_out_buf_ofs = 0;
    memset(
      &d->m_huff_count[0][0],
      0,
      sizeof(d->m_huff_count[0][0]) * TDEFL_MAX_HUFF_SYMBOLS_0);
    memset(
      &d->m_huff_count[1][0],
      0,
      sizeof(d->m_huff_count[1][0]) * TDEFL_MAX_HUFF_SYMBOLS_1);
    return TDEFL_STATUS_OKAY;
  }

  tdefl_status tdefl_get_prev_return_status(tdefl_compressor* d)
  {
    return d->m_prev_return_status;
  }

  mz_uint32 tdefl_get_adler32(tdefl_compressor* d)
  {
    return d->m_adler32;
  }

  mz_bool tdefl_compress_mem_to_output(
    const void* pBuf,
    size_t buf_len,
    tdefl_put_buf_func_ptr pPut_buf_func,
    void* pPut_buf_user,
    int flags)
  {
    tdefl_compressor* pComp;
    mz_bool succeeded;
    if (((buf_len) && (!pBuf)) || (!pPut_buf_func))
      return MZ_FALSE;
    pComp = (tdefl_compressor*)MZ_MALLOC(sizeof(tdefl_compressor));
    if (!pComp)
      return MZ_FALSE;
    succeeded =
      (tdefl_init(pComp, pPut_buf_func, pPut_buf_user, flags) ==
       TDEFL_STATUS_OKAY);
    succeeded = succeeded &&
      (tdefl_compress_buffer(pComp, pBuf, buf_len, TDEFL_FINISH) ==
       TDEFL_STATUS_DONE);
    MZ_FREE(pComp);
    return succeeded;
  }

  typedef struct
  {
    size_t m_size, m_capacity;
    mz_uint8* m_pBuf;
    mz_bool m_expandable;
  } tdefl_output_buffer;

  static mz_bool
  tdefl_output_buffer_putter(const void* pBuf, int len, void* pUser)
  {
    tdefl_output_buffer* p = (tdefl_output_buffer*)pUser;
    size_t new_size = p->m_size + len;
    if (new_size > p->m_capacity)
    {
      size_t new_capacity = p->m_capacity;
      mz_uint8* pNew_buf;
      if (!p->m_expandable)
        return MZ_FALSE;
      do
      {
        new_capacity = MZ_MAX(128U, new_capacity << 1U);
      } while (new_size > new_capacity);
      pNew_buf = (mz_uint8*)MZ_REALLOC(p->m_pBuf, new_capacity);
      if (!pNew_buf)
        return MZ_FALSE;
      p->m_pBuf = pNew_buf;
      p->m_capacity = new_capacity;
    }
    memcpy((mz_uint8*)p->m_pBuf + p->m_size, pBuf, len);
    p->m_size = new_size;
    return MZ_TRUE;
  }

  void* tdefl_compress_mem_to_heap(
    const void* pSrc_buf, size_t src_buf_len, size_t* pOut_len, int flags)
  {
    tdefl_output_buffer out_buf;
    MZ_CLEAR_OBJ(out_buf);
    if (!pOut_len)
      return MZ_FALSE;
    else
      *pOut_len = 0;
    out_buf.m_expandable = MZ_TRUE;
    if (!tdefl_compress_mem_to_output(
          pSrc_buf, src_buf_len, tdefl_output_buffer_putter, &out_buf, flags))
      return NULL;
    *pOut_len = out_buf.m_size;
    return out_buf.m_pBuf;
  }

  size_t tdefl_compress_mem_to_mem(
    void* pOut_buf,
    size_t out_buf_len,
    const void* pSrc_buf,
    size_t src_buf_len,
    int flags)
  {
    tdefl_output_buffer out_buf;
    MZ_CLEAR_OBJ(out_buf);
    if (!pOut_buf)
      return 0;
    out_buf.m_pBuf = (mz_uint8*)pOut_buf;
    out_buf.m_capacity = out_buf_len;
    if (!tdefl_compress_mem_to_output(
          pSrc_buf, src_buf_len, tdefl_output_buffer_putter, &out_buf, flags))
      return 0;
    return out_buf.m_size;
  }

#  ifndef MINIZ_NO_ZLIB_APIS
  static const mz_uint s_tdefl_num_probes[11] = {
    0, 1, 6, 32, 16, 32, 128, 256, 512, 768, 1500};

  // level may actually range from [0,10] (10 is a "hidden" max level, where we
  // want a bit more compression and it's fine if throughput to fall off a cliff
  // on some files).
  mz_uint tdefl_create_comp_flags_from_zip_params(
    int level, int window_bits, int strategy)
  {
    mz_uint comp_flags =
      s_tdefl_num_probes[(level >= 0) ? MZ_MIN(10, level) : MZ_DEFAULT_LEVEL] |
      ((level <= 3) ? TDEFL_GREEDY_PARSING_FLAG : 0);
    if (window_bits > 0)
      comp_flags |= TDEFL_WRITE_ZLIB_HEADER;

    if (!level)
      comp_flags |= TDEFL_FORCE_ALL_RAW_BLOCKS;
    else if (strategy == MZ_FILTERED)
      comp_flags |= TDEFL_FILTER_MATCHES;
    else if (strategy == MZ_HUFFMAN_ONLY)
      comp_flags &= ~TDEFL_MAX_PROBES_MASK;
    else if (strategy == MZ_FIXED)
      comp_flags |= TDEFL_FORCE_ALL_STATIC_BLOCKS;
    else if (strategy == MZ_RLE)
      comp_flags |= TDEFL_RLE_MATCHES;

    return comp_flags;
  }
#  endif // MINIZ_NO_ZLIB_APIS

#  ifdef _MSC_VER
#    pragma warning(push)
#    pragma warning(disable : 4204) // nonstandard extension used : non-constant
                                    // aggregate initializer (also supported by
                                    // GNU C and C99, so no big deal)
#  endif

  // Simple PNG writer function by Alex Evans, 2011. Released into the public
  // domain: https://gist.github.com/908299, more context at
  // http://altdevblogaday.org/2011/04/06/a-smaller-jpg-encoder/.
  // This is actually a modification of Alex's original code so PNG files
  // generated by this function pass pngcheck.
  void* tdefl_write_image_to_png_file_in_memory_ex(
    const void* pImage,
    int w,
    int h,
    int num_chans,
    size_t* pLen_out,
    mz_uint level,
    mz_bool flip)
  {
    // Using a local copy of this array here in case MINIZ_NO_ZLIB_APIS was
    // defined.
    static const mz_uint s_tdefl_png_num_probes[11] = {
      0, 1, 6, 32, 16, 32, 128, 256, 512, 768, 1500};
    tdefl_compressor* pComp =
      (tdefl_compressor*)MZ_MALLOC(sizeof(tdefl_compressor));
    tdefl_output_buffer out_buf;
    int i, bpl = w * num_chans, y, z;
    mz_uint32 c;
    *pLen_out = 0;
    if (!pComp)
      return NULL;
    MZ_CLEAR_OBJ(out_buf);
    out_buf.m_expandable = MZ_TRUE;
    out_buf.m_capacity = 57 + MZ_MAX(64, (1 + bpl) * h);
    if (NULL == (out_buf.m_pBuf = (mz_uint8*)MZ_MALLOC(out_buf.m_capacity)))
    {
      MZ_FREE(pComp);
      return NULL;
    }
    // write dummy header
    for (z = 41; z; --z)
      tdefl_output_buffer_putter(&z, 1, &out_buf);
    // compress image data
    tdefl_init(
      pComp,
      tdefl_output_buffer_putter,
      &out_buf,
      s_tdefl_png_num_probes[MZ_MIN(10, level)] | TDEFL_WRITE_ZLIB_HEADER);
    for (y = 0; y < h; ++y)
    {
      tdefl_compress_buffer(pComp, &z, 1, TDEFL_NO_FLUSH);
      tdefl_compress_buffer(
        pComp,
        (mz_uint8*)pImage + (flip ? (h - 1 - y) : y) * bpl,
        bpl,
        TDEFL_NO_FLUSH);
    }
    if (
      tdefl_compress_buffer(pComp, NULL, 0, TDEFL_FINISH) != TDEFL_STATUS_DONE)
    {
      MZ_FREE(pComp);
      MZ_FREE(out_buf.m_pBuf);
      return NULL;
    }
    // write real header
    *pLen_out = out_buf.m_size - 41;
    {
      static const mz_uint8 chans[] = {0x00, 0x00, 0x04, 0x02, 0x06};
      mz_uint8 pnghdr[41] = {
        0x89,
        0x50,
        0x4e,
        0x47,
        0x0d,
        0x0a,
        0x1a,
        0x0a,
        0x00,
        0x00,
        0x00,
        0x0d,
        0x49,
        0x48,
        0x44,
        0x52,
        0,
        0,
        (mz_uint8)(w >> 8),
        (mz_uint8)w,
        0,
        0,
        (mz_uint8)(h >> 8),
        (mz_uint8)h,
        8,
        chans[num_chans],
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        (mz_uint8)(*pLen_out >> 24),
        (mz_uint8)(*pLen_out >> 16),
        (mz_uint8)(*pLen_out >> 8),
        (mz_uint8)*pLen_out,
        0x49,
        0x44,
        0x41,
        0x54};
      c = (mz_uint32)mz_crc32(MZ_CRC32_INIT, pnghdr + 12, 17);
      for (i = 0; i < 4; ++i, c <<= 8)
        ((mz_uint8*)(pnghdr + 29))[i] = (mz_uint8)(c >> 24);
      memcpy(out_buf.m_pBuf, pnghdr, 41);
    }
    // write footer (IDAT CRC-32, followed by IEND chunk)
    if (!tdefl_output_buffer_putter(
          "\0\0\0\0\0\0\0\0\x49\x45\x4e\x44\xae\x42\x60\x82", 16, &out_buf))
    {
      *pLen_out = 0;
      MZ_FREE(pComp);
      MZ_FREE(out_buf.m_pBuf);
      return NULL;
    }
    c = (mz_uint32)mz_crc32(
      MZ_CRC32_INIT, out_buf.m_pBuf + 41 - 4, *pLen_out + 4);
    for (i = 0; i < 4; ++i, c <<= 8)
      (out_buf.m_pBuf + out_buf.m_size - 16)[i] = (mz_uint8)(c >> 24);
    // compute final size of file, grab compressed data buffer and return
    *pLen_out += 57;
    MZ_FREE(pComp);
    return out_buf.m_pBuf;
  }
  void* tdefl_write_image_to_png_file_in_memory(
    const void* pImage, int w, int h, int num_chans, size_t* pLen_out)
  {
    // Level 6 corresponds to TDEFL_DEFAULT_MAX_PROBES or MZ_DEFAULT_LEVEL (but
    // we can't depend on MZ_DEFAULT_LEVEL being available in case the zlib
    // API's where #defined out)
    return tdefl_write_image_to_png_file_in_memory_ex(
      pImage, w, h, num_chans, pLen_out, 6, MZ_FALSE);
  }

#  ifdef _MSC_VER
#    pragma warning(pop)
#  endif

  // ------------------- .ZIP archive reading

#  ifndef MINIZ_NO_ARCHIVE_APIS

#    ifdef MINIZ_NO_STDIO
#      define MZ_FILE void*
#    else
#      include <stdio.h>
#      include <sys/stat.h>

#      if defined(_MSC_VER) || defined(__MINGW64__)
  static FILE* mz_fopen(const char* pFilename, const char* pMode)
  {
    FILE* pFile = NULL;
    fopen_s(&pFile, pFilename, pMode);
    return pFile;
  }
  static FILE* mz_freopen(const char* pPath, const char* pMode, FILE* pStream)
  {
    FILE* pFile = NULL;
    if (freopen_s(&pFile, pPath, pMode, pStream))
      return NULL;
    return pFile;
  }
#        ifndef MINIZ_NO_TIME
#          include <sys/utime.h>
#        endif
#        define MZ_FILE FILE
#        define MZ_FOPEN mz_fopen
#        define MZ_FCLOSE fclose
#        define MZ_FREAD fread
#        define MZ_FWRITE fwrite
#        define MZ_FTELL64 _ftelli64
#        define MZ_FSEEK64 _fseeki64
#        define MZ_FILE_STAT_STRUCT _stat
#        define MZ_FILE_STAT _stat
#        define MZ_FFLUSH fflush
#        define MZ_FREOPEN mz_freopen
#        define MZ_DELETE_FILE remove
#      elif defined(__MINGW32__)
#        ifndef MINIZ_NO_TIME
#          include <sys/utime.h>
#        endif
#        define MZ_FILE FILE
#        define MZ_FOPEN(f, m) fopen(f, m)
#        define MZ_FCLOSE fclose
#        define MZ_FREAD fread
#        define MZ_FWRITE fwrite
#        define MZ_FTELL64 ftello64
#        define MZ_FSEEK64 fseeko64
#        define MZ_FILE_STAT_STRUCT _stat
#        define MZ_FILE_STAT _stat
#        define MZ_FFLUSH fflush
#        define MZ_FREOPEN(f, m, s) freopen(f, m, s)
#        define MZ_DELETE_FILE remove
#      elif defined(__TINYC__)
#        ifndef MINIZ_NO_TIME
#          include <sys/utime.h>
#        endif
#        define MZ_FILE FILE
#        define MZ_FOPEN(f, m) fopen(f, m)
#        define MZ_FCLOSE fclose
#        define MZ_FREAD fread
#        define MZ_FWRITE fwrite
#        define MZ_FTELL64 ftell
#        define MZ_FSEEK64 fseek
#        define MZ_FILE_STAT_STRUCT stat
#        define MZ_FILE_STAT stat
#        define MZ_FFLUSH fflush
#        define MZ_FREOPEN(f, m, s) freopen(f, m, s)
#        define MZ_DELETE_FILE remove
#      elif defined(__GNUC__) && _LARGEFILE64_SOURCE
#        ifndef MINIZ_NO_TIME
#          include <utime.h>
#        endif
#        define MZ_FILE FILE
#        define MZ_FOPEN(f, m) fopen64(f, m)
#        define MZ_FCLOSE fclose
#        define MZ_FREAD fread
#        define MZ_FWRITE fwrite
#        define MZ_FTELL64 ftello64
#        define MZ_FSEEK64 fseeko64
#        define MZ_FILE_STAT_STRUCT stat64
#        define MZ_FILE_STAT stat64
#        define MZ_FFLUSH fflush
#        define MZ_FREOPEN(p, m, s) freopen64(p, m, s)
#        define MZ_DELETE_FILE remove
#      else
#        ifndef MINIZ_NO_TIME
#          include <utime.h>
#        endif
#        define MZ_FILE FILE
#        define MZ_FOPEN(f, m) fopen(f, m)
#        define MZ_FCLOSE fclose
#        define MZ_FREAD fread
#        define MZ_FWRITE fwrite
#        define MZ_FTELL64 ftello
#        define MZ_FSEEK64 fseeko
#        define MZ_FILE_STAT_STRUCT stat
#        define MZ_FILE_STAT stat
#        define MZ_FFLUSH fflush
#        define MZ_FREOPEN(f, m, s) freopen(f, m, s)
#        define MZ_DELETE_FILE remove
#      endif // #ifdef _MSC_VER
#    endif // #ifdef MINIZ_NO_STDIO

#    define MZ_TOLOWER(c) \
      ((((c) >= 'A') && ((c) <= 'Z')) ? ((c) - 'A' + 'a') : (c))

  // Various ZIP archive enums. To completely avoid cross platform compiler
  // alignment and platform endian issues, miniz.c doesn't use structs for any
  // of this stuff.
  enum
  {
    // ZIP archive identifiers and record sizes
    MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG = 0x06054b50,
    MZ_ZIP_CENTRAL_DIR_HEADER_SIG = 0x02014b50,
    MZ_ZIP_LOCAL_DIR_HEADER_SIG = 0x04034b50,
    MZ_ZIP_LOCAL_DIR_HEADER_SIZE = 30,
    MZ_ZIP_CENTRAL_DIR_HEADER_SIZE = 46,
    MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE = 22,
    // Central directory header record offsets
    MZ_ZIP_CDH_SIG_OFS = 0,
    MZ_ZIP_CDH_VERSION_MADE_BY_OFS = 4,
    MZ_ZIP_CDH_VERSION_NEEDED_OFS = 6,
    MZ_ZIP_CDH_BIT_FLAG_OFS = 8,
    MZ_ZIP_CDH_METHOD_OFS = 10,
    MZ_ZIP_CDH_FILE_TIME_OFS = 12,
    MZ_ZIP_CDH_FILE_DATE_OFS = 14,
    MZ_ZIP_CDH_CRC32_OFS = 16,
    MZ_ZIP_CDH_COMPRESSED_SIZE_OFS = 20,
    MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS = 24,
    MZ_ZIP_CDH_FILENAME_LEN_OFS = 28,
    MZ_ZIP_CDH_EXTRA_LEN_OFS = 30,
    MZ_ZIP_CDH_COMMENT_LEN_OFS = 32,
    MZ_ZIP_CDH_DISK_START_OFS = 34,
    MZ_ZIP_CDH_INTERNAL_ATTR_OFS = 36,
    MZ_ZIP_CDH_EXTERNAL_ATTR_OFS = 38,
    MZ_ZIP_CDH_LOCAL_HEADER_OFS = 42,
    // Local directory header offsets
    MZ_ZIP_LDH_SIG_OFS = 0,
    MZ_ZIP_LDH_VERSION_NEEDED_OFS = 4,
    MZ_ZIP_LDH_BIT_FLAG_OFS = 6,
    MZ_ZIP_LDH_METHOD_OFS = 8,
    MZ_ZIP_LDH_FILE_TIME_OFS = 10,
    MZ_ZIP_LDH_FILE_DATE_OFS = 12,
    MZ_ZIP_LDH_CRC32_OFS = 14,
    MZ_ZIP_LDH_COMPRESSED_SIZE_OFS = 18,
    MZ_ZIP_LDH_DECOMPRESSED_SIZE_OFS = 22,
    MZ_ZIP_LDH_FILENAME_LEN_OFS = 26,
    MZ_ZIP_LDH_EXTRA_LEN_OFS = 28,
    // End of central directory offsets
    MZ_ZIP_ECDH_SIG_OFS = 0,
    MZ_ZIP_ECDH_NUM_THIS_DISK_OFS = 4,
    MZ_ZIP_ECDH_NUM_DISK_CDIR_OFS = 6,
    MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS = 8,
    MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS = 10,
    MZ_ZIP_ECDH_CDIR_SIZE_OFS = 12,
    MZ_ZIP_ECDH_CDIR_OFS_OFS = 16,
    MZ_ZIP_ECDH_COMMENT_SIZE_OFS = 20,
  };

  typedef struct
  {
    void* m_p;
    size_t m_size, m_capacity;
    mz_uint m_element_size;
  } mz_zip_array;

  struct mz_zip_internal_state_tag
  {
    mz_zip_array m_central_dir;
    mz_zip_array m_central_dir_offsets;
    mz_zip_array m_sorted_central_dir_offsets;
    MZ_FILE* m_pFile;
    void* m_pMem;
    size_t m_mem_size;
    size_t m_mem_capacity;
  };

#    define MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(array_ptr, element_size) \
      (array_ptr)->m_element_size = element_size
#    define MZ_ZIP_ARRAY_ELEMENT(array_ptr, element_type, index) \
      ((element_type*)((array_ptr)->m_p))[index]

  static MZ_FORCEINLINE void
  mz_zip_array_clear(mz_zip_archive* pZip, mz_zip_array* pArray)
  {
    pZip->m_pFree(pZip->m_pAlloc_opaque, pArray->m_p);
    memset(pArray, 0, sizeof(mz_zip_array));
  }

  static mz_bool mz_zip_array_ensure_capacity(
    mz_zip_archive* pZip,
    mz_zip_array* pArray,
    size_t min_new_capacity,
    mz_uint growing)
  {
    void* pNew_p;
    size_t new_capacity = min_new_capacity;
    MZ_ASSERT(pArray->m_element_size);
    if (pArray->m_capacity >= min_new_capacity)
      return MZ_TRUE;
    if (growing)
    {
      new_capacity = MZ_MAX(1, pArray->m_capacity);
      while (new_capacity < min_new_capacity)
        new_capacity *= 2;
    }
    if (
      NULL ==
      (pNew_p = pZip->m_pRealloc(
         pZip->m_pAlloc_opaque,
         pArray->m_p,
         pArray->m_element_size,
         new_capacity)))
      return MZ_FALSE;
    pArray->m_p = pNew_p;
    pArray->m_capacity = new_capacity;
    return MZ_TRUE;
  }

  static MZ_FORCEINLINE mz_bool mz_zip_array_reserve(
    mz_zip_archive* pZip,
    mz_zip_array* pArray,
    size_t new_capacity,
    mz_uint growing)
  {
    if (new_capacity > pArray->m_capacity)
    {
      if (!mz_zip_array_ensure_capacity(pZip, pArray, new_capacity, growing))
        return MZ_FALSE;
    }
    return MZ_TRUE;
  }

  static MZ_FORCEINLINE mz_bool mz_zip_array_resize(
    mz_zip_archive* pZip,
    mz_zip_array* pArray,
    size_t new_size,
    mz_uint growing)
  {
    if (new_size > pArray->m_capacity)
    {
      if (!mz_zip_array_ensure_capacity(pZip, pArray, new_size, growing))
        return MZ_FALSE;
    }
    pArray->m_size = new_size;
    return MZ_TRUE;
  }

  static MZ_FORCEINLINE mz_bool
  mz_zip_array_ensure_room(mz_zip_archive* pZip, mz_zip_array* pArray, size_t n)
  {
    return mz_zip_array_reserve(pZip, pArray, pArray->m_size + n, MZ_TRUE);
  }

  static MZ_FORCEINLINE mz_bool mz_zip_array_push_back(
    mz_zip_archive* pZip, mz_zip_array* pArray, const void* pElements, size_t n)
  {
    size_t orig_size = pArray->m_size;
    if (!mz_zip_array_resize(pZip, pArray, orig_size + n, MZ_TRUE))
      return MZ_FALSE;
    memcpy(
      (mz_uint8*)pArray->m_p + orig_size * pArray->m_element_size,
      pElements,
      n * pArray->m_element_size);
    return MZ_TRUE;
  }

#    ifndef MINIZ_NO_TIME
  static time_t mz_zip_dos_to_time_t(int dos_time, int dos_date)
  {
    struct tm tm;
    memset(&tm, 0, sizeof(tm));
    tm.tm_isdst = -1;
    tm.tm_year = ((dos_date >> 9) & 127) + 1980 - 1900;
    tm.tm_mon = ((dos_date >> 5) & 15) - 1;
    tm.tm_mday = dos_date & 31;
    tm.tm_hour = (dos_time >> 11) & 31;
    tm.tm_min = (dos_time >> 5) & 63;
    tm.tm_sec = (dos_time << 1) & 62;
    return mktime(&tm);
  }

  static void mz_zip_time_to_dos_time(
    time_t time, mz_uint16* pDOS_time, mz_uint16* pDOS_date)
  {
#      ifdef _MSC_VER
    struct tm tm_struct;
    struct tm* tm = &tm_struct;
    errno_t err = localtime_s(tm, &time);
    if (err)
    {
      *pDOS_date = 0;
      *pDOS_time = 0;
      return;
    }
#      else
    struct tm* tm = localtime(&time);
#      endif
    *pDOS_time = (mz_uint16)(
      ((tm->tm_hour) << 11) + ((tm->tm_min) << 5) + ((tm->tm_sec) >> 1));
    *pDOS_date = (mz_uint16)(
      ((tm->tm_year + 1900 - 1980) << 9) + ((tm->tm_mon + 1) << 5) +
      tm->tm_mday);
  }
#    endif

#    ifndef MINIZ_NO_STDIO
  static mz_bool mz_zip_get_file_modified_time(
    const char* pFilename, mz_uint16* pDOS_time, mz_uint16* pDOS_date)
  {
#      ifdef MINIZ_NO_TIME
    (void)pFilename;
    *pDOS_date = *pDOS_time = 0;
#      else
    struct MZ_FILE_STAT_STRUCT file_stat;
    // On Linux with x86 glibc, this call will fail on large files (>=
    // 0x80000000 bytes) unless you compiled with _LARGEFILE64_SOURCE. Argh.
    if (MZ_FILE_STAT(pFilename, &file_stat) != 0)
      return MZ_FALSE;
    mz_zip_time_to_dos_time(file_stat.st_mtime, pDOS_time, pDOS_date);
#      endif // #ifdef MINIZ_NO_TIME
    return MZ_TRUE;
  }

#      ifndef MINIZ_NO_TIME
  static mz_bool mz_zip_set_file_times(
    const char* pFilename, time_t access_time, time_t modified_time)
  {
    struct utimbuf t;
    t.actime = access_time;
    t.modtime = modified_time;
    return !utime(pFilename, &t);
  }
#      endif // #ifndef MINIZ_NO_TIME
#    endif // #ifndef MINIZ_NO_STDIO

  static mz_bool
  mz_zip_reader_init_internal(mz_zip_archive* pZip, mz_uint32 flags)
  {
    (void)flags;
    if (
      (!pZip) || (pZip->m_pState) || (pZip->m_zip_mode != MZ_ZIP_MODE_INVALID))
      return MZ_FALSE;

    if (!pZip->m_pAlloc)
      pZip->m_pAlloc = def_alloc_func;
    if (!pZip->m_pFree)
      pZip->m_pFree = def_free_func;
    if (!pZip->m_pRealloc)
      pZip->m_pRealloc = def_realloc_func;

    pZip->m_zip_mode = MZ_ZIP_MODE_READING;
    pZip->m_archive_size = 0;
    pZip->m_central_directory_file_ofs = 0;
    pZip->m_total_files = 0;

    if (
      NULL ==
      (pZip->m_pState = (mz_zip_internal_state*)pZip->m_pAlloc(
         pZip->m_pAlloc_opaque, 1, sizeof(mz_zip_internal_state))))
      return MZ_FALSE;
    memset(pZip->m_pState, 0, sizeof(mz_zip_internal_state));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_central_dir, sizeof(mz_uint8));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_central_dir_offsets, sizeof(mz_uint32));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_sorted_central_dir_offsets, sizeof(mz_uint32));
    return MZ_TRUE;
  }

  static MZ_FORCEINLINE mz_bool mz_zip_reader_filename_less(
    const mz_zip_array* pCentral_dir_array,
    const mz_zip_array* pCentral_dir_offsets,
    mz_uint l_index,
    mz_uint r_index)
  {
    const mz_uint8 *pL = &MZ_ZIP_ARRAY_ELEMENT(
                     pCentral_dir_array,
                     mz_uint8,
                     MZ_ZIP_ARRAY_ELEMENT(
                       pCentral_dir_offsets, mz_uint32, l_index)),
                   *pE;
    const mz_uint8* pR = &MZ_ZIP_ARRAY_ELEMENT(
      pCentral_dir_array,
      mz_uint8,
      MZ_ZIP_ARRAY_ELEMENT(pCentral_dir_offsets, mz_uint32, r_index));
    mz_uint l_len = MZ_READ_LE16(pL + MZ_ZIP_CDH_FILENAME_LEN_OFS),
            r_len = MZ_READ_LE16(pR + MZ_ZIP_CDH_FILENAME_LEN_OFS);
    mz_uint8 l = 0, r = 0;
    pL += MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
    pR += MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
    pE = pL + MZ_MIN(l_len, r_len);
    while (pL < pE)
    {
      if ((l = MZ_TOLOWER(*pL)) != (r = MZ_TOLOWER(*pR)))
        break;
      pL++;
      pR++;
    }
    return (pL == pE) ? (l_len < r_len) : (l < r);
  }

#    define MZ_SWAP_UINT32(a, b) \
      do \
      { \
        mz_uint32 t = a; \
        a = b; \
        b = t; \
      } \
      MZ_MACRO_END

  // Heap sort of lowercased filenames, used to help accelerate plain central
  // directory searches by mz_zip_reader_locate_file(). (Could also use qsort(),
  // but it could allocate memory.)
  static void
  mz_zip_reader_sort_central_dir_offsets_by_filename(mz_zip_archive* pZip)
  {
    mz_zip_internal_state* pState = pZip->m_pState;
    const mz_zip_array* pCentral_dir_offsets = &pState->m_central_dir_offsets;
    const mz_zip_array* pCentral_dir = &pState->m_central_dir;
    mz_uint32* pIndices = &MZ_ZIP_ARRAY_ELEMENT(
      &pState->m_sorted_central_dir_offsets, mz_uint32, 0);
    const int size = pZip->m_total_files;
    int start = (size - 2) >> 1, end;
    while (start >= 0)
    {
      int child, root = start;
      for (;;)
      {
        if ((child = (root << 1) + 1) >= size)
          break;
        child +=
          (((child + 1) < size) &&
           (mz_zip_reader_filename_less(
             pCentral_dir,
             pCentral_dir_offsets,
             pIndices[child],
             pIndices[child + 1])));
        if (!mz_zip_reader_filename_less(
              pCentral_dir,
              pCentral_dir_offsets,
              pIndices[root],
              pIndices[child]))
          break;
        MZ_SWAP_UINT32(pIndices[root], pIndices[child]);
        root = child;
      }
      start--;
    }

    end = size - 1;
    while (end > 0)
    {
      int child, root = 0;
      MZ_SWAP_UINT32(pIndices[end], pIndices[0]);
      for (;;)
      {
        if ((child = (root << 1) + 1) >= end)
          break;
        child +=
          (((child + 1) < end) &&
           mz_zip_reader_filename_less(
             pCentral_dir,
             pCentral_dir_offsets,
             pIndices[child],
             pIndices[child + 1]));
        if (!mz_zip_reader_filename_less(
              pCentral_dir,
              pCentral_dir_offsets,
              pIndices[root],
              pIndices[child]))
          break;
        MZ_SWAP_UINT32(pIndices[root], pIndices[child]);
        root = child;
      }
      end--;
    }
  }

  static mz_bool
  mz_zip_reader_read_central_dir(mz_zip_archive* pZip, mz_uint32 flags)
  {
    mz_uint cdir_size, num_this_disk, cdir_disk_index;
    mz_uint64 cdir_ofs;
    mz_int64 cur_file_ofs;
    const mz_uint8* p;
    mz_uint32 buf_u32[4096 / sizeof(mz_uint32)];
    mz_uint8* pBuf = (mz_uint8*)buf_u32;
    mz_bool sort_central_dir =
      ((flags & MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY) == 0);
    // Basic sanity checks - reject files which are too small, and check the
    // first 4 bytes of the file to make sure a local header is there.
    if (pZip->m_archive_size < MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    // Find the end of central directory record by scanning the file from the
    // end towards the beginning.
    cur_file_ofs =
      MZ_MAX((mz_int64)pZip->m_archive_size - (mz_int64)sizeof(buf_u32), 0);
    for (;;)
    {
      int i,
        n = (int)MZ_MIN(sizeof(buf_u32), pZip->m_archive_size - cur_file_ofs);
      if (
        pZip->m_pRead(pZip->m_pIO_opaque, cur_file_ofs, pBuf, n) != (mz_uint)n)
        return MZ_FALSE;
      for (i = n - 4; i >= 0; --i)
        if (MZ_READ_LE32(pBuf + i) == MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG)
          break;
      if (i >= 0)
      {
        cur_file_ofs += i;
        break;
      }
      if (
        (!cur_file_ofs) ||
        ((pZip->m_archive_size - cur_file_ofs) >=
         (0xFFFF + MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)))
        return MZ_FALSE;
      cur_file_ofs = MZ_MAX(cur_file_ofs - (sizeof(buf_u32) - 3), 0);
    }
    // Read and verify the end of central directory record.
    if (
      pZip->m_pRead(
        pZip->m_pIO_opaque,
        cur_file_ofs,
        pBuf,
        MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE) !=
      MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    if (
      (MZ_READ_LE32(pBuf + MZ_ZIP_ECDH_SIG_OFS) !=
       MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG) ||
      ((pZip->m_total_files =
          MZ_READ_LE16(pBuf + MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS)) !=
       MZ_READ_LE16(pBuf + MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS)))
      return MZ_FALSE;

    num_this_disk = MZ_READ_LE16(pBuf + MZ_ZIP_ECDH_NUM_THIS_DISK_OFS);
    cdir_disk_index = MZ_READ_LE16(pBuf + MZ_ZIP_ECDH_NUM_DISK_CDIR_OFS);
    if (
      ((num_this_disk | cdir_disk_index) != 0) &&
      ((num_this_disk != 1) || (cdir_disk_index != 1)))
      return MZ_FALSE;

    if (
      (cdir_size = MZ_READ_LE32(pBuf + MZ_ZIP_ECDH_CDIR_SIZE_OFS)) <
      pZip->m_total_files * MZ_ZIP_CENTRAL_DIR_HEADER_SIZE)
      return MZ_FALSE;

    cdir_ofs = MZ_READ_LE32(pBuf + MZ_ZIP_ECDH_CDIR_OFS_OFS);
    if ((cdir_ofs + (mz_uint64)cdir_size) > pZip->m_archive_size)
      return MZ_FALSE;

    pZip->m_central_directory_file_ofs = cdir_ofs;

    if (pZip->m_total_files)
    {
      mz_uint i, n;

      // Read the entire central directory into a heap block, and allocate
      // another heap block to hold the unsorted central dir file record
      // offsets, and another to hold the sorted indices.
      if (
        (!mz_zip_array_resize(
          pZip, &pZip->m_pState->m_central_dir, cdir_size, MZ_FALSE)) ||
        (!mz_zip_array_resize(
          pZip,
          &pZip->m_pState->m_central_dir_offsets,
          pZip->m_total_files,
          MZ_FALSE)))
        return MZ_FALSE;

      if (sort_central_dir)
      {
        if (!mz_zip_array_resize(
              pZip,
              &pZip->m_pState->m_sorted_central_dir_offsets,
              pZip->m_total_files,
              MZ_FALSE))
          return MZ_FALSE;
      }

      if (
        pZip->m_pRead(
          pZip->m_pIO_opaque,
          cdir_ofs,
          pZip->m_pState->m_central_dir.m_p,
          cdir_size) != cdir_size)
        return MZ_FALSE;

      // Now create an index into the central directory file records, do some
      // basic sanity checking on each record, and check for zip64 entries
      // (which are not yet supported).
      p = (const mz_uint8*)pZip->m_pState->m_central_dir.m_p;
      for (n = cdir_size, i = 0; i < pZip->m_total_files; ++i)
      {
        mz_uint total_header_size, comp_size, decomp_size, disk_index;
        if (
          (n < MZ_ZIP_CENTRAL_DIR_HEADER_SIZE) ||
          (MZ_READ_LE32(p) != MZ_ZIP_CENTRAL_DIR_HEADER_SIG))
          return MZ_FALSE;
        MZ_ZIP_ARRAY_ELEMENT(
          &pZip->m_pState->m_central_dir_offsets, mz_uint32, i) =
          (mz_uint32)(p - (const mz_uint8*)pZip->m_pState->m_central_dir.m_p);
        if (sort_central_dir)
          MZ_ZIP_ARRAY_ELEMENT(
            &pZip->m_pState->m_sorted_central_dir_offsets, mz_uint32, i) = i;
        comp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS);
        decomp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS);
        if (
          ((!MZ_READ_LE32(p + MZ_ZIP_CDH_METHOD_OFS)) &&
           (decomp_size != comp_size)) ||
          (decomp_size && !comp_size) || (decomp_size == 0xFFFFFFFF) ||
          (comp_size == 0xFFFFFFFF))
          return MZ_FALSE;
        disk_index = MZ_READ_LE16(p + MZ_ZIP_CDH_DISK_START_OFS);
        if ((disk_index != num_this_disk) && (disk_index != 1))
          return MZ_FALSE;
        if (
          ((mz_uint64)MZ_READ_LE32(p + MZ_ZIP_CDH_LOCAL_HEADER_OFS) +
           MZ_ZIP_LOCAL_DIR_HEADER_SIZE + comp_size) > pZip->m_archive_size)
          return MZ_FALSE;
        if (
          (total_header_size = MZ_ZIP_CENTRAL_DIR_HEADER_SIZE +
             MZ_READ_LE16(p + MZ_ZIP_CDH_FILENAME_LEN_OFS) +
             MZ_READ_LE16(p + MZ_ZIP_CDH_EXTRA_LEN_OFS) +
             MZ_READ_LE16(p + MZ_ZIP_CDH_COMMENT_LEN_OFS)) > n)
          return MZ_FALSE;
        n -= total_header_size;
        p += total_header_size;
      }
    }

    if (sort_central_dir)
      mz_zip_reader_sort_central_dir_offsets_by_filename(pZip);

    return MZ_TRUE;
  }

  mz_bool
  mz_zip_reader_init(mz_zip_archive* pZip, mz_uint64 size, mz_uint32 flags)
  {
    if ((!pZip) || (!pZip->m_pRead))
      return MZ_FALSE;
    if (!mz_zip_reader_init_internal(pZip, flags))
      return MZ_FALSE;
    pZip->m_archive_size = size;
    if (!mz_zip_reader_read_central_dir(pZip, flags))
    {
      mz_zip_reader_end(pZip);
      return MZ_FALSE;
    }
    return MZ_TRUE;
  }

  static size_t
  mz_zip_mem_read_func(void* pOpaque, mz_uint64 file_ofs, void* pBuf, size_t n)
  {
    mz_zip_archive* pZip = (mz_zip_archive*)pOpaque;
    size_t s = (file_ofs >= pZip->m_archive_size) ?
      0 :
      (size_t)MZ_MIN(pZip->m_archive_size - file_ofs, n);
    memcpy(pBuf, (const mz_uint8*)pZip->m_pState->m_pMem + file_ofs, s);
    return s;
  }

  mz_bool mz_zip_reader_init_mem(
    mz_zip_archive* pZip, const void* pMem, size_t size, mz_uint32 flags)
  {
    if (!mz_zip_reader_init_internal(pZip, flags))
      return MZ_FALSE;
    pZip->m_archive_size = size;
    pZip->m_pRead = mz_zip_mem_read_func;
    pZip->m_pIO_opaque = pZip;
#    ifdef __cplusplus
    pZip->m_pState->m_pMem = const_cast<void*>(pMem);
#    else
    pZip->m_pState->m_pMem = (void*)pMem;
#    endif
    pZip->m_pState->m_mem_size = size;
    if (!mz_zip_reader_read_central_dir(pZip, flags))
    {
      mz_zip_reader_end(pZip);
      return MZ_FALSE;
    }
    return MZ_TRUE;
  }

#    ifndef MINIZ_NO_STDIO
  static size_t
  mz_zip_file_read_func(void* pOpaque, mz_uint64 file_ofs, void* pBuf, size_t n)
  {
    mz_zip_archive* pZip = (mz_zip_archive*)pOpaque;
    mz_int64 cur_ofs = MZ_FTELL64(pZip->m_pState->m_pFile);
    if (
      ((mz_int64)file_ofs < 0) ||
      (((cur_ofs != (mz_int64)file_ofs)) &&
       (MZ_FSEEK64(pZip->m_pState->m_pFile, (mz_int64)file_ofs, SEEK_SET))))
      return 0;
    return MZ_FREAD(pBuf, 1, n, pZip->m_pState->m_pFile);
  }

  mz_bool mz_zip_reader_init_file(
    mz_zip_archive* pZip, const char* pFilename, mz_uint32 flags)
  {
    mz_uint64 file_size;
    MZ_FILE* pFile = MZ_FOPEN(pFilename, "rb");
    if (!pFile)
      return MZ_FALSE;
    if (MZ_FSEEK64(pFile, 0, SEEK_END))
    {
      MZ_FCLOSE(pFile);
      return MZ_FALSE;
    }
    file_size = MZ_FTELL64(pFile);
    if (!mz_zip_reader_init_internal(pZip, flags))
    {
      MZ_FCLOSE(pFile);
      return MZ_FALSE;
    }
    pZip->m_pRead = mz_zip_file_read_func;
    pZip->m_pIO_opaque = pZip;
    pZip->m_pState->m_pFile = pFile;
    pZip->m_archive_size = file_size;
    if (!mz_zip_reader_read_central_dir(pZip, flags))
    {
      mz_zip_reader_end(pZip);
      return MZ_FALSE;
    }
    return MZ_TRUE;
  }
#    endif // #ifndef MINIZ_NO_STDIO

  mz_uint mz_zip_reader_get_num_files(mz_zip_archive* pZip)
  {
    return pZip ? pZip->m_total_files : 0;
  }

  static MZ_FORCEINLINE const mz_uint8*
  mz_zip_reader_get_cdh(mz_zip_archive* pZip, mz_uint file_index)
  {
    if (
      (!pZip) || (!pZip->m_pState) || (file_index >= pZip->m_total_files) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_READING))
      return NULL;
    return &MZ_ZIP_ARRAY_ELEMENT(
      &pZip->m_pState->m_central_dir,
      mz_uint8,
      MZ_ZIP_ARRAY_ELEMENT(
        &pZip->m_pState->m_central_dir_offsets, mz_uint32, file_index));
  }

  mz_bool
  mz_zip_reader_is_file_encrypted(mz_zip_archive* pZip, mz_uint file_index)
  {
    mz_uint m_bit_flag;
    const mz_uint8* p = mz_zip_reader_get_cdh(pZip, file_index);
    if (!p)
      return MZ_FALSE;
    m_bit_flag = MZ_READ_LE16(p + MZ_ZIP_CDH_BIT_FLAG_OFS);
    return (m_bit_flag & 1);
  }

  mz_bool
  mz_zip_reader_is_file_a_directory(mz_zip_archive* pZip, mz_uint file_index)
  {
    mz_uint filename_len, external_attr;
    const mz_uint8* p = mz_zip_reader_get_cdh(pZip, file_index);
    if (!p)
      return MZ_FALSE;

    // First see if the filename ends with a '/' character.
    filename_len = MZ_READ_LE16(p + MZ_ZIP_CDH_FILENAME_LEN_OFS);
    if (filename_len)
    {
      if (*(p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + filename_len - 1) == '/')
        return MZ_TRUE;
    }

    // Bugfix: This code was also checking if the internal attribute was
    // non-zero, which wasn't correct. Most/all zip writers (hopefully) set DOS
    // file/directory attributes in the low 16-bits, so check for the DOS
    // directory flag and ignore the source OS ID in the created by field.
    // FIXME: Remove this check? Is it necessary - we already check the
    // filename.
    external_attr = MZ_READ_LE32(p + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS);
    if ((external_attr & 0x10) != 0)
      return MZ_TRUE;

    return MZ_FALSE;
  }

  mz_bool mz_zip_reader_file_stat(
    mz_zip_archive* pZip, mz_uint file_index, mz_zip_archive_file_stat* pStat)
  {
    mz_uint n;
    const mz_uint8* p = mz_zip_reader_get_cdh(pZip, file_index);
    if ((!p) || (!pStat))
      return MZ_FALSE;

    // Unpack the central directory record.
    pStat->m_file_index = file_index;
    pStat->m_central_dir_ofs = MZ_ZIP_ARRAY_ELEMENT(
      &pZip->m_pState->m_central_dir_offsets, mz_uint32, file_index);
    pStat->m_version_made_by = MZ_READ_LE16(p + MZ_ZIP_CDH_VERSION_MADE_BY_OFS);
    pStat->m_version_needed = MZ_READ_LE16(p + MZ_ZIP_CDH_VERSION_NEEDED_OFS);
    pStat->m_bit_flag = MZ_READ_LE16(p + MZ_ZIP_CDH_BIT_FLAG_OFS);
    pStat->m_method = MZ_READ_LE16(p + MZ_ZIP_CDH_METHOD_OFS);
#    ifndef MINIZ_NO_TIME
    pStat->m_time = mz_zip_dos_to_time_t(
      MZ_READ_LE16(p + MZ_ZIP_CDH_FILE_TIME_OFS),
      MZ_READ_LE16(p + MZ_ZIP_CDH_FILE_DATE_OFS));
#    endif
    pStat->m_crc32 = MZ_READ_LE32(p + MZ_ZIP_CDH_CRC32_OFS);
    pStat->m_comp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS);
    pStat->m_uncomp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS);
    pStat->m_internal_attr = MZ_READ_LE16(p + MZ_ZIP_CDH_INTERNAL_ATTR_OFS);
    pStat->m_external_attr = MZ_READ_LE32(p + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS);
    pStat->m_local_header_ofs = MZ_READ_LE32(p + MZ_ZIP_CDH_LOCAL_HEADER_OFS);

    // Copy as much of the filename and comment as possible.
    n = MZ_READ_LE16(p + MZ_ZIP_CDH_FILENAME_LEN_OFS);
    n = MZ_MIN(n, MZ_ZIP_MAX_ARCHIVE_FILENAME_SIZE - 1);
    memcpy(pStat->m_filename, p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE, n);
    pStat->m_filename[n] = '\0';

    n = MZ_READ_LE16(p + MZ_ZIP_CDH_COMMENT_LEN_OFS);
    n = MZ_MIN(n, MZ_ZIP_MAX_ARCHIVE_FILE_COMMENT_SIZE - 1);
    pStat->m_comment_size = n;
    memcpy(
      pStat->m_comment,
      p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE +
        MZ_READ_LE16(p + MZ_ZIP_CDH_FILENAME_LEN_OFS) +
        MZ_READ_LE16(p + MZ_ZIP_CDH_EXTRA_LEN_OFS),
      n);
    pStat->m_comment[n] = '\0';

    return MZ_TRUE;
  }

  mz_uint mz_zip_reader_get_filename(
    mz_zip_archive* pZip,
    mz_uint file_index,
    char* pFilename,
    mz_uint filename_buf_size)
  {
    mz_uint n;
    const mz_uint8* p = mz_zip_reader_get_cdh(pZip, file_index);
    if (!p)
    {
      if (filename_buf_size)
        pFilename[0] = '\0';
      return 0;
    }
    n = MZ_READ_LE16(p + MZ_ZIP_CDH_FILENAME_LEN_OFS);
    if (filename_buf_size)
    {
      n = MZ_MIN(n, filename_buf_size - 1);
      memcpy(pFilename, p + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE, n);
      pFilename[n] = '\0';
    }
    return n + 1;
  }

  static MZ_FORCEINLINE mz_bool mz_zip_reader_string_equal(
    const char* pA, const char* pB, mz_uint len, mz_uint flags)
  {
    mz_uint i;
    if (flags & MZ_ZIP_FLAG_CASE_SENSITIVE)
      return 0 == memcmp(pA, pB, len);
    for (i = 0; i < len; ++i)
      if (MZ_TOLOWER(pA[i]) != MZ_TOLOWER(pB[i]))
        return MZ_FALSE;
    return MZ_TRUE;
  }

  static MZ_FORCEINLINE int mz_zip_reader_filename_compare(
    const mz_zip_array* pCentral_dir_array,
    const mz_zip_array* pCentral_dir_offsets,
    mz_uint l_index,
    const char* pR,
    mz_uint r_len)
  {
    const mz_uint8 *pL = &MZ_ZIP_ARRAY_ELEMENT(
                     pCentral_dir_array,
                     mz_uint8,
                     MZ_ZIP_ARRAY_ELEMENT(
                       pCentral_dir_offsets, mz_uint32, l_index)),
                   *pE;
    mz_uint l_len = MZ_READ_LE16(pL + MZ_ZIP_CDH_FILENAME_LEN_OFS);
    mz_uint8 l = 0, r = 0;
    pL += MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
    pE = pL + MZ_MIN(l_len, r_len);
    while (pL < pE)
    {
      if ((l = MZ_TOLOWER(*pL)) != (r = MZ_TOLOWER(*pR)))
        break;
      pL++;
      pR++;
    }
    return (pL == pE) ? (int)(l_len - r_len) : (l - r);
  }

  static int mz_zip_reader_locate_file_binary_search(
    mz_zip_archive* pZip, const char* pFilename)
  {
    mz_zip_internal_state* pState = pZip->m_pState;
    const mz_zip_array* pCentral_dir_offsets = &pState->m_central_dir_offsets;
    const mz_zip_array* pCentral_dir = &pState->m_central_dir;
    mz_uint32* pIndices = &MZ_ZIP_ARRAY_ELEMENT(
      &pState->m_sorted_central_dir_offsets, mz_uint32, 0);
    const int size = pZip->m_total_files;
    const mz_uint filename_len = (mz_uint)strlen(pFilename);
    int l = 0, h = size - 1;
    while (l <= h)
    {
      int m = (l + h) >> 1, file_index = pIndices[m],
          comp = mz_zip_reader_filename_compare(
            pCentral_dir,
            pCentral_dir_offsets,
            file_index,
            pFilename,
            filename_len);
      if (!comp)
        return file_index;
      else if (comp < 0)
        l = m + 1;
      else
        h = m - 1;
    }
    return -1;
  }

  int mz_zip_reader_locate_file(
    mz_zip_archive* pZip,
    const char* pName,
    const char* pComment,
    mz_uint flags)
  {
    mz_uint file_index;
    size_t name_len, comment_len;
    if (
      (!pZip) || (!pZip->m_pState) || (!pName) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_READING))
      return -1;
    if (
      ((flags & (MZ_ZIP_FLAG_IGNORE_PATH | MZ_ZIP_FLAG_CASE_SENSITIVE)) == 0) &&
      (!pComment) && (pZip->m_pState->m_sorted_central_dir_offsets.m_size))
      return mz_zip_reader_locate_file_binary_search(pZip, pName);
    name_len = strlen(pName);
    if (name_len > 0xFFFF)
      return -1;
    comment_len = pComment ? strlen(pComment) : 0;
    if (comment_len > 0xFFFF)
      return -1;
    for (file_index = 0; file_index < pZip->m_total_files; file_index++)
    {
      const mz_uint8* pHeader = &MZ_ZIP_ARRAY_ELEMENT(
        &pZip->m_pState->m_central_dir,
        mz_uint8,
        MZ_ZIP_ARRAY_ELEMENT(
          &pZip->m_pState->m_central_dir_offsets, mz_uint32, file_index));
      mz_uint filename_len =
        MZ_READ_LE16(pHeader + MZ_ZIP_CDH_FILENAME_LEN_OFS);
      const char* pFilename =
        (const char*)pHeader + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE;
      if (filename_len < name_len)
        continue;
      if (comment_len)
      {
        mz_uint file_extra_len =
                  MZ_READ_LE16(pHeader + MZ_ZIP_CDH_EXTRA_LEN_OFS),
                file_comment_len =
                  MZ_READ_LE16(pHeader + MZ_ZIP_CDH_COMMENT_LEN_OFS);
        const char* pFile_comment = pFilename + filename_len + file_extra_len;
        if (
          (file_comment_len != comment_len) ||
          (!mz_zip_reader_string_equal(
            pComment, pFile_comment, file_comment_len, flags)))
          continue;
      }
      if ((flags & MZ_ZIP_FLAG_IGNORE_PATH) && (filename_len))
      {
        int ofs = filename_len - 1;
        do
        {
          if (
            (pFilename[ofs] == '/') || (pFilename[ofs] == '\\') ||
            (pFilename[ofs] == ':'))
            break;
        } while (--ofs >= 0);
        ofs++;
        pFilename += ofs;
        filename_len -= ofs;
      }
      if (
        (filename_len == name_len) &&
        (mz_zip_reader_string_equal(pName, pFilename, filename_len, flags)))
        return file_index;
    }
    return -1;
  }

  mz_bool mz_zip_reader_extract_to_mem_no_alloc(
    mz_zip_archive* pZip,
    mz_uint file_index,
    void* pBuf,
    size_t buf_size,
    mz_uint flags,
    void* pUser_read_buf,
    size_t user_read_buf_size)
  {
    int status = TINFL_STATUS_DONE;
    mz_uint64 needed_size, cur_file_ofs, comp_remaining,
      out_buf_ofs = 0, read_buf_size, read_buf_ofs = 0, read_buf_avail;
    mz_zip_archive_file_stat file_stat;
    void* pRead_buf;
    mz_uint32 local_header_u32
      [(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(mz_uint32) - 1) /
       sizeof(mz_uint32)];
    mz_uint8* pLocal_header = (mz_uint8*)local_header_u32;
    tinfl_decompressor inflator;

    if ((buf_size) && (!pBuf))
      return MZ_FALSE;

    if (!mz_zip_reader_file_stat(pZip, file_index, &file_stat))
      return MZ_FALSE;

    // Empty file, or a directory (but not always a directory - I've seen odd
    // zips with directories that have compressed data which inflates to 0
    // bytes)
    if (!file_stat.m_comp_size)
      return MZ_TRUE;

    // Entry is a subdirectory (I've seen old zips with dir entries which have
    // compressed deflate data which inflates to 0 bytes, but these entries
    // claim to uncompress to 512 bytes in the headers). I'm torn how to handle
    // this case - should it fail instead?
    if (mz_zip_reader_is_file_a_directory(pZip, file_index))
      return MZ_TRUE;

    // Encryption and patch files are not supported.
    if (file_stat.m_bit_flag & (1 | 32))
      return MZ_FALSE;

    // This function only supports stored and deflate.
    if (
      (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (file_stat.m_method != 0) &&
      (file_stat.m_method != MZ_DEFLATED))
      return MZ_FALSE;

    // Ensure supplied output buffer is large enough.
    needed_size = (flags & MZ_ZIP_FLAG_COMPRESSED_DATA) ?
      file_stat.m_comp_size :
      file_stat.m_uncomp_size;
    if (buf_size < needed_size)
      return MZ_FALSE;

    // Read and parse the local directory entry.
    cur_file_ofs = file_stat.m_local_header_ofs;
    if (
      pZip->m_pRead(
        pZip->m_pIO_opaque,
        cur_file_ofs,
        pLocal_header,
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE) != MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    if (MZ_READ_LE32(pLocal_header) != MZ_ZIP_LOCAL_DIR_HEADER_SIG)
      return MZ_FALSE;

    cur_file_ofs += MZ_ZIP_LOCAL_DIR_HEADER_SIZE +
      MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS) +
      MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS);
    if ((cur_file_ofs + file_stat.m_comp_size) > pZip->m_archive_size)
      return MZ_FALSE;

    if ((flags & MZ_ZIP_FLAG_COMPRESSED_DATA) || (!file_stat.m_method))
    {
      // The file is stored or the caller has requested the compressed data.
      if (
        pZip->m_pRead(
          pZip->m_pIO_opaque, cur_file_ofs, pBuf, (size_t)needed_size) !=
        needed_size)
        return MZ_FALSE;
      return ((flags & MZ_ZIP_FLAG_COMPRESSED_DATA) != 0) ||
        (mz_crc32(
           MZ_CRC32_INIT,
           (const mz_uint8*)pBuf,
           (size_t)file_stat.m_uncomp_size) == file_stat.m_crc32);
    }

    // Decompress the file either directly from memory or from a file input
    // buffer.
    tinfl_init(&inflator);

    if (pZip->m_pState->m_pMem)
    {
      // Read directly from the archive in memory.
      pRead_buf = (mz_uint8*)pZip->m_pState->m_pMem + cur_file_ofs;
      read_buf_size = read_buf_avail = file_stat.m_comp_size;
      comp_remaining = 0;
    }
    else if (pUser_read_buf)
    {
      // Use a user provided read buffer.
      if (!user_read_buf_size)
        return MZ_FALSE;
      pRead_buf = (mz_uint8*)pUser_read_buf;
      read_buf_size = user_read_buf_size;
      read_buf_avail = 0;
      comp_remaining = file_stat.m_comp_size;
    }
    else
    {
      // Temporarily allocate a read buffer.
      read_buf_size = MZ_MIN(file_stat.m_comp_size, MZ_ZIP_MAX_IO_BUF_SIZE);
#    ifdef _MSC_VER
      if (
        ((0, sizeof(size_t) == sizeof(mz_uint32))) &&
        (read_buf_size > 0x7FFFFFFF))
#    else
      if (
        ((sizeof(size_t) == sizeof(mz_uint32))) && (read_buf_size > 0x7FFFFFFF))
#    endif
        return MZ_FALSE;
      if (
        NULL ==
        (pRead_buf =
           pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, (size_t)read_buf_size)))
        return MZ_FALSE;
      read_buf_avail = 0;
      comp_remaining = file_stat.m_comp_size;
    }

    do
    {
      size_t in_buf_size,
        out_buf_size = (size_t)(file_stat.m_uncomp_size - out_buf_ofs);
      if ((!read_buf_avail) && (!pZip->m_pState->m_pMem))
      {
        read_buf_avail = MZ_MIN(read_buf_size, comp_remaining);
        if (
          pZip->m_pRead(
            pZip->m_pIO_opaque,
            cur_file_ofs,
            pRead_buf,
            (size_t)read_buf_avail) != read_buf_avail)
        {
          status = TINFL_STATUS_FAILED;
          break;
        }
        cur_file_ofs += read_buf_avail;
        comp_remaining -= read_buf_avail;
        read_buf_ofs = 0;
      }
      in_buf_size = (size_t)read_buf_avail;
      status = tinfl_decompress(
        &inflator,
        (mz_uint8*)pRead_buf + read_buf_ofs,
        &in_buf_size,
        (mz_uint8*)pBuf,
        (mz_uint8*)pBuf + out_buf_ofs,
        &out_buf_size,
        TINFL_FLAG_USING_NON_WRAPPING_OUTPUT_BUF |
          (comp_remaining ? TINFL_FLAG_HAS_MORE_INPUT : 0));
      read_buf_avail -= in_buf_size;
      read_buf_ofs += in_buf_size;
      out_buf_ofs += out_buf_size;
    } while (status == TINFL_STATUS_NEEDS_MORE_INPUT);

    if (status == TINFL_STATUS_DONE)
    {
      // Make sure the entire file was decompressed, and check its CRC.
      if (
        (out_buf_ofs != file_stat.m_uncomp_size) ||
        (mz_crc32(
           MZ_CRC32_INIT,
           (const mz_uint8*)pBuf,
           (size_t)file_stat.m_uncomp_size) != file_stat.m_crc32))
        status = TINFL_STATUS_FAILED;
    }

    if ((!pZip->m_pState->m_pMem) && (!pUser_read_buf))
      pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);

    return status == TINFL_STATUS_DONE;
  }

  mz_bool mz_zip_reader_extract_file_to_mem_no_alloc(
    mz_zip_archive* pZip,
    const char* pFilename,
    void* pBuf,
    size_t buf_size,
    mz_uint flags,
    void* pUser_read_buf,
    size_t user_read_buf_size)
  {
    int file_index = mz_zip_reader_locate_file(pZip, pFilename, NULL, flags);
    if (file_index < 0)
      return MZ_FALSE;
    return mz_zip_reader_extract_to_mem_no_alloc(
      pZip,
      file_index,
      pBuf,
      buf_size,
      flags,
      pUser_read_buf,
      user_read_buf_size);
  }

  mz_bool mz_zip_reader_extract_to_mem(
    mz_zip_archive* pZip,
    mz_uint file_index,
    void* pBuf,
    size_t buf_size,
    mz_uint flags)
  {
    return mz_zip_reader_extract_to_mem_no_alloc(
      pZip, file_index, pBuf, buf_size, flags, NULL, 0);
  }

  mz_bool mz_zip_reader_extract_file_to_mem(
    mz_zip_archive* pZip,
    const char* pFilename,
    void* pBuf,
    size_t buf_size,
    mz_uint flags)
  {
    return mz_zip_reader_extract_file_to_mem_no_alloc(
      pZip, pFilename, pBuf, buf_size, flags, NULL, 0);
  }

  void* mz_zip_reader_extract_to_heap(
    mz_zip_archive* pZip, mz_uint file_index, size_t* pSize, mz_uint flags)
  {
    mz_uint64 comp_size, uncomp_size, alloc_size;
    const mz_uint8* p = mz_zip_reader_get_cdh(pZip, file_index);
    void* pBuf;

    if (pSize)
      *pSize = 0;
    if (!p)
      return NULL;

    comp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS);
    uncomp_size = MZ_READ_LE32(p + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS);

    alloc_size =
      (flags & MZ_ZIP_FLAG_COMPRESSED_DATA) ? comp_size : uncomp_size;
#    ifdef _MSC_VER
    if (((0, sizeof(size_t) == sizeof(mz_uint32))) && (alloc_size > 0x7FFFFFFF))
#    else
    if (((sizeof(size_t) == sizeof(mz_uint32))) && (alloc_size > 0x7FFFFFFF))
#    endif
      return NULL;
    if (
      NULL ==
      (pBuf = pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, (size_t)alloc_size)))
      return NULL;

    if (!mz_zip_reader_extract_to_mem(
          pZip, file_index, pBuf, (size_t)alloc_size, flags))
    {
      pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);
      return NULL;
    }

    if (pSize)
      *pSize = (size_t)alloc_size;
    return pBuf;
  }

  void* mz_zip_reader_extract_file_to_heap(
    mz_zip_archive* pZip, const char* pFilename, size_t* pSize, mz_uint flags)
  {
    int file_index = mz_zip_reader_locate_file(pZip, pFilename, NULL, flags);
    if (file_index < 0)
    {
      if (pSize)
        *pSize = 0;
      return MZ_FALSE;
    }
    return mz_zip_reader_extract_to_heap(pZip, file_index, pSize, flags);
  }

  mz_bool mz_zip_reader_extract_to_callback(
    mz_zip_archive* pZip,
    mz_uint file_index,
    mz_file_write_func pCallback,
    void* pOpaque,
    mz_uint flags)
  {
    int status = TINFL_STATUS_DONE;
    mz_uint file_crc32 = MZ_CRC32_INIT;
    mz_uint64 read_buf_size, read_buf_ofs = 0, read_buf_avail, comp_remaining,
                             out_buf_ofs = 0, cur_file_ofs;
    mz_zip_archive_file_stat file_stat;
    void* pRead_buf = NULL;
    void* pWrite_buf = NULL;
    mz_uint32 local_header_u32
      [(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(mz_uint32) - 1) /
       sizeof(mz_uint32)];
    mz_uint8* pLocal_header = (mz_uint8*)local_header_u32;

    if (!mz_zip_reader_file_stat(pZip, file_index, &file_stat))
      return MZ_FALSE;

    // Empty file, or a directory (but not always a directory - I've seen odd
    // zips with directories that have compressed data which inflates to 0
    // bytes)
    if (!file_stat.m_comp_size)
      return MZ_TRUE;

    // Entry is a subdirectory (I've seen old zips with dir entries which have
    // compressed deflate data which inflates to 0 bytes, but these entries
    // claim to uncompress to 512 bytes in the headers). I'm torn how to handle
    // this case - should it fail instead?
    if (mz_zip_reader_is_file_a_directory(pZip, file_index))
      return MZ_TRUE;

    // Encryption and patch files are not supported.
    if (file_stat.m_bit_flag & (1 | 32))
      return MZ_FALSE;

    // This function only supports stored and deflate.
    if (
      (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (file_stat.m_method != 0) &&
      (file_stat.m_method != MZ_DEFLATED))
      return MZ_FALSE;

    // Read and parse the local directory entry.
    cur_file_ofs = file_stat.m_local_header_ofs;
    if (
      pZip->m_pRead(
        pZip->m_pIO_opaque,
        cur_file_ofs,
        pLocal_header,
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE) != MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    if (MZ_READ_LE32(pLocal_header) != MZ_ZIP_LOCAL_DIR_HEADER_SIG)
      return MZ_FALSE;

    cur_file_ofs += MZ_ZIP_LOCAL_DIR_HEADER_SIZE +
      MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS) +
      MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS);
    if ((cur_file_ofs + file_stat.m_comp_size) > pZip->m_archive_size)
      return MZ_FALSE;

    // Decompress the file either directly from memory or from a file input
    // buffer.
    if (pZip->m_pState->m_pMem)
    {
      pRead_buf = (mz_uint8*)pZip->m_pState->m_pMem + cur_file_ofs;
      read_buf_size = read_buf_avail = file_stat.m_comp_size;
      comp_remaining = 0;
    }
    else
    {
      read_buf_size = MZ_MIN(file_stat.m_comp_size, MZ_ZIP_MAX_IO_BUF_SIZE);
      if (
        NULL ==
        (pRead_buf =
           pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, (size_t)read_buf_size)))
        return MZ_FALSE;
      read_buf_avail = 0;
      comp_remaining = file_stat.m_comp_size;
    }

    if ((flags & MZ_ZIP_FLAG_COMPRESSED_DATA) || (!file_stat.m_method))
    {
      // The file is stored or the caller has requested the compressed data.
      if (pZip->m_pState->m_pMem)
      {
#    ifdef _MSC_VER
        if (
          ((0, sizeof(size_t) == sizeof(mz_uint32))) &&
          (file_stat.m_comp_size > 0xFFFFFFFF))
#    else
        if (
          ((sizeof(size_t) == sizeof(mz_uint32))) &&
          (file_stat.m_comp_size > 0xFFFFFFFF))
#    endif
          return MZ_FALSE;
        if (
          pCallback(
            pOpaque, out_buf_ofs, pRead_buf, (size_t)file_stat.m_comp_size) !=
          file_stat.m_comp_size)
          status = TINFL_STATUS_FAILED;
        else if (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
          file_crc32 = (mz_uint32)mz_crc32(
            file_crc32,
            (const mz_uint8*)pRead_buf,
            (size_t)file_stat.m_comp_size);
        cur_file_ofs += file_stat.m_comp_size;
        out_buf_ofs += file_stat.m_comp_size;
        comp_remaining = 0;
      }
      else
      {
        while (comp_remaining)
        {
          read_buf_avail = MZ_MIN(read_buf_size, comp_remaining);
          if (
            pZip->m_pRead(
              pZip->m_pIO_opaque,
              cur_file_ofs,
              pRead_buf,
              (size_t)read_buf_avail) != read_buf_avail)
          {
            status = TINFL_STATUS_FAILED;
            break;
          }

          if (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
            file_crc32 = (mz_uint32)mz_crc32(
              file_crc32, (const mz_uint8*)pRead_buf, (size_t)read_buf_avail);

          if (
            pCallback(
              pOpaque, out_buf_ofs, pRead_buf, (size_t)read_buf_avail) !=
            read_buf_avail)
          {
            status = TINFL_STATUS_FAILED;
            break;
          }
          cur_file_ofs += read_buf_avail;
          out_buf_ofs += read_buf_avail;
          comp_remaining -= read_buf_avail;
        }
      }
    }
    else
    {
      tinfl_decompressor inflator;
      tinfl_init(&inflator);

      if (
        NULL ==
        (pWrite_buf =
           pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, TINFL_LZ_DICT_SIZE)))
        status = TINFL_STATUS_FAILED;
      else
      {
        do
        {
          mz_uint8* pWrite_buf_cur =
            (mz_uint8*)pWrite_buf + (out_buf_ofs & (TINFL_LZ_DICT_SIZE - 1));
          size_t in_buf_size,
            out_buf_size =
              TINFL_LZ_DICT_SIZE - (out_buf_ofs & (TINFL_LZ_DICT_SIZE - 1));
          if ((!read_buf_avail) && (!pZip->m_pState->m_pMem))
          {
            read_buf_avail = MZ_MIN(read_buf_size, comp_remaining);
            if (
              pZip->m_pRead(
                pZip->m_pIO_opaque,
                cur_file_ofs,
                pRead_buf,
                (size_t)read_buf_avail) != read_buf_avail)
            {
              status = TINFL_STATUS_FAILED;
              break;
            }
            cur_file_ofs += read_buf_avail;
            comp_remaining -= read_buf_avail;
            read_buf_ofs = 0;
          }

          in_buf_size = (size_t)read_buf_avail;
          status = tinfl_decompress(
            &inflator,
            (const mz_uint8*)pRead_buf + read_buf_ofs,
            &in_buf_size,
            (mz_uint8*)pWrite_buf,
            pWrite_buf_cur,
            &out_buf_size,
            comp_remaining ? TINFL_FLAG_HAS_MORE_INPUT : 0);
          read_buf_avail -= in_buf_size;
          read_buf_ofs += in_buf_size;

          if (out_buf_size)
          {
            if (
              pCallback(pOpaque, out_buf_ofs, pWrite_buf_cur, out_buf_size) !=
              out_buf_size)
            {
              status = TINFL_STATUS_FAILED;
              break;
            }
            file_crc32 =
              (mz_uint32)mz_crc32(file_crc32, pWrite_buf_cur, out_buf_size);
            if ((out_buf_ofs += out_buf_size) > file_stat.m_uncomp_size)
            {
              status = TINFL_STATUS_FAILED;
              break;
            }
          }
        } while ((status == TINFL_STATUS_NEEDS_MORE_INPUT) ||
                 (status == TINFL_STATUS_HAS_MORE_OUTPUT));
      }
    }

    if (
      (status == TINFL_STATUS_DONE) && (!(flags & MZ_ZIP_FLAG_COMPRESSED_DATA)))
    {
      // Make sure the entire file was decompressed, and check its CRC.
      if (
        (out_buf_ofs != file_stat.m_uncomp_size) ||
        (file_crc32 != file_stat.m_crc32))
        status = TINFL_STATUS_FAILED;
    }

    if (!pZip->m_pState->m_pMem)
      pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
    if (pWrite_buf)
      pZip->m_pFree(pZip->m_pAlloc_opaque, pWrite_buf);

    return status == TINFL_STATUS_DONE;
  }

  mz_bool mz_zip_reader_extract_file_to_callback(
    mz_zip_archive* pZip,
    const char* pFilename,
    mz_file_write_func pCallback,
    void* pOpaque,
    mz_uint flags)
  {
    int file_index = mz_zip_reader_locate_file(pZip, pFilename, NULL, flags);
    if (file_index < 0)
      return MZ_FALSE;
    return mz_zip_reader_extract_to_callback(
      pZip, file_index, pCallback, pOpaque, flags);
  }

#    ifndef MINIZ_NO_STDIO
  static size_t mz_zip_file_write_callback(
    void* pOpaque, mz_uint64 ofs, const void* pBuf, size_t n)
  {
    (void)ofs;
    return MZ_FWRITE(pBuf, 1, n, (MZ_FILE*)pOpaque);
  }

  mz_bool mz_zip_reader_extract_to_file(
    mz_zip_archive* pZip,
    mz_uint file_index,
    const char* pDst_filename,
    mz_uint flags)
  {
    mz_bool status;
    mz_zip_archive_file_stat file_stat;
    MZ_FILE* pFile;
    if (!mz_zip_reader_file_stat(pZip, file_index, &file_stat))
      return MZ_FALSE;
    pFile = MZ_FOPEN(pDst_filename, "wb");
    if (!pFile)
      return MZ_FALSE;
    status = mz_zip_reader_extract_to_callback(
      pZip, file_index, mz_zip_file_write_callback, pFile, flags);
    if (MZ_FCLOSE(pFile) == EOF)
      return MZ_FALSE;
#      ifndef MINIZ_NO_TIME
    if (status)
      mz_zip_set_file_times(pDst_filename, file_stat.m_time, file_stat.m_time);
#      endif
    return status;
  }
#    endif // #ifndef MINIZ_NO_STDIO

  mz_bool mz_zip_reader_end(mz_zip_archive* pZip)
  {
    if (
      (!pZip) || (!pZip->m_pState) || (!pZip->m_pAlloc) || (!pZip->m_pFree) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_READING))
      return MZ_FALSE;

    if (pZip->m_pState)
    {
      mz_zip_internal_state* pState = pZip->m_pState;
      pZip->m_pState = NULL;
      mz_zip_array_clear(pZip, &pState->m_central_dir);
      mz_zip_array_clear(pZip, &pState->m_central_dir_offsets);
      mz_zip_array_clear(pZip, &pState->m_sorted_central_dir_offsets);

#    ifndef MINIZ_NO_STDIO
      if (pState->m_pFile)
      {
        MZ_FCLOSE(pState->m_pFile);
        pState->m_pFile = NULL;
      }
#    endif // #ifndef MINIZ_NO_STDIO

      pZip->m_pFree(pZip->m_pAlloc_opaque, pState);
    }
    pZip->m_zip_mode = MZ_ZIP_MODE_INVALID;

    return MZ_TRUE;
  }

#    ifndef MINIZ_NO_STDIO
  mz_bool mz_zip_reader_extract_file_to_file(
    mz_zip_archive* pZip,
    const char* pArchive_filename,
    const char* pDst_filename,
    mz_uint flags)
  {
    int file_index =
      mz_zip_reader_locate_file(pZip, pArchive_filename, NULL, flags);
    if (file_index < 0)
      return MZ_FALSE;
    return mz_zip_reader_extract_to_file(
      pZip, file_index, pDst_filename, flags);
  }
#    endif

  // ------------------- .ZIP archive writing

#    ifndef MINIZ_NO_ARCHIVE_WRITING_APIS

  static void mz_write_le16(mz_uint8* p, mz_uint16 v)
  {
    p[0] = (mz_uint8)v;
    p[1] = (mz_uint8)(v >> 8);
  }
  static void mz_write_le32(mz_uint8* p, mz_uint32 v)
  {
    p[0] = (mz_uint8)v;
    p[1] = (mz_uint8)(v >> 8);
    p[2] = (mz_uint8)(v >> 16);
    p[3] = (mz_uint8)(v >> 24);
  }
#      define MZ_WRITE_LE16(p, v) mz_write_le16((mz_uint8*)(p), (mz_uint16)(v))
#      define MZ_WRITE_LE32(p, v) mz_write_le32((mz_uint8*)(p), (mz_uint32)(v))

  mz_bool mz_zip_writer_init(mz_zip_archive* pZip, mz_uint64 existing_size)
  {
    if (
      (!pZip) || (pZip->m_pState) || (!pZip->m_pWrite) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_INVALID))
      return MZ_FALSE;

    if (pZip->m_file_offset_alignment)
    {
      // Ensure user specified file offset alignment is a power of 2.
      if (pZip->m_file_offset_alignment & (pZip->m_file_offset_alignment - 1))
        return MZ_FALSE;
    }

    if (!pZip->m_pAlloc)
      pZip->m_pAlloc = def_alloc_func;
    if (!pZip->m_pFree)
      pZip->m_pFree = def_free_func;
    if (!pZip->m_pRealloc)
      pZip->m_pRealloc = def_realloc_func;

    pZip->m_zip_mode = MZ_ZIP_MODE_WRITING;
    pZip->m_archive_size = existing_size;
    pZip->m_central_directory_file_ofs = 0;
    pZip->m_total_files = 0;

    if (
      NULL ==
      (pZip->m_pState = (mz_zip_internal_state*)pZip->m_pAlloc(
         pZip->m_pAlloc_opaque, 1, sizeof(mz_zip_internal_state))))
      return MZ_FALSE;
    memset(pZip->m_pState, 0, sizeof(mz_zip_internal_state));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_central_dir, sizeof(mz_uint8));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_central_dir_offsets, sizeof(mz_uint32));
    MZ_ZIP_ARRAY_SET_ELEMENT_SIZE(
      &pZip->m_pState->m_sorted_central_dir_offsets, sizeof(mz_uint32));
    return MZ_TRUE;
  }

  static size_t mz_zip_heap_write_func(
    void* pOpaque, mz_uint64 file_ofs, const void* pBuf, size_t n)
  {
    mz_zip_archive* pZip = (mz_zip_archive*)pOpaque;
    mz_zip_internal_state* pState = pZip->m_pState;
    mz_uint64 new_size = MZ_MAX(file_ofs + n, pState->m_mem_size);
#      ifdef _MSC_VER
    if (
      (!n) ||
      ((0, sizeof(size_t) == sizeof(mz_uint32)) && (new_size > 0x7FFFFFFF)))
#      else
    if (
      (!n) ||
      ((sizeof(size_t) == sizeof(mz_uint32)) && (new_size > 0x7FFFFFFF)))
#      endif
      return 0;
    if (new_size > pState->m_mem_capacity)
    {
      void* pNew_block;
      size_t new_capacity = MZ_MAX(64, pState->m_mem_capacity);
      while (new_capacity < new_size)
        new_capacity *= 2;
      if (
        NULL ==
        (pNew_block = pZip->m_pRealloc(
           pZip->m_pAlloc_opaque, pState->m_pMem, 1, new_capacity)))
        return 0;
      pState->m_pMem = pNew_block;
      pState->m_mem_capacity = new_capacity;
    }
    memcpy((mz_uint8*)pState->m_pMem + file_ofs, pBuf, n);
    pState->m_mem_size = (size_t)new_size;
    return n;
  }

  mz_bool mz_zip_writer_init_heap(
    mz_zip_archive* pZip,
    size_t size_to_reserve_at_beginning,
    size_t initial_allocation_size)
  {
    pZip->m_pWrite = mz_zip_heap_write_func;
    pZip->m_pIO_opaque = pZip;
    if (!mz_zip_writer_init(pZip, size_to_reserve_at_beginning))
      return MZ_FALSE;
    if (
      0 !=
      (initial_allocation_size =
         MZ_MAX(initial_allocation_size, size_to_reserve_at_beginning)))
    {
      if (
        NULL ==
        (pZip->m_pState->m_pMem =
           pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, initial_allocation_size)))
      {
        mz_zip_writer_end(pZip);
        return MZ_FALSE;
      }
      pZip->m_pState->m_mem_capacity = initial_allocation_size;
    }
    return MZ_TRUE;
  }

#      ifndef MINIZ_NO_STDIO
  static size_t mz_zip_file_write_func(
    void* pOpaque, mz_uint64 file_ofs, const void* pBuf, size_t n)
  {
    mz_zip_archive* pZip = (mz_zip_archive*)pOpaque;
    mz_int64 cur_ofs = MZ_FTELL64(pZip->m_pState->m_pFile);
    if (
      ((mz_int64)file_ofs < 0) ||
      (((cur_ofs != (mz_int64)file_ofs)) &&
       (MZ_FSEEK64(pZip->m_pState->m_pFile, (mz_int64)file_ofs, SEEK_SET))))
      return 0;
    return MZ_FWRITE(pBuf, 1, n, pZip->m_pState->m_pFile);
  }

  mz_bool mz_zip_writer_init_file(
    mz_zip_archive* pZip,
    const char* pFilename,
    mz_uint64 size_to_reserve_at_beginning)
  {
    MZ_FILE* pFile;
    pZip->m_pWrite = mz_zip_file_write_func;
    pZip->m_pIO_opaque = pZip;
    if (!mz_zip_writer_init(pZip, size_to_reserve_at_beginning))
      return MZ_FALSE;
    if (NULL == (pFile = MZ_FOPEN(pFilename, "wb")))
    {
      mz_zip_writer_end(pZip);
      return MZ_FALSE;
    }
    pZip->m_pState->m_pFile = pFile;
    if (size_to_reserve_at_beginning)
    {
      mz_uint64 cur_ofs = 0;
      char buf[4096];
      MZ_CLEAR_OBJ(buf);
      do
      {
        size_t n = (size_t)MZ_MIN(sizeof(buf), size_to_reserve_at_beginning);
        if (pZip->m_pWrite(pZip->m_pIO_opaque, cur_ofs, buf, n) != n)
        {
          mz_zip_writer_end(pZip);
          return MZ_FALSE;
        }
        cur_ofs += n;
        size_to_reserve_at_beginning -= n;
      } while (size_to_reserve_at_beginning);
    }
    return MZ_TRUE;
  }
#      endif // #ifndef MINIZ_NO_STDIO

  mz_bool
  mz_zip_writer_init_from_reader(mz_zip_archive* pZip, const char* pFilename)
  {
    mz_zip_internal_state* pState;
    if (
      (!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode != MZ_ZIP_MODE_READING))
      return MZ_FALSE;
    // No sense in trying to write to an archive that's already at the support
    // max size
    if (
      (pZip->m_total_files == 0xFFFF) ||
      ((pZip->m_archive_size + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE +
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE) > 0xFFFFFFFF))
      return MZ_FALSE;

    pState = pZip->m_pState;

    if (pState->m_pFile)
    {
#      ifdef MINIZ_NO_STDIO
      pFilename;
      return MZ_FALSE;
#      else
      // Archive is being read from stdio - try to reopen as writable.
      if (pZip->m_pIO_opaque != pZip)
        return MZ_FALSE;
      if (!pFilename)
        return MZ_FALSE;
      pZip->m_pWrite = mz_zip_file_write_func;
      if (
        NULL ==
        (pState->m_pFile = MZ_FREOPEN(pFilename, "r+b", pState->m_pFile)))
      {
        // The mz_zip_archive is now in a bogus state because pState->m_pFile is
        // NULL, so just close it.
        mz_zip_reader_end(pZip);
        return MZ_FALSE;
      }
#      endif // #ifdef MINIZ_NO_STDIO
    }
    else if (pState->m_pMem)
    {
      // Archive lives in a memory block. Assume it's from the heap that we can
      // resize using the realloc callback.
      if (pZip->m_pIO_opaque != pZip)
        return MZ_FALSE;
      pState->m_mem_capacity = pState->m_mem_size;
      pZip->m_pWrite = mz_zip_heap_write_func;
    }
    // Archive is being read via a user provided read function - make sure the
    // user has specified a write function too.
    else if (!pZip->m_pWrite)
      return MZ_FALSE;

    // Start writing new files at the archive's current central directory
    // location.
    pZip->m_archive_size = pZip->m_central_directory_file_ofs;
    pZip->m_zip_mode = MZ_ZIP_MODE_WRITING;
    pZip->m_central_directory_file_ofs = 0;

    return MZ_TRUE;
  }

  mz_bool mz_zip_writer_add_mem(
    mz_zip_archive* pZip,
    const char* pArchive_name,
    const void* pBuf,
    size_t buf_size,
    mz_uint level_and_flags)
  {
    return mz_zip_writer_add_mem_ex(
      pZip, pArchive_name, pBuf, buf_size, NULL, 0, level_and_flags, 0, 0);
  }

  typedef struct
  {
    mz_zip_archive* m_pZip;
    mz_uint64 m_cur_archive_file_ofs;
    mz_uint64 m_comp_size;
  } mz_zip_writer_add_state;

  static mz_bool
  mz_zip_writer_add_put_buf_callback(const void* pBuf, int len, void* pUser)
  {
    mz_zip_writer_add_state* pState = (mz_zip_writer_add_state*)pUser;
    if (
      (int)pState->m_pZip->m_pWrite(
        pState->m_pZip->m_pIO_opaque,
        pState->m_cur_archive_file_ofs,
        pBuf,
        len) != len)
      return MZ_FALSE;
    pState->m_cur_archive_file_ofs += len;
    pState->m_comp_size += len;
    return MZ_TRUE;
  }

  static mz_bool mz_zip_writer_create_local_dir_header(
    mz_zip_archive* pZip,
    mz_uint8* pDst,
    mz_uint16 filename_size,
    mz_uint16 extra_size,
    mz_uint64 uncomp_size,
    mz_uint64 comp_size,
    mz_uint32 uncomp_crc32,
    mz_uint16 method,
    mz_uint16 bit_flags,
    mz_uint16 dos_time,
    mz_uint16 dos_date)
  {
    (void)pZip;
    memset(pDst, 0, MZ_ZIP_LOCAL_DIR_HEADER_SIZE);
    MZ_WRITE_LE32(pDst + MZ_ZIP_LDH_SIG_OFS, MZ_ZIP_LOCAL_DIR_HEADER_SIG);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_VERSION_NEEDED_OFS, method ? 20 : 0);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_BIT_FLAG_OFS, bit_flags);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_METHOD_OFS, method);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_FILE_TIME_OFS, dos_time);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_FILE_DATE_OFS, dos_date);
    MZ_WRITE_LE32(pDst + MZ_ZIP_LDH_CRC32_OFS, uncomp_crc32);
    MZ_WRITE_LE32(pDst + MZ_ZIP_LDH_COMPRESSED_SIZE_OFS, comp_size);
    MZ_WRITE_LE32(pDst + MZ_ZIP_LDH_DECOMPRESSED_SIZE_OFS, uncomp_size);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_FILENAME_LEN_OFS, filename_size);
    MZ_WRITE_LE16(pDst + MZ_ZIP_LDH_EXTRA_LEN_OFS, extra_size);
    return MZ_TRUE;
  }

  static mz_bool mz_zip_writer_create_central_dir_header(
    mz_zip_archive* pZip,
    mz_uint8* pDst,
    mz_uint16 filename_size,
    mz_uint16 extra_size,
    mz_uint16 comment_size,
    mz_uint64 uncomp_size,
    mz_uint64 comp_size,
    mz_uint32 uncomp_crc32,
    mz_uint16 method,
    mz_uint16 bit_flags,
    mz_uint16 dos_time,
    mz_uint16 dos_date,
    mz_uint64 local_header_ofs,
    mz_uint32 ext_attributes)
  {
    (void)pZip;
    memset(pDst, 0, MZ_ZIP_CENTRAL_DIR_HEADER_SIZE);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_SIG_OFS, MZ_ZIP_CENTRAL_DIR_HEADER_SIG);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_VERSION_NEEDED_OFS, method ? 20 : 0);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_BIT_FLAG_OFS, bit_flags);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_METHOD_OFS, method);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_FILE_TIME_OFS, dos_time);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_FILE_DATE_OFS, dos_date);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_CRC32_OFS, uncomp_crc32);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS, comp_size);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_DECOMPRESSED_SIZE_OFS, uncomp_size);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_FILENAME_LEN_OFS, filename_size);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_EXTRA_LEN_OFS, extra_size);
    MZ_WRITE_LE16(pDst + MZ_ZIP_CDH_COMMENT_LEN_OFS, comment_size);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_EXTERNAL_ATTR_OFS, ext_attributes);
    MZ_WRITE_LE32(pDst + MZ_ZIP_CDH_LOCAL_HEADER_OFS, local_header_ofs);
    return MZ_TRUE;
  }

  static mz_bool mz_zip_writer_add_to_central_dir(
    mz_zip_archive* pZip,
    const char* pFilename,
    mz_uint16 filename_size,
    const void* pExtra,
    mz_uint16 extra_size,
    const void* pComment,
    mz_uint16 comment_size,
    mz_uint64 uncomp_size,
    mz_uint64 comp_size,
    mz_uint32 uncomp_crc32,
    mz_uint16 method,
    mz_uint16 bit_flags,
    mz_uint16 dos_time,
    mz_uint16 dos_date,
    mz_uint64 local_header_ofs,
    mz_uint32 ext_attributes)
  {
    mz_zip_internal_state* pState = pZip->m_pState;
    mz_uint32 central_dir_ofs = (mz_uint32)pState->m_central_dir.m_size;
    size_t orig_central_dir_size = pState->m_central_dir.m_size;
    mz_uint8 central_dir_header[MZ_ZIP_CENTRAL_DIR_HEADER_SIZE];

    // No zip64 support yet
    if (
      (local_header_ofs > 0xFFFFFFFF) ||
      (((mz_uint64)pState->m_central_dir.m_size +
        MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + filename_size + extra_size +
        comment_size) > 0xFFFFFFFF))
      return MZ_FALSE;

    if (!mz_zip_writer_create_central_dir_header(
          pZip,
          central_dir_header,
          filename_size,
          extra_size,
          comment_size,
          uncomp_size,
          comp_size,
          uncomp_crc32,
          method,
          bit_flags,
          dos_time,
          dos_date,
          local_header_ofs,
          ext_attributes))
      return MZ_FALSE;

    if (
      (!mz_zip_array_push_back(
        pZip,
        &pState->m_central_dir,
        central_dir_header,
        MZ_ZIP_CENTRAL_DIR_HEADER_SIZE)) ||
      (!mz_zip_array_push_back(
        pZip, &pState->m_central_dir, pFilename, filename_size)) ||
      (!mz_zip_array_push_back(
        pZip, &pState->m_central_dir, pExtra, extra_size)) ||
      (!mz_zip_array_push_back(
        pZip, &pState->m_central_dir, pComment, comment_size)) ||
      (!mz_zip_array_push_back(
        pZip, &pState->m_central_dir_offsets, &central_dir_ofs, 1)))
    {
      // Try to push the central directory array back into its original state.
      mz_zip_array_resize(
        pZip, &pState->m_central_dir, orig_central_dir_size, MZ_FALSE);
      return MZ_FALSE;
    }

    return MZ_TRUE;
  }

  static mz_bool mz_zip_writer_validate_archive_name(const char* pArchive_name)
  {
    // Basic ZIP archive filename validity checks: Valid filenames cannot start
    // with a forward slash, cannot contain a drive letter, and cannot use
    // DOS-style backward slashes.
    if (*pArchive_name == '/')
      return MZ_FALSE;
    while (*pArchive_name)
    {
      if ((*pArchive_name == '\\') || (*pArchive_name == ':'))
        return MZ_FALSE;
      pArchive_name++;
    }
    return MZ_TRUE;
  }

  static mz_uint
  mz_zip_writer_compute_padding_needed_for_file_alignment(mz_zip_archive* pZip)
  {
    mz_uint32 n;
    if (!pZip->m_file_offset_alignment)
      return 0;
    n = (mz_uint32)(pZip->m_archive_size & (pZip->m_file_offset_alignment - 1));
    return (pZip->m_file_offset_alignment - n) &
      (pZip->m_file_offset_alignment - 1);
  }

  static mz_bool mz_zip_writer_write_zeros(
    mz_zip_archive* pZip, mz_uint64 cur_file_ofs, mz_uint32 n)
  {
    char buf[4096];
    memset(buf, 0, MZ_MIN(sizeof(buf), n));
    while (n)
    {
      mz_uint32 s = MZ_MIN(sizeof(buf), n);
      if (pZip->m_pWrite(pZip->m_pIO_opaque, cur_file_ofs, buf, s) != s)
        return MZ_FALSE;
      cur_file_ofs += s;
      n -= s;
    }
    return MZ_TRUE;
  }

  mz_bool mz_zip_writer_add_mem_ex(
    mz_zip_archive* pZip,
    const char* pArchive_name,
    const void* pBuf,
    size_t buf_size,
    const void* pComment,
    mz_uint16 comment_size,
    mz_uint level_and_flags,
    mz_uint64 uncomp_size,
    mz_uint32 uncomp_crc32)
  {
    mz_uint16 method = 0, dos_time = 0, dos_date = 0;
    mz_uint level, ext_attributes = 0, num_alignment_padding_bytes;
    mz_uint64 local_dir_header_ofs = pZip->m_archive_size,
              cur_archive_file_ofs = pZip->m_archive_size, comp_size = 0;
    size_t archive_name_size;
    mz_uint8 local_dir_header[MZ_ZIP_LOCAL_DIR_HEADER_SIZE];
    tdefl_compressor* pComp = NULL;
    mz_bool store_data_uncompressed;
    mz_zip_internal_state* pState;

    if ((int)level_and_flags < 0)
      level_and_flags = MZ_DEFAULT_LEVEL;
    level = level_and_flags & 0xF;
    store_data_uncompressed =
      ((!level) || (level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA));

    if (
      (!pZip) || (!pZip->m_pState) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_WRITING) || ((buf_size) && (!pBuf)) ||
      (!pArchive_name) || ((comment_size) && (!pComment)) ||
      (pZip->m_total_files == 0xFFFF) || (level > MZ_UBER_COMPRESSION))
      return MZ_FALSE;

    pState = pZip->m_pState;

    if ((!(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)) && (uncomp_size))
      return MZ_FALSE;
    // No zip64 support yet
    if ((buf_size > 0xFFFFFFFF) || (uncomp_size > 0xFFFFFFFF))
      return MZ_FALSE;
    if (!mz_zip_writer_validate_archive_name(pArchive_name))
      return MZ_FALSE;

#      ifndef MINIZ_NO_TIME
    {
      time_t cur_time;
      time(&cur_time);
      mz_zip_time_to_dos_time(cur_time, &dos_time, &dos_date);
    }
#      endif // #ifndef MINIZ_NO_TIME

    archive_name_size = strlen(pArchive_name);
    if (archive_name_size > 0xFFFF)
      return MZ_FALSE;

    num_alignment_padding_bytes =
      mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);

    // no zip64 support yet
    if (
      (pZip->m_total_files == 0xFFFF) ||
      ((pZip->m_archive_size + num_alignment_padding_bytes +
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE +
        comment_size + archive_name_size) > 0xFFFFFFFF))
      return MZ_FALSE;

    if ((archive_name_size) && (pArchive_name[archive_name_size - 1] == '/'))
    {
      // Set DOS Subdirectory attribute bit.
      ext_attributes |= 0x10;
      // Subdirectories cannot contain data.
      if ((buf_size) || (uncomp_size))
        return MZ_FALSE;
    }

    // Try to do any allocations before writing to the archive, so if an
    // allocation fails the file remains unmodified. (A good idea if we're doing
    // an in-place modification.)
    if (
      (!mz_zip_array_ensure_room(
        pZip,
        &pState->m_central_dir,
        MZ_ZIP_CENTRAL_DIR_HEADER_SIZE + archive_name_size + comment_size)) ||
      (!mz_zip_array_ensure_room(pZip, &pState->m_central_dir_offsets, 1)))
      return MZ_FALSE;

    if ((!store_data_uncompressed) && (buf_size))
    {
      if (
        NULL ==
        (pComp = (tdefl_compressor*)pZip->m_pAlloc(
           pZip->m_pAlloc_opaque, 1, sizeof(tdefl_compressor))))
        return MZ_FALSE;
    }

    if (!mz_zip_writer_write_zeros(
          pZip,
          cur_archive_file_ofs,
          num_alignment_padding_bytes + sizeof(local_dir_header)))
    {
      pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
      return MZ_FALSE;
    }
    local_dir_header_ofs += num_alignment_padding_bytes;
    if (pZip->m_file_offset_alignment)
    {
      MZ_ASSERT(
        (local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0);
    }
    cur_archive_file_ofs +=
      num_alignment_padding_bytes + sizeof(local_dir_header);

    MZ_CLEAR_OBJ(local_dir_header);
    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque,
        cur_archive_file_ofs,
        pArchive_name,
        archive_name_size) != archive_name_size)
    {
      pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
      return MZ_FALSE;
    }
    cur_archive_file_ofs += archive_name_size;

    if (!(level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA))
    {
      uncomp_crc32 =
        (mz_uint32)mz_crc32(MZ_CRC32_INIT, (const mz_uint8*)pBuf, buf_size);
      uncomp_size = buf_size;
      if (uncomp_size <= 3)
      {
        level = 0;
        store_data_uncompressed = MZ_TRUE;
      }
    }

    if (store_data_uncompressed)
    {
      if (
        pZip->m_pWrite(
          pZip->m_pIO_opaque, cur_archive_file_ofs, pBuf, buf_size) != buf_size)
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
        return MZ_FALSE;
      }

      cur_archive_file_ofs += buf_size;
      comp_size = buf_size;

      if (level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)
        method = MZ_DEFLATED;
    }
    else if (buf_size)
    {
      mz_zip_writer_add_state state;

      state.m_pZip = pZip;
      state.m_cur_archive_file_ofs = cur_archive_file_ofs;
      state.m_comp_size = 0;

      if (
        (tdefl_init(
           pComp,
           mz_zip_writer_add_put_buf_callback,
           &state,
           tdefl_create_comp_flags_from_zip_params(
             level, -15, MZ_DEFAULT_STRATEGY)) != TDEFL_STATUS_OKAY) ||
        (tdefl_compress_buffer(pComp, pBuf, buf_size, TDEFL_FINISH) !=
         TDEFL_STATUS_DONE))
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
        return MZ_FALSE;
      }

      comp_size = state.m_comp_size;
      cur_archive_file_ofs = state.m_cur_archive_file_ofs;

      method = MZ_DEFLATED;
    }

    pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
    pComp = NULL;

    // no zip64 support yet
    if ((comp_size > 0xFFFFFFFF) || (cur_archive_file_ofs > 0xFFFFFFFF))
      return MZ_FALSE;

    if (!mz_zip_writer_create_local_dir_header(
          pZip,
          local_dir_header,
          (mz_uint16)archive_name_size,
          0,
          uncomp_size,
          comp_size,
          uncomp_crc32,
          method,
          0,
          dos_time,
          dos_date))
      return MZ_FALSE;

    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque,
        local_dir_header_ofs,
        local_dir_header,
        sizeof(local_dir_header)) != sizeof(local_dir_header))
      return MZ_FALSE;

    if (!mz_zip_writer_add_to_central_dir(
          pZip,
          pArchive_name,
          (mz_uint16)archive_name_size,
          NULL,
          0,
          pComment,
          comment_size,
          uncomp_size,
          comp_size,
          uncomp_crc32,
          method,
          0,
          dos_time,
          dos_date,
          local_dir_header_ofs,
          ext_attributes))
      return MZ_FALSE;

    pZip->m_total_files++;
    pZip->m_archive_size = cur_archive_file_ofs;

    return MZ_TRUE;
  }

#      ifndef MINIZ_NO_STDIO
  mz_bool mz_zip_writer_add_file(
    mz_zip_archive* pZip,
    const char* pArchive_name,
    const char* pSrc_filename,
    const void* pComment,
    mz_uint16 comment_size,
    mz_uint level_and_flags)
  {
    mz_uint uncomp_crc32 = MZ_CRC32_INIT, level, num_alignment_padding_bytes;
    mz_uint16 method = 0, dos_time = 0, dos_date = 0, ext_attributes = 0;
    mz_uint64 local_dir_header_ofs = pZip->m_archive_size,
              cur_archive_file_ofs = pZip->m_archive_size, uncomp_size = 0,
              comp_size = 0;
    size_t archive_name_size;
    mz_uint8 local_dir_header[MZ_ZIP_LOCAL_DIR_HEADER_SIZE];
    MZ_FILE* pSrc_file = NULL;

    if ((int)level_and_flags < 0)
      level_and_flags = MZ_DEFAULT_LEVEL;
    level = level_and_flags & 0xF;

    if (
      (!pZip) || (!pZip->m_pState) ||
      (pZip->m_zip_mode != MZ_ZIP_MODE_WRITING) || (!pArchive_name) ||
      ((comment_size) && (!pComment)) || (level > MZ_UBER_COMPRESSION))
      return MZ_FALSE;
    if (level_and_flags & MZ_ZIP_FLAG_COMPRESSED_DATA)
      return MZ_FALSE;
    if (!mz_zip_writer_validate_archive_name(pArchive_name))
      return MZ_FALSE;

    archive_name_size = strlen(pArchive_name);
    if (archive_name_size > 0xFFFF)
      return MZ_FALSE;

    num_alignment_padding_bytes =
      mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);

    // no zip64 support yet
    if (
      (pZip->m_total_files == 0xFFFF) ||
      ((pZip->m_archive_size + num_alignment_padding_bytes +
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE +
        comment_size + archive_name_size) > 0xFFFFFFFF))
      return MZ_FALSE;

    if (!mz_zip_get_file_modified_time(pSrc_filename, &dos_time, &dos_date))
      return MZ_FALSE;

    pSrc_file = MZ_FOPEN(pSrc_filename, "rb");
    if (!pSrc_file)
      return MZ_FALSE;
    MZ_FSEEK64(pSrc_file, 0, SEEK_END);
    uncomp_size = MZ_FTELL64(pSrc_file);
    MZ_FSEEK64(pSrc_file, 0, SEEK_SET);

    if (uncomp_size > 0xFFFFFFFF)
    {
      // No zip64 support yet
      MZ_FCLOSE(pSrc_file);
      return MZ_FALSE;
    }
    if (uncomp_size <= 3)
      level = 0;

    if (!mz_zip_writer_write_zeros(
          pZip,
          cur_archive_file_ofs,
          num_alignment_padding_bytes + sizeof(local_dir_header)))
    {
      MZ_FCLOSE(pSrc_file);
      return MZ_FALSE;
    }
    local_dir_header_ofs += num_alignment_padding_bytes;
    if (pZip->m_file_offset_alignment)
    {
      MZ_ASSERT(
        (local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0);
    }
    cur_archive_file_ofs +=
      num_alignment_padding_bytes + sizeof(local_dir_header);

    MZ_CLEAR_OBJ(local_dir_header);
    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque,
        cur_archive_file_ofs,
        pArchive_name,
        archive_name_size) != archive_name_size)
    {
      MZ_FCLOSE(pSrc_file);
      return MZ_FALSE;
    }
    cur_archive_file_ofs += archive_name_size;

    if (uncomp_size)
    {
      mz_uint64 uncomp_remaining = uncomp_size;
      void* pRead_buf =
        pZip->m_pAlloc(pZip->m_pAlloc_opaque, 1, MZ_ZIP_MAX_IO_BUF_SIZE);
      if (!pRead_buf)
      {
        MZ_FCLOSE(pSrc_file);
        return MZ_FALSE;
      }

      if (!level)
      {
        while (uncomp_remaining)
        {
          mz_uint n = (mz_uint)MZ_MIN(MZ_ZIP_MAX_IO_BUF_SIZE, uncomp_remaining);
          if (
            (MZ_FREAD(pRead_buf, 1, n, pSrc_file) != n) ||
            (pZip->m_pWrite(
               pZip->m_pIO_opaque, cur_archive_file_ofs, pRead_buf, n) != n))
          {
            pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
            MZ_FCLOSE(pSrc_file);
            return MZ_FALSE;
          }
          uncomp_crc32 =
            (mz_uint32)mz_crc32(uncomp_crc32, (const mz_uint8*)pRead_buf, n);
          uncomp_remaining -= n;
          cur_archive_file_ofs += n;
        }
        comp_size = uncomp_size;
      }
      else
      {
        mz_bool result = MZ_FALSE;
        mz_zip_writer_add_state state;
        tdefl_compressor* pComp = (tdefl_compressor*)pZip->m_pAlloc(
          pZip->m_pAlloc_opaque, 1, sizeof(tdefl_compressor));
        if (!pComp)
        {
          pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
          MZ_FCLOSE(pSrc_file);
          return MZ_FALSE;
        }

        state.m_pZip = pZip;
        state.m_cur_archive_file_ofs = cur_archive_file_ofs;
        state.m_comp_size = 0;

        if (
          tdefl_init(
            pComp,
            mz_zip_writer_add_put_buf_callback,
            &state,
            tdefl_create_comp_flags_from_zip_params(
              level, -15, MZ_DEFAULT_STRATEGY)) != TDEFL_STATUS_OKAY)
        {
          pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);
          pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
          MZ_FCLOSE(pSrc_file);
          return MZ_FALSE;
        }

        for (;;)
        {
          size_t in_buf_size =
            (mz_uint32)MZ_MIN(uncomp_remaining, MZ_ZIP_MAX_IO_BUF_SIZE);
          tdefl_status status;

          if (MZ_FREAD(pRead_buf, 1, in_buf_size, pSrc_file) != in_buf_size)
            break;

          uncomp_crc32 = (mz_uint32)mz_crc32(
            uncomp_crc32, (const mz_uint8*)pRead_buf, in_buf_size);
          uncomp_remaining -= in_buf_size;

          status = tdefl_compress_buffer(
            pComp,
            pRead_buf,
            in_buf_size,
            uncomp_remaining ? TDEFL_NO_FLUSH : TDEFL_FINISH);
          if (status == TDEFL_STATUS_DONE)
          {
            result = MZ_TRUE;
            break;
          }
          else if (status != TDEFL_STATUS_OKAY)
            break;
        }

        pZip->m_pFree(pZip->m_pAlloc_opaque, pComp);

        if (!result)
        {
          pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
          MZ_FCLOSE(pSrc_file);
          return MZ_FALSE;
        }

        comp_size = state.m_comp_size;
        cur_archive_file_ofs = state.m_cur_archive_file_ofs;

        method = MZ_DEFLATED;
      }

      pZip->m_pFree(pZip->m_pAlloc_opaque, pRead_buf);
    }

    MZ_FCLOSE(pSrc_file);
    pSrc_file = NULL;

    // no zip64 support yet
    if ((comp_size > 0xFFFFFFFF) || (cur_archive_file_ofs > 0xFFFFFFFF))
      return MZ_FALSE;

    if (!mz_zip_writer_create_local_dir_header(
          pZip,
          local_dir_header,
          (mz_uint16)archive_name_size,
          0,
          uncomp_size,
          comp_size,
          uncomp_crc32,
          method,
          0,
          dos_time,
          dos_date))
      return MZ_FALSE;

    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque,
        local_dir_header_ofs,
        local_dir_header,
        sizeof(local_dir_header)) != sizeof(local_dir_header))
      return MZ_FALSE;

    if (!mz_zip_writer_add_to_central_dir(
          pZip,
          pArchive_name,
          (mz_uint16)archive_name_size,
          NULL,
          0,
          pComment,
          comment_size,
          uncomp_size,
          comp_size,
          uncomp_crc32,
          method,
          0,
          dos_time,
          dos_date,
          local_dir_header_ofs,
          ext_attributes))
      return MZ_FALSE;

    pZip->m_total_files++;
    pZip->m_archive_size = cur_archive_file_ofs;

    return MZ_TRUE;
  }
#      endif // #ifndef MINIZ_NO_STDIO

  mz_bool mz_zip_writer_add_from_zip_reader(
    mz_zip_archive* pZip, mz_zip_archive* pSource_zip, mz_uint file_index)
  {
    mz_uint n, bit_flags, num_alignment_padding_bytes;
    mz_uint64 comp_bytes_remaining, local_dir_header_ofs;
    mz_uint64 cur_src_file_ofs, cur_dst_file_ofs;
    mz_uint32 local_header_u32
      [(MZ_ZIP_LOCAL_DIR_HEADER_SIZE + sizeof(mz_uint32) - 1) /
       sizeof(mz_uint32)];
    mz_uint8* pLocal_header = (mz_uint8*)local_header_u32;
    mz_uint8 central_header[MZ_ZIP_CENTRAL_DIR_HEADER_SIZE];
    size_t orig_central_dir_size;
    mz_zip_internal_state* pState;
    void* pBuf;
    const mz_uint8* pSrc_central_header;

    if (
      (!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode != MZ_ZIP_MODE_WRITING))
      return MZ_FALSE;
    if (
      NULL ==
      (pSrc_central_header = mz_zip_reader_get_cdh(pSource_zip, file_index)))
      return MZ_FALSE;
    pState = pZip->m_pState;

    num_alignment_padding_bytes =
      mz_zip_writer_compute_padding_needed_for_file_alignment(pZip);

    // no zip64 support yet
    if (
      (pZip->m_total_files == 0xFFFF) ||
      ((pZip->m_archive_size + num_alignment_padding_bytes +
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE) >
       0xFFFFFFFF))
      return MZ_FALSE;

    cur_src_file_ofs =
      MZ_READ_LE32(pSrc_central_header + MZ_ZIP_CDH_LOCAL_HEADER_OFS);
    cur_dst_file_ofs = pZip->m_archive_size;

    if (
      pSource_zip->m_pRead(
        pSource_zip->m_pIO_opaque,
        cur_src_file_ofs,
        pLocal_header,
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE) != MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    if (MZ_READ_LE32(pLocal_header) != MZ_ZIP_LOCAL_DIR_HEADER_SIG)
      return MZ_FALSE;
    cur_src_file_ofs += MZ_ZIP_LOCAL_DIR_HEADER_SIZE;

    if (!mz_zip_writer_write_zeros(
          pZip, cur_dst_file_ofs, num_alignment_padding_bytes))
      return MZ_FALSE;
    cur_dst_file_ofs += num_alignment_padding_bytes;
    local_dir_header_ofs = cur_dst_file_ofs;
    if (pZip->m_file_offset_alignment)
    {
      MZ_ASSERT(
        (local_dir_header_ofs & (pZip->m_file_offset_alignment - 1)) == 0);
    }

    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque,
        cur_dst_file_ofs,
        pLocal_header,
        MZ_ZIP_LOCAL_DIR_HEADER_SIZE) != MZ_ZIP_LOCAL_DIR_HEADER_SIZE)
      return MZ_FALSE;
    cur_dst_file_ofs += MZ_ZIP_LOCAL_DIR_HEADER_SIZE;

    n = MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_FILENAME_LEN_OFS) +
      MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_EXTRA_LEN_OFS);
    comp_bytes_remaining =
      n + MZ_READ_LE32(pSrc_central_header + MZ_ZIP_CDH_COMPRESSED_SIZE_OFS);

    if (
      NULL ==
      (pBuf = pZip->m_pAlloc(
         pZip->m_pAlloc_opaque,
         1,
         (size_t)MZ_MAX(
           sizeof(mz_uint32) * 4,
           MZ_MIN(MZ_ZIP_MAX_IO_BUF_SIZE, comp_bytes_remaining)))))
      return MZ_FALSE;

    while (comp_bytes_remaining)
    {
      n = (mz_uint)MZ_MIN(MZ_ZIP_MAX_IO_BUF_SIZE, comp_bytes_remaining);
      if (
        pSource_zip->m_pRead(
          pSource_zip->m_pIO_opaque, cur_src_file_ofs, pBuf, n) != n)
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);
        return MZ_FALSE;
      }
      cur_src_file_ofs += n;

      if (pZip->m_pWrite(pZip->m_pIO_opaque, cur_dst_file_ofs, pBuf, n) != n)
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);
        return MZ_FALSE;
      }
      cur_dst_file_ofs += n;

      comp_bytes_remaining -= n;
    }

    bit_flags = MZ_READ_LE16(pLocal_header + MZ_ZIP_LDH_BIT_FLAG_OFS);
    if (bit_flags & 8)
    {
      // Copy data descriptor
      if (
        pSource_zip->m_pRead(
          pSource_zip->m_pIO_opaque,
          cur_src_file_ofs,
          pBuf,
          sizeof(mz_uint32) * 4) != sizeof(mz_uint32) * 4)
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);
        return MZ_FALSE;
      }

      n = sizeof(mz_uint32) * ((MZ_READ_LE32(pBuf) == 0x08074b50) ? 4 : 3);
      if (pZip->m_pWrite(pZip->m_pIO_opaque, cur_dst_file_ofs, pBuf, n) != n)
      {
        pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);
        return MZ_FALSE;
      }

      cur_src_file_ofs += n;
      cur_dst_file_ofs += n;
    }
    pZip->m_pFree(pZip->m_pAlloc_opaque, pBuf);

    // no zip64 support yet
    if (cur_dst_file_ofs > 0xFFFFFFFF)
      return MZ_FALSE;

    orig_central_dir_size = pState->m_central_dir.m_size;

    memcpy(central_header, pSrc_central_header, MZ_ZIP_CENTRAL_DIR_HEADER_SIZE);
    MZ_WRITE_LE32(
      central_header + MZ_ZIP_CDH_LOCAL_HEADER_OFS, local_dir_header_ofs);
    if (!mz_zip_array_push_back(
          pZip,
          &pState->m_central_dir,
          central_header,
          MZ_ZIP_CENTRAL_DIR_HEADER_SIZE))
      return MZ_FALSE;

    n = MZ_READ_LE16(pSrc_central_header + MZ_ZIP_CDH_FILENAME_LEN_OFS) +
      MZ_READ_LE16(pSrc_central_header + MZ_ZIP_CDH_EXTRA_LEN_OFS) +
      MZ_READ_LE16(pSrc_central_header + MZ_ZIP_CDH_COMMENT_LEN_OFS);
    if (!mz_zip_array_push_back(
          pZip,
          &pState->m_central_dir,
          pSrc_central_header + MZ_ZIP_CENTRAL_DIR_HEADER_SIZE,
          n))
    {
      mz_zip_array_resize(
        pZip, &pState->m_central_dir, orig_central_dir_size, MZ_FALSE);
      return MZ_FALSE;
    }

    if (pState->m_central_dir.m_size > 0xFFFFFFFF)
      return MZ_FALSE;
    n = (mz_uint32)orig_central_dir_size;
    if (!mz_zip_array_push_back(pZip, &pState->m_central_dir_offsets, &n, 1))
    {
      mz_zip_array_resize(
        pZip, &pState->m_central_dir, orig_central_dir_size, MZ_FALSE);
      return MZ_FALSE;
    }

    pZip->m_total_files++;
    pZip->m_archive_size = cur_dst_file_ofs;

    return MZ_TRUE;
  }

  mz_bool mz_zip_writer_finalize_archive(mz_zip_archive* pZip)
  {
    mz_zip_internal_state* pState;
    mz_uint64 central_dir_ofs, central_dir_size;
    mz_uint8 hdr[MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE];

    if (
      (!pZip) || (!pZip->m_pState) || (pZip->m_zip_mode != MZ_ZIP_MODE_WRITING))
      return MZ_FALSE;

    pState = pZip->m_pState;

    // no zip64 support yet
    if (
      (pZip->m_total_files > 0xFFFF) ||
      ((pZip->m_archive_size + pState->m_central_dir.m_size +
        MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIZE) > 0xFFFFFFFF))
      return MZ_FALSE;

    central_dir_ofs = 0;
    central_dir_size = 0;
    if (pZip->m_total_files)
    {
      // Write central directory
      central_dir_ofs = pZip->m_archive_size;
      central_dir_size = pState->m_central_dir.m_size;
      pZip->m_central_directory_file_ofs = central_dir_ofs;
      if (
        pZip->m_pWrite(
          pZip->m_pIO_opaque,
          central_dir_ofs,
          pState->m_central_dir.m_p,
          (size_t)central_dir_size) != central_dir_size)
        return MZ_FALSE;
      pZip->m_archive_size += central_dir_size;
    }

    // Write end of central directory record
    MZ_CLEAR_OBJ(hdr);
    MZ_WRITE_LE32(
      hdr + MZ_ZIP_ECDH_SIG_OFS, MZ_ZIP_END_OF_CENTRAL_DIR_HEADER_SIG);
    MZ_WRITE_LE16(
      hdr + MZ_ZIP_ECDH_CDIR_NUM_ENTRIES_ON_DISK_OFS, pZip->m_total_files);
    MZ_WRITE_LE16(
      hdr + MZ_ZIP_ECDH_CDIR_TOTAL_ENTRIES_OFS, pZip->m_total_files);
    MZ_WRITE_LE32(hdr + MZ_ZIP_ECDH_CDIR_SIZE_OFS, central_dir_size);
    MZ_WRITE_LE32(hdr + MZ_ZIP_ECDH_CDIR_OFS_OFS, central_dir_ofs);

    if (
      pZip->m_pWrite(
        pZip->m_pIO_opaque, pZip->m_archive_size, hdr, sizeof(hdr)) !=
      sizeof(hdr))
      return MZ_FALSE;
#      ifndef MINIZ_NO_STDIO
    if ((pState->m_pFile) && (MZ_FFLUSH(pState->m_pFile) == EOF))
      return MZ_FALSE;
#      endif // #ifndef MINIZ_NO_STDIO

    pZip->m_archive_size += sizeof(hdr);

    pZip->m_zip_mode = MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED;
    return MZ_TRUE;
  }

  mz_bool mz_zip_writer_finalize_heap_archive(
    mz_zip_archive* pZip, void** pBuf, size_t* pSize)
  {
    if ((!pZip) || (!pZip->m_pState) || (!pBuf) || (!pSize))
      return MZ_FALSE;
    if (pZip->m_pWrite != mz_zip_heap_write_func)
      return MZ_FALSE;
    if (!mz_zip_writer_finalize_archive(pZip))
      return MZ_FALSE;

    *pBuf = pZip->m_pState->m_pMem;
    *pSize = pZip->m_pState->m_mem_size;
    pZip->m_pState->m_pMem = NULL;
    pZip->m_pState->m_mem_size = pZip->m_pState->m_mem_capacity = 0;
    return MZ_TRUE;
  }

  mz_bool mz_zip_writer_end(mz_zip_archive* pZip)
  {
    mz_zip_internal_state* pState;
    mz_bool status = MZ_TRUE;
    if (
      (!pZip) || (!pZip->m_pState) || (!pZip->m_pAlloc) || (!pZip->m_pFree) ||
      ((pZip->m_zip_mode != MZ_ZIP_MODE_WRITING) &&
       (pZip->m_zip_mode != MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED)))
      return MZ_FALSE;

    pState = pZip->m_pState;
    pZip->m_pState = NULL;
    mz_zip_array_clear(pZip, &pState->m_central_dir);
    mz_zip_array_clear(pZip, &pState->m_central_dir_offsets);
    mz_zip_array_clear(pZip, &pState->m_sorted_central_dir_offsets);

#      ifndef MINIZ_NO_STDIO
    if (pState->m_pFile)
    {
      MZ_FCLOSE(pState->m_pFile);
      pState->m_pFile = NULL;
    }
#      endif // #ifndef MINIZ_NO_STDIO

    if ((pZip->m_pWrite == mz_zip_heap_write_func) && (pState->m_pMem))
    {
      pZip->m_pFree(pZip->m_pAlloc_opaque, pState->m_pMem);
      pState->m_pMem = NULL;
    }

    pZip->m_pFree(pZip->m_pAlloc_opaque, pState);
    pZip->m_zip_mode = MZ_ZIP_MODE_INVALID;
    return status;
  }

#      ifndef MINIZ_NO_STDIO
  mz_bool mz_zip_add_mem_to_archive_file_in_place(
    const char* pZip_filename,
    const char* pArchive_name,
    const void* pBuf,
    size_t buf_size,
    const void* pComment,
    mz_uint16 comment_size,
    mz_uint level_and_flags)
  {
    mz_bool status, created_new_archive = MZ_FALSE;
    mz_zip_archive zip_archive;
    struct MZ_FILE_STAT_STRUCT file_stat;
    MZ_CLEAR_OBJ(zip_archive);
    if ((int)level_and_flags < 0)
      level_and_flags = MZ_DEFAULT_LEVEL;
    if (
      (!pZip_filename) || (!pArchive_name) || ((buf_size) && (!pBuf)) ||
      ((comment_size) && (!pComment)) ||
      ((level_and_flags & 0xF) > MZ_UBER_COMPRESSION))
      return MZ_FALSE;
    if (!mz_zip_writer_validate_archive_name(pArchive_name))
      return MZ_FALSE;
    if (MZ_FILE_STAT(pZip_filename, &file_stat) != 0)
    {
      // Create a new archive.
      if (!mz_zip_writer_init_file(&zip_archive, pZip_filename, 0))
        return MZ_FALSE;
      created_new_archive = MZ_TRUE;
    }
    else
    {
      // Append to an existing archive.
      if (!mz_zip_reader_init_file(
            &zip_archive,
            pZip_filename,
            level_and_flags | MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY))
        return MZ_FALSE;
      if (!mz_zip_writer_init_from_reader(&zip_archive, pZip_filename))
      {
        mz_zip_reader_end(&zip_archive);
        return MZ_FALSE;
      }
    }
    status = mz_zip_writer_add_mem_ex(
      &zip_archive,
      pArchive_name,
      pBuf,
      buf_size,
      pComment,
      comment_size,
      level_and_flags,
      0,
      0);
    // Always finalize, even if adding failed for some reason, so we have a
    // valid central directory. (This may not always succeed, but we can try.)
    if (!mz_zip_writer_finalize_archive(&zip_archive))
      status = MZ_FALSE;
    if (!mz_zip_writer_end(&zip_archive))
      status = MZ_FALSE;
    if ((!status) && (created_new_archive))
    {
      // It's a new archive and something went wrong, so just delete it.
      int ignoredStatus = MZ_DELETE_FILE(pZip_filename);
      (void)ignoredStatus;
    }
    return status;
  }

  void* mz_zip_extract_archive_file_to_heap(
    const char* pZip_filename,
    const char* pArchive_name,
    size_t* pSize,
    mz_uint flags)
  {
    int file_index;
    mz_zip_archive zip_archive;
    void* p = NULL;

    if (pSize)
      *pSize = 0;

    if ((!pZip_filename) || (!pArchive_name))
      return NULL;

    MZ_CLEAR_OBJ(zip_archive);
    if (!mz_zip_reader_init_file(
          &zip_archive,
          pZip_filename,
          flags | MZ_ZIP_FLAG_DO_NOT_SORT_CENTRAL_DIRECTORY))
      return NULL;

    if (
      (file_index = mz_zip_reader_locate_file(
         &zip_archive, pArchive_name, NULL, flags)) >= 0)
      p = mz_zip_reader_extract_to_heap(&zip_archive, file_index, pSize, flags);

    mz_zip_reader_end(&zip_archive);
    return p;
  }

#      endif // #ifndef MINIZ_NO_STDIO

#    endif // #ifndef MINIZ_NO_ARCHIVE_WRITING_APIS

#  endif // #ifndef MINIZ_NO_ARCHIVE_APIS

#  ifdef __cplusplus
}
#  endif

#endif // MINIZ_HEADER_FILE_ONLY

/*
  This is free and unencumbered software released into the public domain.

  Anyone is free to copy, modify, publish, use, compile, sell, or
  distribute this software, either in source code form or as a compiled
  binary, for any purpose, commercial or non-commercial, and by any
  means.

  In jurisdictions that recognize copyright laws, the author or authors
  of this software dedicate any and all copyright interest in the
  software to the public domain. We make this dedication for the benefit
  of the public at large and to the detriment of our heirs and
  successors. We intend this dedication to be an overt act of
  relinquishment in perpetuity of all present and future rights to this
  software under copyright law.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
  OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.

  For more information, please refer to <http://unlicense.org/>
*/

namespace scenepic
{
  namespace miniz_cpp
  {
    namespace detail
    {
#ifdef _WIN32
      char directory_separator = '\\';
      char alt_directory_separator = '/';
#else
      char directory_separator = '/';
      char alt_directory_separator = '\\';
#endif

      std::string join_path(const std::vector<std::string>& parts)
      {
        std::string joined;
        std::size_t i = 0;
        for (auto part : parts)
        {
          joined.append(part);

          if (i++ != parts.size() - 1)
          {
            joined.append(1, '/');
          }
        }
        return joined;
      }

      std::vector<std::string>
      split_path(const std::string& path, char delim = directory_separator)
      {
        std::vector<std::string> split;
        std::string::size_type previous_index = 0;
        auto separator_index = path.find(delim);

        while (separator_index != std::string::npos)
        {
          auto part =
            path.substr(previous_index, separator_index - previous_index);
          if (part != "..")
          {
            split.push_back(part);
          }
          else
          {
            split.pop_back();
          }
          previous_index = separator_index + 1;
          separator_index = path.find(delim, previous_index);
        }

        split.push_back(path.substr(previous_index));

        if (split.size() == 1 && delim == directory_separator)
        {
          auto alternative = split_path(path, alt_directory_separator);
          if (alternative.size() > 1)
          {
            return alternative;
          }
        }

        return split;
      }

      uint32_t crc32buf(const char* buf, std::size_t len)
      {
        uint32_t oldcrc32 = 0xFFFFFFFF;

        uint32_t crc_32_tab[] = {
          /* CRC polynomial 0xedb88320 */
          0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
          0x706af48f, 0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4,
          0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07,
          0x90bf1d91, 0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
          0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856,
          0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
          0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
          0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
          0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3,
          0x45df5c75, 0xdcd60dcf, 0xabd13d59, 0x26d930ac, 0x51de003a,
          0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599,
          0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
          0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
          0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f,
          0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0x0f00f934, 0x9609a88e,
          0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
          0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed,
          0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
          0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
          0xfbd44c65, 0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
          0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a,
          0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5,
          0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa, 0xbe0b1010,
          0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
          0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
          0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6,
          0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615,
          0x73dc1683, 0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
          0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1, 0xf00f9344,
          0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
          0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
          0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
          0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1,
          0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c,
          0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef,
          0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
          0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
          0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31,
          0x2cd99e8b, 0x5bdeae1d, 0x9b64c2b0, 0xec63f226, 0x756aa39c,
          0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
          0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b,
          0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
          0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
          0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
          0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 0xa00ae278,
          0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7,
          0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66,
          0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
          0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
          0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8,
          0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b,
          0x2d02ef8d};

#define UPDC32(octet, crc) \
  (crc_32_tab[((crc) ^ static_cast<uint8_t>(octet)) & 0xff] ^ ((crc) >> 8))

        for (; len; --len, ++buf)
        {
          oldcrc32 = UPDC32(*buf, oldcrc32);
        }

        return ~oldcrc32;
      }

      tm safe_localtime(const time_t& t)
      {
#ifdef _WIN32
        tm time;
        localtime_s(&time, &t);
        return time;
#else
        tm* time = localtime(&t);
        assert(time != nullptr);
        return *time;
#endif
      }

      std::size_t write_callback(
        void* opaque, mz_uint64 file_ofs, const void* pBuf, std::size_t n)
      {
        auto buffer = static_cast<std::vector<char>*>(opaque);

        if (file_ofs + n > buffer->size())
        {
          auto new_size =
            static_cast<std::vector<char>::size_type>(file_ofs + n);
          buffer->resize(new_size);
        }

        for (std::size_t i = 0; i < n; i++)
        {
          (*buffer)[static_cast<std::size_t>(file_ofs + i)] =
            (static_cast<const char*>(pBuf))[i];
        }

        return n;
      }

    } // namespace detail

    struct zip_info
    {
      std::string filename;

      struct
      {
        int year = 1980;
        int month = 0;
        int day = 0;
        int hours = 0;
        int minutes = 0;
        int seconds = 0;
      } date_time;

      std::string comment;
      std::string extra;
      uint16_t create_system = 0;
      uint16_t create_version = 0;
      uint16_t extract_version = 0;
      uint16_t flag_bits = 0;
      std::size_t volume = 0;
      uint32_t internal_attr = 0;
      uint32_t external_attr = 0;
      std::size_t header_offset = 0;
      uint32_t crc = 0;
      std::size_t compress_size = 0;
      std::size_t file_size = 0;
    };

    class zip_file
    {
    public:
      zip_file() : archive_(new mz_zip_archive())
      {
        reset();
      }

      zip_file(const std::string& filename) : zip_file()
      {
        load(filename);
      }

      zip_file(std::istream& stream) : zip_file()
      {
        load(stream);
      }

      zip_file(const std::vector<unsigned char>& bytes) : zip_file()
      {
        load(bytes);
      }

      ~zip_file()
      {
        reset();
      }

      void load(std::istream& stream)
      {
        reset();
        buffer_.assign(
          std::istreambuf_iterator<char>(stream),
          std::istreambuf_iterator<char>());
        remove_comment();
        start_read();
      }

      void load(const std::string& filename)
      {
        filename_ = filename;
        std::ifstream stream(filename, std::ios::binary);
        load(stream);
      }

      void load(const std::vector<unsigned char>& bytes)
      {
        reset();
        buffer_.assign(bytes.begin(), bytes.end());
        remove_comment();
        start_read();
      }

      void save(const std::string& filename)
      {
        filename_ = filename;
        std::ofstream stream(filename, std::ios::binary);
        save(stream);
      }

      void save(std::ostream& stream)
      {
        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING)
        {
          mz_zip_writer_finalize_archive(archive_.get());
        }

        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED)
        {
          mz_zip_writer_end(archive_.get());
        }

        if (archive_->m_zip_mode == MZ_ZIP_MODE_INVALID)
        {
          start_read();
        }

        append_comment();
        stream.write(buffer_.data(), static_cast<long>(buffer_.size()));
      }

      void save(std::vector<unsigned char>& bytes)
      {
        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING)
        {
          mz_zip_writer_finalize_archive(archive_.get());
        }

        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED)
        {
          mz_zip_writer_end(archive_.get());
        }

        if (archive_->m_zip_mode == MZ_ZIP_MODE_INVALID)
        {
          start_read();
        }

        append_comment();
        bytes.assign(buffer_.begin(), buffer_.end());
      }

      void reset()
      {
        switch (archive_->m_zip_mode)
        {
          case MZ_ZIP_MODE_READING:
            mz_zip_reader_end(archive_.get());
            break;
          case MZ_ZIP_MODE_WRITING:
            mz_zip_writer_finalize_archive(archive_.get());
            mz_zip_writer_end(archive_.get());
            break;
          case MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED:
            mz_zip_writer_end(archive_.get());
            break;
          case MZ_ZIP_MODE_INVALID:
            break;
        }

        if (archive_->m_zip_mode != MZ_ZIP_MODE_INVALID)
        {
          throw std::runtime_error("");
        }

        buffer_.clear();
        comment.clear();

        start_write();
        mz_zip_writer_finalize_archive(archive_.get());
        mz_zip_writer_end(archive_.get());
      }

      bool has_file(const std::string& name)
      {
        if (archive_->m_zip_mode != MZ_ZIP_MODE_READING)
        {
          start_read();
        }

        int index =
          mz_zip_reader_locate_file(archive_.get(), name.c_str(), nullptr, 0);

        return index != -1;
      }

      bool has_file(const zip_info& name)
      {
        return has_file(name.filename);
      }

      zip_info getinfo(const std::string& name)
      {
        if (archive_->m_zip_mode != MZ_ZIP_MODE_READING)
        {
          start_read();
        }

        int index =
          mz_zip_reader_locate_file(archive_.get(), name.c_str(), nullptr, 0);

        if (index == -1)
        {
          throw std::runtime_error("not found");
        }

        return getinfo(index);
      }

      std::vector<zip_info> infolist()
      {
        if (archive_->m_zip_mode != MZ_ZIP_MODE_READING)
        {
          start_read();
        }

        std::vector<zip_info> info;

        for (std::size_t i = 0; i < mz_zip_reader_get_num_files(archive_.get());
             i++)
        {
          info.push_back(getinfo(static_cast<int>(i)));
        }

        return info;
      }

      std::vector<std::string> namelist()
      {
        std::vector<std::string> names;

        for (auto& info : infolist())
        {
          names.push_back(info.filename);
        }

        return names;
      }

      std::ostream& open(const std::string& name)
      {
        return open(getinfo(name));
      }

      std::ostream& open(const zip_info& name)
      {
        auto data = read(name);
        std::string data_string(data.begin(), data.end());
        open_stream_ << data_string;
        return open_stream_;
      }

      void extract(const std::string& member, const std::string& path)
      {
        std::fstream stream(
          detail::join_path({path, member}), std::ios::binary | std::ios::out);
        stream << open(member).rdbuf();
      }

      void extract(const zip_info& member, const std::string& path)
      {
        std::fstream stream(
          detail::join_path({path, member.filename}),
          std::ios::binary | std::ios::out);
        stream << open(member).rdbuf();
      }

      void extractall(const std::string& path)
      {
        extractall(path, infolist());
      }

      void extractall(
        const std::string& path, const std::vector<std::string>& members)
      {
        for (auto& member : members)
        {
          extract(member, path);
        }
      }

      void
      extractall(const std::string& path, const std::vector<zip_info>& members)
      {
        for (auto& member : members)
        {
          extract(member, path);
        }
      }

      void printdir()
      {
        printdir(std::cout);
      }

      void printdir(std::ostream& stream)
      {
        stream << "  Length "
               << "  "
               << "   "
               << "Date"
               << "   "
               << " "
               << "Time "
               << "   "
               << "Name" << std::endl;
        stream << "---------  ---------- -----   ----" << std::endl;

        std::size_t sum_length = 0;
        std::size_t file_count = 0;

        for (auto& member : infolist())
        {
          sum_length += member.file_size;
          file_count++;

          std::string length_string = std::to_string(member.file_size);
          while (length_string.length() < 9)
          {
            length_string = " " + length_string;
          }
          stream << length_string;

          stream << "  ";
          stream << (member.date_time.month < 10 ? "0" : "")
                 << member.date_time.month;
          stream << "/";
          stream << (member.date_time.day < 10 ? "0" : "")
                 << member.date_time.day;
          stream << "/";
          stream << member.date_time.year;
          stream << " ";
          stream << (member.date_time.hours < 10 ? "0" : "")
                 << member.date_time.hours;
          stream << ":";
          stream << (member.date_time.minutes < 10 ? "0" : "")
                 << member.date_time.minutes;
          stream << "   ";
          stream << member.filename;
          stream << std::endl;
        }

        stream << "---------                     -------" << std::endl;

        std::string length_string = std::to_string(sum_length);
        while (length_string.length() < 9)
        {
          length_string = " " + length_string;
        }
        stream << length_string << "                     " << file_count << " "
               << (file_count == 1 ? "file" : "files");
        stream << std::endl;
      }

      std::string read(const zip_info& info)
      {
        std::size_t size;
        char* data = static_cast<char*>(mz_zip_reader_extract_file_to_heap(
          archive_.get(), info.filename.c_str(), &size, 0));
        if (data == nullptr)
        {
          throw std::runtime_error("file couldn't be read");
        }
        std::string extracted(data, data + size);
        mz_free(data);
        return extracted;
      }

      std::string read(const std::string& name)
      {
        return read(getinfo(name));
      }

      std::pair<bool, std::string> testzip()
      {
        if (archive_->m_zip_mode == MZ_ZIP_MODE_INVALID)
        {
          throw std::runtime_error("not open");
        }

        for (auto& file : infolist())
        {
          auto content = read(file);
          auto crc = detail::crc32buf(content.c_str(), content.size());

          if (crc != file.crc)
          {
            return {false, file.filename};
          }
        }

        return {true, ""};
      }

      void write(const std::string& filename)
      {
        auto split = detail::split_path(filename);
        if (split.size() > 1)
        {
          split.erase(split.begin());
        }
        auto arcname = detail::join_path(split);
        write(filename, arcname);
      }

      void write(const std::string& filename, const std::string& arcname)
      {
        std::fstream file(filename, std::ios::binary | std::ios::in);
        std::stringstream ss;
        ss << file.rdbuf();
        std::string bytes = ss.str();

        writestr(arcname, bytes);
      }

      void writestr(const std::string& arcname, const std::string& bytes)
      {
        if (archive_->m_zip_mode != MZ_ZIP_MODE_WRITING)
        {
          start_write();
        }

        if (!mz_zip_writer_add_mem(
              archive_.get(),
              arcname.c_str(),
              bytes.data(),
              bytes.size(),
              MZ_BEST_COMPRESSION))
        {
          throw std::runtime_error("write error");
        }
      }

      void writestr(const zip_info& info, const std::string& bytes)
      {
        if (info.filename.empty() || info.date_time.year < 1980)
        {
          throw std::runtime_error(
            "must specify a filename and valid date (year >= 1980");
        }

        if (archive_->m_zip_mode != MZ_ZIP_MODE_WRITING)
        {
          start_write();
        }

        auto crc = detail::crc32buf(bytes.c_str(), bytes.size());

        if (!mz_zip_writer_add_mem_ex(
              archive_.get(),
              info.filename.c_str(),
              bytes.data(),
              bytes.size(),
              info.comment.c_str(),
              static_cast<mz_uint16>(info.comment.size()),
              MZ_BEST_COMPRESSION,
              0,
              crc))
        {
          throw std::runtime_error("write error");
        }
      }

      std::string get_filename() const
      {
        return filename_;
      }

      std::string comment;

    private:
      void start_read()
      {
        if (archive_->m_zip_mode == MZ_ZIP_MODE_READING)
          return;

        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING)
        {
          mz_zip_writer_finalize_archive(archive_.get());
        }

        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED)
        {
          mz_zip_writer_end(archive_.get());
        }

        if (!mz_zip_reader_init_mem(
              archive_.get(), buffer_.data(), buffer_.size(), 0))
        {
          throw std::runtime_error("bad zip");
        }
      }

      void start_write()
      {
        if (archive_->m_zip_mode == MZ_ZIP_MODE_WRITING)
          return;

        switch (archive_->m_zip_mode)
        {
          case MZ_ZIP_MODE_READING:
          {
            mz_zip_archive archive_copy;
            std::memset(&archive_copy, 0, sizeof(mz_zip_archive));
            std::vector<char> buffer_copy(buffer_.begin(), buffer_.end());

            if (!mz_zip_reader_init_mem(
                  &archive_copy, buffer_copy.data(), buffer_copy.size(), 0))
            {
              throw std::runtime_error("bad zip");
            }

            mz_zip_reader_end(archive_.get());

            archive_->m_pWrite = &detail::write_callback;
            archive_->m_pIO_opaque = &buffer_;
            buffer_ = std::vector<char>();

            if (!mz_zip_writer_init(archive_.get(), 0))
            {
              throw std::runtime_error("bad zip");
            }

            for (unsigned int i = 0;
                 i < static_cast<unsigned int>(archive_copy.m_total_files);
                 i++)
            {
              if (!mz_zip_writer_add_from_zip_reader(
                    archive_.get(), &archive_copy, i))
              {
                throw std::runtime_error("fail");
              }
            }

            mz_zip_reader_end(&archive_copy);
            return;
          }
          case MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED:
            mz_zip_writer_end(archive_.get());
            break;
          case MZ_ZIP_MODE_INVALID:
          case MZ_ZIP_MODE_WRITING:
            break;
        }

        archive_->m_pWrite = &detail::write_callback;
        archive_->m_pIO_opaque = &buffer_;

        if (!mz_zip_writer_init(archive_.get(), 0))
        {
          throw std::runtime_error("bad zip");
        }
      }

      void append_comment()
      {
        if (!comment.empty())
        {
          auto comment_length = std::min(
            static_cast<uint16_t>(comment.length()),
            std::numeric_limits<uint16_t>::max());
          buffer_[buffer_.size() - 2] = static_cast<char>(comment_length);
          buffer_[buffer_.size() - 1] = static_cast<char>(comment_length >> 8);
          std::copy(
            comment.begin(), comment.end(), std::back_inserter(buffer_));
        }
      }

      void remove_comment()
      {
        if (buffer_.empty())
          return;

        std::size_t position = buffer_.size() - 1;

        for (; position >= 3; position--)
        {
          if (
            buffer_[position - 3] == 'P' && buffer_[position - 2] == 'K' &&
            buffer_[position - 1] == '\x05' && buffer_[position] == '\x06')
          {
            position = position + 17;
            break;
          }
        }

        if (position == 3)
        {
          throw std::runtime_error(
            "didn't find end of central directory signature");
        }

        uint16_t length = static_cast<uint16_t>(buffer_[position + 1]);
        length = static_cast<uint16_t>(length << 8) +
          static_cast<uint16_t>(buffer_[position]);
        position += 2;

        if (length != 0)
        {
          comment = std::string(
            buffer_.data() + position, buffer_.data() + position + length);
          buffer_.resize(buffer_.size() - length);
          buffer_[buffer_.size() - 1] = 0;
          buffer_[buffer_.size() - 2] = 0;
        }
      }

      zip_info getinfo(int index)
      {
        if (archive_->m_zip_mode != MZ_ZIP_MODE_READING)
        {
          start_read();
        }

        mz_zip_archive_file_stat stat;
        mz_zip_reader_file_stat(
          archive_.get(), static_cast<mz_uint>(index), &stat);

        zip_info result;

        result.filename = std::string(
          stat.m_filename, stat.m_filename + std::strlen(stat.m_filename));
        result.comment =
          std::string(stat.m_comment, stat.m_comment + stat.m_comment_size);
        result.compress_size = static_cast<std::size_t>(stat.m_comp_size);
        result.file_size = static_cast<std::size_t>(stat.m_uncomp_size);
        result.header_offset =
          static_cast<std::size_t>(stat.m_local_header_ofs);
        result.crc = stat.m_crc32;
        auto time = detail::safe_localtime(stat.m_time);
        result.date_time.year = 1900 + time.tm_year;
        result.date_time.month = 1 + time.tm_mon;
        result.date_time.day = time.tm_mday;
        result.date_time.hours = time.tm_hour;
        result.date_time.minutes = time.tm_min;
        result.date_time.seconds = time.tm_sec;
        result.flag_bits = stat.m_bit_flag;
        result.internal_attr = stat.m_internal_attr;
        result.external_attr = stat.m_external_attr;
        result.extract_version = stat.m_version_needed;
        result.create_version = stat.m_version_made_by;
        result.volume = stat.m_file_index;
        result.create_system = stat.m_method;

        return result;
      }

      std::unique_ptr<mz_zip_archive> archive_;
      std::vector<char> buffer_;
      std::stringstream open_stream_;
      std::string filename_;
    };

  } // namespace miniz_cpp
} // namespace scenepic
